/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008
*
****************************************************************************/
package inversion_applet;

import java.util.ArrayList;
import java.io.File;
import javax.swing.JTextArea;
import javax.swing.JTable;
import javax.swing.JPanel;

/************************************************************************
* This class contains all of the data for the applet.  By centralising
* the data in this applet, it becomes easier and more transparent for
* the different parts to communicate.
************************************************************************/
public class DataCenter {

  /**********************************************************************
  * A constructor.
  **********************************************************************/
  public DataCenter() {
    // variables relating to rotational inversion:
    n_gridR = 200;
    RLambda = 1e-3;
    RTheta = 0.1;
    RDelta = 0.2;
    RNoiseLevel = 1e-3;
    RNoiseSelected = false;
    RSloaded = false;
    targetRotaLoaded = false;
    rotationInversionUpToDate = true;

    // variables relating to structural inversion:
    kernelType = kernelTypes[0];
    n_gridS = 200;
    n_surf = 6;
    SLambda = 1e-3;
    Beta1 = 1.0;
    Beta2 = 1.0;
    STheta1 = 0.1;
    STheta2 = 0.1;
    SDelta1 = 0.2;
    SDelta2 = 0.2;
    SNoiseLevel = 1e-3;
    SNoiseSelected = false;
    DFloaded = false;
    targetStructLoaded = false;
    structInversionUpToDate = true;

    // other variables:
    modelLoaded = false;
    kernelsLoaded = false;
    currentDirectory = new File("."); // this doesn't seem to work on a Mac:
  }

  /**********************************************************************
  * This tells whether the program is running as an applet (isApplet ==
  * true) or as a stand-alone application (isApplet == false).
  **********************************************************************/
  public boolean isApplet;

  /**********************************************************************
  * This keeps track of the current directory, which is used when opening
  * and saving files.
  **********************************************************************/
  public File currentDirectory;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Information on observed pulsation modes (frequency shifts +
  // mode identification):
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** This variable is set to 'true' when frequency shifts are loaded and
  *   'false' otherwise. */
  public boolean DFloaded;
  /** Radial order associated with a frequency shift data */
  public ArrayList<Integer> nDF = new ArrayList<Integer>();
  /** Harmonic degree associated with a frequency shift data */
  public ArrayList<Integer> lDF = new ArrayList<Integer>();
  /** DF = Observed frequency - theoretical frequency from unperturbed
  *   model. */
  public ArrayList<Double> DF = new ArrayList<Double>();
  /** Error associated with a frequency shift data */
  public ArrayList<Double> ErrorDF = new ArrayList<Double>();
  /** This is used for sorting the frequencie shift data according to (n,l) */
  public ArrayList<Integer> indexDF = new ArrayList<Integer>();
  /** This keeps track of the frequency shifts page so that other
  *   components can update this page. */
  public JPanel frequencyShiftsPage;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Information on rotation splittings ( + mode identification):
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** This variable is set to 'true' when rotational splittings are loaded
  *   and 'false' otherwise. */
  public boolean RSloaded;
  /** Radial order associated with a rotational splitting */
  public ArrayList<Integer> nRS = new ArrayList<Integer>();
  /** Harmonic degree associated with a rotational splitting */
  public ArrayList<Integer> lRS = new ArrayList<Integer>();
  /** RS = average splitting in a rotationally split multiplet.
  *   For prograde modes, the frequency is increased by |m|RS */
  public ArrayList<Double> RS = new ArrayList<Double>();
  /** Error on observed frequency */
  public ArrayList<Double> ErrorRS = new ArrayList<Double>();
  /** This is used for sorting the rotational splittings according to
  *   (n,l) */
  public ArrayList<Integer> indexRS = new ArrayList<Integer>();
  /** This keeps track of the rotational splittings page so that other
  *   components can update this page. */
  public JPanel rotationalSplittingsPage;


  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Variables with information on kernels:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
 
  /** This variable is set to 'true' when kernels are loaded and
  *   'false' otherwise. */
  public boolean kernelsLoaded;
  /** Radial order of a specific kernel */
  public int[] nK;
  /** Harmonic degree of a specific kernel */
  public int[] lK;
  /** frequency a specific eigen function */
  public double[] freq;
  public double[] freqDimless;
  public double[] Enl;

  /** Rotational kernel. First index = kernel number, Second index =
   *  grid position. */
  public double[][] rotaKernel;

 /* ______________________________________________________ */

	public double[][] c2rhoKernel;
	public double[][] rhoc2Kernel;

	public double[] EnlCoefficient;
	public double[] z1max;

  /* ______________________________________________________ */

	public double EigenFuncStellarRadius;
	public double EigenFuncStellarMass;

 /* ______________________________________________________ */

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Variables with information on eigenmodes:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  
  /** Underlying grid */
  public double[] grid;
  /** Horizontal displacement. First index = kernel number, Second index =
   *  grid position.  */
  public double[][] hDisp;
  /** Vertical displacement. First index = kernel number, Second index =
   *  grid position.  */
  public double[][] vDisp;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Variables relating to the model:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  
  /** This variable is set to 'true' when a stellar model is loaded and
  *   'false' otherwise. */
  public boolean modelLoaded;
  /** Underlying grid */
  public double[] gridModel;
  /** The stellar radius (in cm) */
  public double StellarRadius;
  /** The stellar mass (in g) */
  public double StellarMass;
  /** Reference value for the pressure. Pref*pressure[] = pressure (in g.cm^-1.s^-2 or dyn.cm^-2) */
  public double Pref;
  /** Reference value for the density. RhoRef*density[] = density (in g.cm^-3) */
  public double RhoRef;

  /** Gamma1 profile. Gamma1 is the adiabatic exponent. */
  public double[] Gamma1;
  /** Normalised density profile */
  public double[] rho;
  /** Normalised pressure profile*/
  public double[] pressure;
  /** Normalised squared sound velocity profile */
  public double[] c2;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Physical constants:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** The gravitational constant (in cm^3.g^-1.s^-2) */
  public static final double G = 6.6716823e-8;
  /** The solar mass in g */
  public static final double SolarMass = 1.98919e33; 
  /** The solar radius in cm */
  public static final double SolarRadius = 6.9599e10;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Variables relating to the rotation inversion:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** This variable states whether or not the rotational inversion is up
  *   to date or not (i.e. that a suitable list of frequency shifts and
  *   kernels  exists and agrees with current variable contents). */
  public boolean rotationInversionUpToDate;
  /** This contains the grid used in the inversion.  It doesn't need
  *   to contain more than 200 points and should be spaced such that
  *   c(r)Delta(r) = cnst */
  public double[] gridRInversion;
  /** The number of points used in the above grid */
  public int n_gridR;
  /** This regulates the trade-off between an exact fit to data and
  *   regularisation of the solution in the RLS method. */
  public double RLambda;
  /** This regulates the trade-off between fitting to the data points
  *   data points and regularising the solution in the OLA method. */
  public double RTheta;
  /** This regulates the width of the Gaussian target functions
  *   in the SOLA method. */
  public double RDelta;   
  /** If this parameter is set to true, than artiticial noise will
  *   be added to the data and used in the inversion. */
  public boolean RNoiseSelected;
  /** This regulates the noise level */
  public double RNoiseLevel;
  /** This contains artificial noise to be added to the data */
  public double[] RNoise;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Variables relating to the structural inversion:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** This variable states whether or not the structural inversion
  *   is up to date or not (i.e. that a suitable list of frequency shifts
  *   and kernels exists and agrees with current variable contents). */
  public boolean structInversionUpToDate;
  /** This contains the grid used in the inversion.  It doesn't need
  *   to contain more than 200 points and should be spaced such that
  *   c(r)Delta(r) = cnst */
  public double[] gridSInversion;
  /** The number of points used in the above grid */
  public int n_gridS;
  /** Number of Legendre polynomials to be used in modelling surface terms */
  public int n_surf;
  /** This regulates the trade-off between an exact fit to data and
  *   regularisation of the solution in the RLS method. */
  public double SLambda;
  /** This regulates the trade-off between fitting to the data points
  *   data points and regularising the solution in the OLA method. */
  public double STheta1;
  public double STheta2;
  /** This regulates the width of the Gaussian target functions
  *   in the SOLA method. */
  public double SDelta1;
  public double SDelta2;
  /** This regulates the balance between the obtaining "nice" averaging
  *   kernels and reducing cross-term kernels (see e.g.
  *   Rabello-Soares et al., 1999) */
  public double Beta1;
  public double Beta2;
  /** The following string tells type of kernels to use */
  public String kernelType;
  /** The following arrays contain the kernels relevent to the kernel
  *   type selected in the above string */
  public double[][] struct1Kernel;
  public double[][] struct2Kernel;
  /** If this parameter is set to true, than artiticial noise will
  *   be added to the data and used in the inversion. */
  public boolean SNoiseSelected;
  /** This regulates the noise level */
  public double SNoiseLevel;
  /** This contains artificial noise to be added to the data */
  public double[] SNoise;
  
  /** The list of different kernel types */
  public static final String[] kernelTypes = {"c2-rho kernels"};

  /** This keeps track of whether an extra constraint which keeps the
  *   mass of the star constant is needed or not. 0 means no contraints,
  *   1 means a constraint on the first structural inversion, 2 means
  *   a contraint on the second structural inversion, and 3 means a
  *   constraint on both inversions*/
  public int constantMass;

  ///////////////////////////////////////////////////////////////////////
  /** The following method sets up struct1Kernel and struct2Kernel based
  *   on the choice given in kernelType */
  ///////////////////////////////////////////////////////////////////////
  public void findStructKernels() {
    if (!kernelsLoaded) {
      struct1Kernel = null;
      struct2Kernel = null;
      constantMass = 0;
      return;
    }
    
    if (kernelType.equals(kernelTypes[0])) {
      struct1Kernel = c2rhoKernel;
      struct2Kernel = rhoc2Kernel;
      constantMass = 2;
      return;
    }
    
    // the following should never happen ...
    struct1Kernel = null;
    struct2Kernel = null;
  }

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Target functions:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /** This variable is set to 'true' when a targetRota function is
  *   loaded. */
  public boolean targetRotaLoaded;
  /** This contains an optional target rota function which can be compared
  *   with the inversion results, or used to calculate theoretical
  *   frequency shifts */
  public double[] targetRota;
  /** x grid corresponding to the target rota function */
  public double[] gridTargetRota;
  /** This variable is set to 'true' when a target structural functions are
  *   loaded. */
  public boolean targetStructLoaded;
  /** This contains an optional target struct1 function which can be compared
  *   with the inversion results, or used to calculate theoretical
  *   frequency shifts */
  public double[] targetStruct1;
  /** This contains an option target struct2 function which can be compared
  *   with the inversion results, or used to calculate theoretical
  *   frequency shifts */
  public double[] targetStruct2;
  /** x grid corresponding to the target structural functions */
  public double[] gridTargetStruct;

  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////
  // Various utility methods:
  ///////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////

  /******************************************************************************
  * This method produces an index array such that Kernel[index[i]] is sorted
  * according to n and l.  The algorithm is based on the heapsort method.  It is
  * copied and adapted from:
  *
  * "Numerical Recipes in Fortran: The Art of Scientific Computing"
  * Second Edition.
  * William H. Press, Saul A. Teukolsky, William T. Vetterling, Brian P. Flannery
  * Cambridge University Press, 1992.
  ******************************************************************************/

  public int[] sortKernels() {
    if (!kernelsLoaded) return null;
    int n = nK.length;
    int[] index = new int[n];

    // Initialise the index array
    for(int i=0; i<n; i++) index[i] = i;
    
    if (n < 2) return index;

    // The index l will be decremented from its initial value down to 0
    // during the "hiring" (heap creation) phase.  Once it reaches 0, the
    // index ir will be decremented from its initial value down to 0 during
    // the "retirement-and-promotion" (heap selection) phase.

    int l = n/2;
    int ir = n-1;
    int i, j, iindex;
    while(true) {
      if (l > 0) {                    // Still in hiring phase
        l--;
        iindex = index[l];
      } else {                        // In retirement-and-promotion phase.
        iindex = index[ir];           // Clear a space at end of array.
        index[ir] = index[0];         // Retire the top of the heap into it
        ir--;                         // Decrease the size of the corporation.
        if(ir == 0) {                 // Done with the last promotion.
          index[0] = iindex;          // The least competent worker of all!
          return index;
        }
      }
      i=l;                            // Whether in the hiring phase or
                                      // promotion phase, we here set up to
                                      // sift down element rra to its proper
                                      // level.
      j=l+l+1;
      while (j <= ir) {
        if (j < ir) {
          if (compareKernel(index[j],index[j+1])) j++;
                                      // Compare to the better underling
        }
        if (compareKernel(iindex,index[j])) {
          index[i]=index[j];          // Demote ra(iindex)
          i=j;
          j=j+j+1;
        } else {                      // This is rra's level.  Set j to
                                      // terminate the sift-down.
          j=ir+1;
        }
      }
      index[i] = iindex;              // Put rra into its slot.
    }
  }
  
  // This is method returns "true" if Kernel[i1] is "smaller" than
  // Kernel[i2] (i.e. if (nK[i1],lK[i1]) commes before (nK[i2],lK[i2])
  // in lexico-graphical order).
  private boolean compareKernel(int i1, int i2) {
    if (lK[i1] < lK[i2]) return true;
    if ((lK[i1] == lK[i2]) && (nK[i1] < nK[i2])) return true;
    return false;
  }

  /******************************************************************************
  * This method produces an index array such that DF[index[i]] is sorted
  * according to n and l.  The algorithm is based on the heapsort method.  It is
  * copied and adapted from:
  *
  * "Numerical Recipes in Fortran: The Art of Scientific Computing"
  * Second Edition.
  * William H. Press, Saul A. Teukolsky, William T. Vetterling, Brian P. Flannery
  * Cambridge University Press, 1992.
  ******************************************************************************/

  public int[] sortDF() {
    if (!DFloaded) return null;
    int n = nDF.size();
    int[] index = new int[n];

    // Initialise the index array
    for(int i=0; i<n; i++) index[i] = i;
    
    if (n < 2) return index;

    // The index l will be decremented from its initial value down to 0
    // during the "hiring" (heap creation) phase.  Once it reaches 0, the
    // index ir will be decremented from its initial value down to 0 during
    // the "retirement-and-promotion" (heap selection) phase.

    int l = n/2;
    int ir = n-1;
    int i, j, iindex;
    while(true) {
      if (l > 0) {                    // Still in hiring phase
        l--;
        iindex = index[l];
      } else {                        // In retirement-and-promotion phase.
        iindex = index[ir];           // Clear a space at end of array.
        index[ir] = index[0];         // Retire the top of the heap into it
        ir--;                         // Decrease the size of the corporation.
        if(ir == 0) {                 // Done with the last promotion.
          index[0] = iindex;          // The least competent worker of all!
          return index;
        }
      }
      i=l;                            // Whether in the hiring phase or
                                      // promotion phase, we here set up to
                                      // sift down element rra to its proper
                                      // level.
      j=l+l+1;
      while (j <= ir) {
        if (j < ir) {
          if (compareDF(index[j],index[j+1])) j++;
                                      // Compare to the better underling
        }
        if (compareDF(iindex,index[j])) {
          index[i]=index[j];          // Demote ra(iindex)
          i=j;
          j=j+j+1;
        } else {                      // This is rra's level.  Set j to
                                      // terminate the sift-down.
          j=ir+1;
        }
      }
      index[i] = iindex;              // Put rra into its slot.
    }
  }
  
  // This is method returns "true" if DF[i1] is "smaller" than
  // Kernel[i2] (i.e. if (lDF[i1],nDF[i1]) commes before
  // (lDF[i2],nDF[i2]) in lexico-graphical order).
  private boolean compareDF(int i1, int i2) {
    int compareL = compareIntegers(lDF.get(i1),lDF.get(i2));
    if (compareL < 0) return true;
    if (compareL == 0) {
      if (compareIntegers(nDF.get(i1),nDF.get(i2)) <= 0) return true;
    }
    return false;
  }

  /******************************************************************************
  * This method produces an index array such that RS[index[i]] is sorted
  * according to n and l.  The algorithm is based on the heapsort method.  It is
  * copied and adapted from:
  *
  * "Numerical Recipes in Fortran: The Art of Scientific Computing"
  * Second Edition.
  * William H. Press, Saul A. Teukolsky, William T. Vetterling, Brian P. Flannery
  * Cambridge University Press, 1992.
  ******************************************************************************/

  public int[] sortRS() {
    if (!RSloaded) return null;
    int n = nRS.size();
    int[] index = new int[n];

    // Initialise the index array
    for(int i=0; i<n; i++) index[i] = i;
    
    if (n < 2) return index;

    // The index l will be decremented from its initial value down to 0
    // during the "hiring" (heap creation) phase.  Once it reaches 0, the
    // index ir will be decremented from its initial value down to 0 during
    // the "retirement-and-promotion" (heap selection) phase.

    int l = n/2;
    int ir = n-1;
    int i, j, iindex;
    while(true) {
      if (l > 0) {                    // Still in hiring phase
        l--;
        iindex = index[l];
      } else {                        // In retirement-and-promotion phase.
        iindex = index[ir];           // Clear a space at end of array.
        index[ir] = index[0];         // Retire the top of the heap into it
        ir--;                         // Decrease the size of the corporation.
        if(ir == 0) {                 // Done with the last promotion.
          index[0] = iindex;          // The least competent worker of all!
          return index;
        }
      }
      i=l;                            // Whether in the hiring phase or
                                      // promotion phase, we here set up to
                                      // sift down element rra to its proper
                                      // level.
      j=l+l+1;
      while (j <= ir) {
        if (j < ir) {
          if (compareRS(index[j],index[j+1])) j++;
                                      // Compare to the better underling
        }
        if (compareRS(iindex,index[j])) {
          index[i]=index[j];          // Demote ra(iindex)
          i=j;
          j=j+j+1;
        } else {                      // This is rra's level.  Set j to
                                      // terminate the sift-down.
          j=ir+1;
        }
      }
      index[i] = iindex;              // Put rra into its slot.
    }
  }
  
  // This is method returns "true" if RS[i1] is "smaller" than
  // RS[i2] (i.e. if (lRS[i1],nRS[i1]) commes before
  // (lRS[i2],nRS[i2]) in lexico-graphical order).
  private boolean compareRS(int i1, int i2) {
    int compareL = compareIntegers(lRS.get(i1),lRS.get(i2));
    if (compareL < 0) return true;
    if (compareL == 0) {
      if (compareIntegers(nRS.get(i1),nRS.get(i2)) <= 0) return true;
    }
    return false;
  }

  // This compares two Integers and treats null as smaller than
  // all other Integers.
  private int compareIntegers(Integer i, Integer j) {
    if (i == null) {
      if (j == null) {
        return 0;
      } else {
        return -1;
      }
    } else {
      if (j == null) {
        return 1;
      } else {
        return i.compareTo(j);
      }
    }
  }
}
