/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008
*
****************************************************************************/
package inversion_applet;

import inversion_applet.DataCenter;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.util.StringTokenizer;

public class FrequencyShiftsDataModel extends AbstractTableModel {

  // the data with the frequency shifts is in dataCenter;
  private DataCenter dataCenter;
 
  // These contain the parent components, which intervenes
  // when using a JPaneOption object.
  private JComponent parentComponent;

  
  // the names of the columns:
  private String[] columns = {"n","\u2113","\u0394f/f",
                              "Error(\u0394f/f)"};

  // the classes of the columns:
  private Class[] columnClasses = new Class[] 
    {Integer.class, Integer.class, Double.class, Double.class};
 
  /////////////////////////////////////////////////////////
  // a constructor:
  /////////////////////////////////////////////////////////
  protected FrequencyShiftsDataModel(DataCenter dataCenter,
                               JComponent parentComponent) {
    this.dataCenter = dataCenter;
    this.parentComponent = parentComponent;
  }
  
  /////////////////////////////////////////////////////////
  // clear all the data:
  /////////////////////////////////////////////////////////
  protected void clear() {
    dataCenter.nDF.clear();
    dataCenter.lDF.clear();
    dataCenter.DF.clear();
    dataCenter.ErrorDF.clear();
    dataCenter.indexDF.clear();
    dataCenter.structInversionUpToDate = false;
    dataCenter.DFloaded = false;
  }

  /////////////////////////////////////////////////////////
  // load a file:
  /////////////////////////////////////////////////////////
  protected void loadFile(String filename){
    BufferedReader f;
    String line;
    Boolean skipLine = false;
    int nDF, lDF;
    double DF, ErrorDF;
    int n = dataCenter.nDF.size();
    try {
      f = new BufferedReader(new FileReader(filename));
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
                    parentComponent,  // the parent component
                    "Unable to open file \""+filename+"\"",
                    "File not found",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }
    try {
      while (true) {
        line = f.readLine();     // read a line of text
        if (line == null) break; // test for the end of the buffer
        line = trim(line);       // remove comments
        StringTokenizer t = new StringTokenizer(line," ");
        // only read lines with EXACTLY 4 numbers
        if (t.countTokens() == 4) {
          nDF = Integer.parseInt(t.nextToken());
          lDF = Integer.parseInt(t.nextToken());
          DF  = Double.parseDouble(t.nextToken());
          ErrorDF = Double.parseDouble(t.nextToken());
          if (lDF >= 0) {
            dataCenter.nDF.add(nDF);
            dataCenter.lDF.add(lDF);
            dataCenter.DF.add(DF);
            dataCenter.ErrorDF.add(ErrorDF);
            dataCenter.indexDF.add(new Integer(n));
            n++;
          } else {
            skipLine = true;
          }
        } else {
          if (t.countTokens() != 0) skipLine = true;
        }
      }
      f.close();
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
                    parentComponent,  // the parent component
                    "Format not recognised",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }
    if (skipLine) {
      JOptionPane.showMessageDialog(
                    parentComponent,  // the parent component
                    "Some lines with either the wrong\n"+
                    "number of entries or with \u2113 < 0\n"+
                    "were skipped.",
                    "Warning",
                    JOptionPane.WARNING_MESSAGE
                  );
    }
    dataCenter.structInversionUpToDate = false;
    if (n > 0) dataCenter.DFloaded = true;
  }

  /////////////////////////////////////////////////////////
  // This remove comments (i.e. anything following a "#")
  // from a string and is used by the loadFile method:
  /////////////////////////////////////////////////////////
  private static String trim(String str) {
   int len = str.length();
    int i = 0;
    while ((i < len)&&(str.charAt(i) != '#')) i++;
    return str.substring(0,i);
  }
      
  /////////////////////////////////////////////////////////
  // write a file:
  /////////////////////////////////////////////////////////
  protected void writeFile(String filename) {
    File f = new File(filename);
    PrintWriter out;
    try {
      out = new PrintWriter(new FileWriter(f));
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        parentComponent,
        "Unable to write to file \""+filename+"\"",
        "Failure",
        JOptionPane.WARNING_MESSAGE
      );
      return;
    }
    out.println("# n l frequency_shift (f_obs - f_theo) error");
    for(int i = 0; i < dataCenter.nDF.size(); i++) {
      out.printf("%3d %3d %19.12e %19.12e%n",
                 dataCenter.nDF.get(dataCenter.indexDF.get(i).intValue()), 
                 dataCenter.lDF.get(dataCenter.indexDF.get(i).intValue()),
                 dataCenter.DF.get(dataCenter.indexDF.get(i).intValue()),
                 dataCenter.ErrorDF.get(dataCenter.indexDF.get(i).intValue()));
    }
    out.close();
  }


  /////////////////////////////////////////////////////////
  // the number of columns:
  /////////////////////////////////////////////////////////
  public int getColumnCount() {return 4;}
  
  /////////////////////////////////////////////////////////
  // the number of rows:
  /////////////////////////////////////////////////////////
  public int getRowCount() {return dataCenter.nDF.size();}
 
  /////////////////////////////////////////////////////////
  // return the value of a cell:
  /////////////////////////////////////////////////////////
  public Object getValueAt(int row, int col) {
    if (col == 0) return dataCenter.nDF.get(dataCenter.indexDF.get(row).intValue());
    if (col == 1) return dataCenter.lDF.get(dataCenter.indexDF.get(row).intValue());
    if (col == 2) return dataCenter.DF.get(dataCenter.indexDF.get(row).intValue());
    return dataCenter.ErrorDF.get(dataCenter.indexDF.get(row).intValue());
  }

  /////////////////////////////////////////////////////////
  // return the name of a column:
  /////////////////////////////////////////////////////////
  public String getColumnName(int column) {return columns[column];}
  
  /////////////////////////////////////////////////////////
  // return the class of a column:
  /////////////////////////////////////////////////////////
  public Class getColumnClass(int c) {return columnClasses[c];}
  
  /////////////////////////////////////////////////////////
  // this makes all of the cells editable:
  /////////////////////////////////////////////////////////
  public boolean isCellEditable(int row, int col) {return true;}

  /////////////////////////////////////////////////////////
  /** set the value of a particular cell: */
  /////////////////////////////////////////////////////////
  public void setValueAt(Object obj, int row, int col) {
    if (col == 0) dataCenter.nDF.set(dataCenter.indexDF.get(row).intValue(),(Integer)obj);
    if (col == 1) dataCenter.lDF.set(dataCenter.indexDF.get(row).intValue(),(Integer)obj);
    if (col == 2) dataCenter.DF.set(dataCenter.indexDF.get(row).intValue(),(Double)obj);
    if (col == 3) dataCenter.ErrorDF.set(dataCenter.indexDF.get(row).intValue(),(Double)obj);
    dataCenter.structInversionUpToDate = false;
  }
  
  /////////////////////////////////////////////////////////
  // add a blank row:
  /////////////////////////////////////////////////////////
  public void addBlankRow() {
    dataCenter.indexDF.add(dataCenter.nDF.size());
    dataCenter.nDF.add(null);
    dataCenter.lDF.add(null);
    dataCenter.DF.add(null);
    dataCenter.ErrorDF.add(null);
    dataCenter.structInversionUpToDate = false;
    dataCenter.DFloaded = true;
  }

  /////////////////////////////////////////////////////////
  /** remove a row: */
  /////////////////////////////////////////////////////////
  public void removeRow(int i) {
    if (i >= dataCenter.nDF.size()) return;
    int ii = dataCenter.indexDF.get(i).intValue();
    int jj;
    dataCenter.nDF.remove(ii);
    dataCenter.lDF.remove(ii);
    dataCenter.DF.remove(ii);
    dataCenter.ErrorDF.remove(ii);
    dataCenter.indexDF.remove(i);
    for(int j=0; j<dataCenter.indexDF.size(); j++) {
      jj = dataCenter.indexDF.get(j).intValue();
      if (jj > ii) dataCenter.indexDF.set(j,jj-1);
    }
    dataCenter.structInversionUpToDate = false;
  }

  /////////////////////////////////////////////////////////
  /** sort the table: */
  /////////////////////////////////////////////////////////
  public void sort() {
    int[] new_indexDF = dataCenter.sortDF();
    for(int i=0; i<new_indexDF.length; i++)
      dataCenter.indexDF.set(i,new_indexDF[i]);
  }

  /////////////////////////////////////////////////////////
  // Calculate theoretical frequency shifts
  /////////////////////////////////////////////////////////
  public void calculateDFfromTargets(JComponent c){
    if (!dataCenter.targetStructLoaded) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "No structural target functions loaded.  Unable to proceed.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    if (!dataCenter.kernelsLoaded) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "No kernels loaded.  Unable to proceed.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    // check to see if the span covered by the target grid is
    // large enough:
    if ((dataCenter.gridTargetStruct[0]>dataCenter.grid[0])||
        (dataCenter.gridTargetStruct[dataCenter.gridTargetStruct.length-1]<
         dataCenter.grid[dataCenter.grid.length-1])) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "Unable to calculate frequency splittings from\n"+
                    "target functions: the grid does not span a\n"+
                    "large enough domain.  Please enter a new target\n"+
                    "functions defined on a larger domain.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    // interpolate structural target functions onto the same grid as the kernels
    double[] targetStruct1 = Numerics.cubicSplineInterpol(
                             dataCenter.gridTargetStruct,
                             dataCenter.targetStruct1,
                             dataCenter.grid);
    double[] targetStruct2 = Numerics.cubicSplineInterpol(
                             dataCenter.gridTargetStruct,
                             dataCenter.targetStruct2,
                             dataCenter.grid);
    double[] rho = Numerics.cubicSplineInterpol(
                   dataCenter.gridModel,
                   dataCenter.rho,
                   dataCenter.grid);
    double[] weights = Numerics.findIntegrationWeights(dataCenter.grid);
    
    if (dataCenter.constantMass!=0) {
      JOptionPane.showMessageDialog(
        c,
        "Target function(s) may have to be adjusted to\n"+
        "ensure that the stellar mass remains constant",
        "Notification",
        JOptionPane.INFORMATION_MESSAGE
      );
    }
        
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==3)) {
      double num = 0, den = 0;
      for(int i=0; i<weights.length; i++) {
        num+=weights[i]*dataCenter.grid[i]*dataCenter.grid[i]*rho[i]*targetStruct1[i];
        den+=weights[i]*dataCenter.grid[i]*dataCenter.grid[i]*rho[i];
      }
      double A = num/den;
      for(int i=0; i<weights.length; i++) targetStruct1[i]-=A;
      for(int i=0; i<dataCenter.targetStruct1.length; i++)
        dataCenter.targetStruct1[i]-=A;
    }
            
    if ((dataCenter.constantMass==2)||(dataCenter.constantMass==3)) {
      double num = 0, den = 0;
      for(int i=0; i<weights.length; i++) {
        num+=weights[i]*dataCenter.grid[i]*dataCenter.grid[i]*rho[i]*targetStruct2[i];
        den+=weights[i]*dataCenter.grid[i]*dataCenter.grid[i]*rho[i];
      }
      double A = num/den;
      for(int i=0; i<weights.length; i++) targetStruct2[i]-=A;
      for(int i=0; i<dataCenter.targetStruct2.length; i++)
        dataCenter.targetStruct2[i]-=A;
    }

    clear();
    double sum;
    int ii = 0;
    for(int i=0; i<dataCenter.nK.length; i++) {
      sum = 0.0;
      for(int j=0; j<targetStruct1.length; j++) {
        sum+=weights[j]*targetStruct1[j]*dataCenter.struct1Kernel[i][j];
        sum+=weights[j]*targetStruct2[j]*dataCenter.struct2Kernel[i][j];
      }
      if (!Double.isNaN(sum)) {
        dataCenter.nDF.add(dataCenter.nK[i]);
        dataCenter.lDF.add(dataCenter.lK[i]);
        dataCenter.ErrorDF.add(1.0);
        dataCenter.DF.add(sum);
        dataCenter.indexDF.add(ii);
        ii++;
      }
    }
    if (ii > 0) dataCenter.DFloaded = true;
    dataCenter.structInversionUpToDate = false;
  }
}
