/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008
*
****************************************************************************/
package inversion_applet;

import graph.IPlot;
import inversion_applet.DataCenter;
import inversion_applet.Numerics;
import inversion_applet.Utilities;
import java.util.StringTokenizer;
import java.util.ArrayList;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.File;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JFrame;
import javax.swing.BoxLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.BorderLayout;
import java.awt.GridLayout;

public class KernelPage extends JPanel
{
  DataCenter dataCenter;
  int nfunc;
  int ngrid;

  private JLabel currentKernelLabel;
  private JLabel fValue;
  private JLabel cfValue;
  private JLabel nValue;
  private JLabel lValue;
  private JButton getStructKernelButton;
  private JButton readEFunctionsButton;

  private JButton firstKernel;
  private JButton previousKernel;
  private JButton nextKernel;
  private JButton lastKernel;
  private int currentKernel = 0;
  private JPanel buttonBox = new JPanel();

  private JPanel graphArea = new JPanel();
  private IPlot RotationKernelPlot;
  private IPlot StructKernelPlot;
  private IPlot StructKernelPlot2;

  /***************************************************************************
  *  This is the entry point if the program is run as a standalone application
  ***************************************************************************/
  public static void main(String args[]) {
    JFrame f = new JFrame("Kernel Page");
    DataCenter dataCenter = new DataCenter();
    KernelPage kernelPage = new KernelPage(dataCenter);

    kernelPage.init();
    f.add("Center",kernelPage);
    f.setSize(600,600);
    f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    f.setVisible(true);
  }


  /////////////////////////////////////////////////////////////////////
  /** A constructor */
  /////////////////////////////////////////////////////////////////////
  public KernelPage(DataCenter dataCenter) 
  {
    this.dataCenter = dataCenter;
  }

  /////////////////////////////////////////////////////////////////////
  /** This sets up the graphical user interface */
  /////////////////////////////////////////////////////////////////////
  public void init() {
    initReadEFunctionsButton();
    initNavigationButtons();

    RotationKernelPlot = new IPlot("Rotation kernel", "Radius, r/R", "Rotation kernel", 1);
    RotationKernelPlot.setIOenabled(!dataCenter.isApplet);
    StructKernelPlot = new IPlot("Structural kernel", "Radius, r/R", "Struct kernel", 1);
    StructKernelPlot.setIOenabled(!dataCenter.isApplet);
    StructKernelPlot2 = new IPlot("Structural kernel 2", "Radius, r/R", "Struct kernel 2", 1);
    StructKernelPlot2.setIOenabled(!dataCenter.isApplet);


    graphArea.setLayout(new GridLayout(3, 1));
    graphArea.add(RotationKernelPlot);
    graphArea.add(StructKernelPlot);
    graphArea.add(StructKernelPlot2);

    buttonBox.setLayout(new BoxLayout(buttonBox, BoxLayout.Y_AXIS));
    buttonBox.add(readEFunctionsButton);
  
    buttonBox.add(new JLabel(" "));
    buttonBox.add(currentKernelLabel);
    buttonBox.add(fValue);
    buttonBox.add(cfValue);
    buttonBox.add(nValue);
    buttonBox.add(lValue);
    buttonBox.add(new JLabel(" "));
    buttonBox.add(firstKernel);
    buttonBox.add(previousKernel);
    buttonBox.add(nextKernel);
    buttonBox.add(lastKernel);


    setLayout(new BorderLayout());
    add("Center", graphArea);
    add("East",buttonBox);
  }

  /////////////////////////////////////////////////////////////////////
  // init the navigation buttons:
  /////////////////////////////////////////////////////////////////////
  private void initNavigationButtons() {
    firstKernel    = Utilities.createButton("<<");
    previousKernel = Utilities.createButton("<");
    nextKernel     = Utilities.createButton(">");
    lastKernel     = Utilities.createButton(">>");
    currentKernelLabel = new JLabel(" No kernels loaded");
    fValue = new JLabel();
    cfValue = new JLabel();
    nValue = new JLabel();
    lValue = new JLabel();
    
    firstKernel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        currentKernel = 0;
        updateKernelDisplay();
      }
    });

    previousKernel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        currentKernel--;
        if (currentKernel < 0) currentKernel = 0;
        updateKernelDisplay();
      }
    });

    nextKernel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        currentKernel++;
        if (currentKernel >= dataCenter.rotaKernel.length) 
          currentKernel = dataCenter.rotaKernel.length-1;
        updateKernelDisplay();
      }
    });

    lastKernel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        currentKernel = dataCenter.rotaKernel.length-1;
        updateKernelDisplay();
      }
    });
  }

  /////////////////////////////////////////////////////////////////////
  // initialise readEFunctionsButton:
  /////////////////////////////////////////////////////////////////////
  private void initReadEFunctionsButton() {
    readEFunctionsButton = Utilities.createButton("Load eigenfunctions");
    readEFunctionsButton.setEnabled(!dataCenter.isApplet);
    readEFunctionsButton.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (!dataCenter.modelLoaded) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "No model loaded, please load the model first.",
            "No model",
            JOptionPane.ERROR_MESSAGE
          );
          return;
        }
        
        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showOpenDialog(readEFunctionsButton);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            dataCenter.currentDirectory = aFile.getParentFile();
            if (!readEigenFunctions(aFile.getPath())) return;
            
            InterpolateModel();
            buildKernel();
            buildKernel_Kc2rho();
            buildKernel_Krhoc2();
            updateKernelDisplay();
            
            // update some status variables in the dataCenter.
            dataCenter.kernelsLoaded = true;
            dataCenter.rotationInversionUpToDate = false;
            dataCenter.structInversionUpToDate = false;
            dataCenter.findStructKernels();
          } else {
            JOptionPane.showMessageDialog(
              readEFunctionsButton,
              "File \""+aFile.getName()+"\" doesn't exist.",
              "File not found",
              JOptionPane.ERROR_MESSAGE
            );
          }
        }
      }
    });
  }

  /////////////////////////////////////////////////////////////////////
  // This interpolates the model onto the grid given with the
  // eigenfunctions.
  /////////////////////////////////////////////////////////////////////
  private void InterpolateModel() {
    dataCenter.c2=Numerics.cubicSplineInterpol(dataCenter.gridModel, dataCenter.c2, dataCenter.grid);
    dataCenter.rho=Numerics.cubicSplineInterpol(dataCenter.gridModel, dataCenter.rho, dataCenter.grid);
    dataCenter.pressure=Numerics.cubicSplineInterpol(dataCenter.gridModel, dataCenter.pressure, dataCenter.grid);
    dataCenter.Gamma1=Numerics.cubicSplineInterpol(dataCenter.gridModel, dataCenter.Gamma1, dataCenter.grid);
    dataCenter.gridModel = dataCenter.grid.clone();
  }

  /////////////////////////////////////////////////////////////////////
  // update the kernel plots and labels:
  /////////////////////////////////////////////////////////////////////
  private void updateKernelDisplay() {
    updateKernelLabels();
    updateKernelPlot();
    updateStructKernelPlot();
    updateStructKernelPlot2();
  }

  /////////////////////////////////////////////////////////////////////
  // update the kernel labels:
  /////////////////////////////////////////////////////////////////////
  private void updateKernelLabels() {
    currentKernelLabel.setText(String.format(" Current kernel: %d/%d",
      currentKernel+1, dataCenter.rotaKernel.length));
    fValue.setText(String.format("    freq=%f",dataCenter.freq[currentKernel]));
    nValue.setText(String.format("    n=%d",dataCenter.nK[currentKernel]));
    lValue.setText(String.format("    \u2113=%d",dataCenter.lK[currentKernel]));
    cfValue.setText(String.format("    cf=%f",dataCenter.EnlCoefficient[currentKernel]));
  }
  
  /////////////////////////////////////////////////////////////////////
  // this updates the plot with one of the rotation kernels
  /////////////////////////////////////////////////////////////////////
  private void updateKernelPlot() {
    if (dataCenter.rotaKernel != null) {
      RotationKernelPlot.replaceDataSet(0,dataCenter.grid,dataCenter.rotaKernel[currentKernel]);
      RotationKernelPlot.findRange();
      RotationKernelPlot.repaint();
    }
  }

  /////////////////////////////////////////////////////////////////////
  // this updates the plot with one of the c2rho kernels
  /////////////////////////////////////////////////////////////////////
  private void updateStructKernelPlot()
  {
    if (dataCenter.c2rhoKernel != null) {
      StructKernelPlot.replaceDataSet(0, dataCenter.grid, dataCenter.c2rhoKernel[currentKernel]);
      StructKernelPlot.findRange();
      StructKernelPlot.repaint();
    }
  }

  /////////////////////////////////////////////////////////////////////
  // this updates the plot with one of the rhoc2 kernels
  /////////////////////////////////////////////////////////////////////
  private void updateStructKernelPlot2()
  {
    if (dataCenter.rhoc2Kernel != null) {
      StructKernelPlot2.replaceDataSet(0, dataCenter.grid, dataCenter.rhoc2Kernel[currentKernel]);
      StructKernelPlot2.findRange();
      StructKernelPlot2.repaint();
    }
  }

  /////////////////////////////////////////////////////////////////////////
  /** This method reads a file which contains the eigenfunctions.  The
  *   format of the file is an ASCII version of an output file from JCD's
  *   code plus an additional 10-line header. */
  /////////////////////////////////////////////////////////////////////////
  public boolean readEigenFunctions(String Filename) {
    double tol = 1.e-06;
    int i, j;
    String line;
    int nparams = 50;

    // set up some array lists into which to read different quantities:
    ArrayList<double[]> globals = new ArrayList<double[]>();
    ArrayList<double[]> hDisp = new ArrayList<double[]>();
    ArrayList<double[]> vDisp = new ArrayList<double[]>();

    try {
      BufferedReader eigInput = Utilities.OpenFile(Filename);

      // skip the header
      for (i = 0; i < 10; i++) line = eigInput.readLine();

      // get number of grid points
      line = eigInput.readLine();
      StringTokenizer t0 = new StringTokenizer(line," ");
      ngrid = Integer.parseInt(t0.nextToken());
      
      // set up the grid in dataCenter:
      dataCenter.grid = new double[ngrid];
      // FIRST read the grid....
      int count = 0;
      while (count < ngrid)
      {
        line = eigInput.readLine();
        t0 = new StringTokenizer(line," ");
        while (t0.hasMoreTokens())
        {
          dataCenter.grid[count] = Double.parseDouble(t0.nextToken());
          count++;
        }
      }

      // check the grid to see that it is possible to interpolate the
      // model onto the eigenfunction grid:
      if ((dataCenter.gridModel[0]>dataCenter.grid[0])||
          (dataCenter.gridModel[dataCenter.gridModel.length-1]<dataCenter.grid[ngrid-1])) {
        JOptionPane.showMessageDialog(
          readEFunctionsButton,
          "Model and eigenfunctions do not agree, please chose another set....",
          "Eigenfunction-model error",
          JOptionPane.ERROR_MESSAGE
        );
        return false;
      }

      // now read all available eigenfunctions
      nfunc = 0;
      line = eigInput.readLine();  // read a line of text
      while(true) {

        // test for the end of the buffer:
        if (line == null) break; // test for the end of the buffer

        // NOW read global params....
        globals.add(new double[nparams]);
        count = 0;
        while (count < nparams) {
          t0 = new StringTokenizer(line, " ");
          while (t0.hasMoreTokens()) {
            globals.get(nfunc)[count] = Double.parseDouble(t0.nextToken());
            count++;
          }
          // IMPORTANT: is line is null, then applying a string tokenizer will
          // provoke and error, so will be detected.  There is therefore no need
          // to do an extra test to see if line == null.
          line = eigInput.readLine();
        }
        
        // check the stellar mass and radius associated with the eigenfunctions
        // to see if they match that of the star:
        if ((Math.abs(dataCenter.StellarMass-globals.get(nfunc)[1]) > tol)||
            (Math.abs(dataCenter.StellarRadius-globals.get(nfunc)[2]) > tol)) {
          JOptionPane.showMessageDialog(
            readEFunctionsButton,
            "Model and eigenfunctions do not agree, please chose another set....",
            "Eigenfunction-model error",
            JOptionPane.ERROR_MESSAGE
          );
          return false;
        }

        // NOW read eigenfunctions....
        vDisp.add(new double[ngrid]);
        hDisp.add(new double[ngrid]);
        count = 0;
        while (count < 2 * ngrid)
        {
          t0 = new StringTokenizer(line, " ");
          while (t0.hasMoreTokens())
          {
            if (count % 2 == 0)
              vDisp.get(nfunc)[count/2] = Double.parseDouble(t0.nextToken());
            else 
              hDisp.get(nfunc)[count/2] = Double.parseDouble(t0.nextToken());
            count++;
          }
          line = eigInput.readLine();
        }
        nfunc++;
      }
      eigInput.close(); 
    } catch (Exception e) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "Format not recognised.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }

    // check to see that there is some data:
    if (nfunc == 0) {
      JOptionPane.showMessageDialog(
        readEFunctionsButton,
        "No eigenfunctions in file! No data loaded.",
        "Error reading file",
        JOptionPane.ERROR_MESSAGE
      );
      return false;
    }
    
    // find the surface index for the grid...
    double surface=1.0;
    int indSurface=0;
    double err=Math.abs(surface - dataCenter.grid[0]);
    for (i=1;i<ngrid;i++) {
      if (Math.abs(surface - dataCenter.grid[i])< err) {
        indSurface=i;
        err=Math.abs(surface - dataCenter.grid[i]);
      }
    }

    // set up some variables in the DataCenter:
    dataCenter.nK             = new int[nfunc];
    dataCenter.lK             = new int[nfunc];
    dataCenter.z1max          = new double[nfunc];
    dataCenter.freqDimless    = new double[nfunc];
    dataCenter.freq           = new double[nfunc];
    dataCenter.Enl            = new double[nfunc];
    dataCenter.EnlCoefficient = new double[nfunc];
    dataCenter.hDisp          = new double[nfunc][ngrid];
    dataCenter.vDisp          = new double[nfunc][ngrid];
    dataCenter.EigenFuncStellarRadius=globals.get(0)[2];
    dataCenter.EigenFuncStellarMass=globals.get(0)[1];

    for(int nn = 0; nn < nfunc; nn++) {
      dataCenter.hDisp[nn] = hDisp.get(nn);
      dataCenter.vDisp[nn] = vDisp.get(nn);
      dataCenter.nK[nn] = (int)globals.get(nn)[18];
      dataCenter.lK[nn] = (int)globals.get(nn)[17];
      dataCenter.z1max[nn] = globals.get(nn)[33];
      dataCenter.freqDimless[nn] = Math.sqrt(globals.get(nn)[19]);
      dataCenter.freq[nn] = Math.sqrt(globals.get(nn)[26])*1.e-03;
      dataCenter.Enl[nn] = globals.get(nn)[23];
      dataCenter.EnlCoefficient[nn]=(dataCenter.vDisp[nn][indSurface]*dataCenter.vDisp[nn][indSurface]+  
         dataCenter.hDisp[nn][indSurface]*dataCenter.hDisp[nn][indSurface])/
        (dataCenter.StellarRadius * dataCenter.RhoRef * dataCenter.rho[indSurface]);
    }
    
    // this is to avoid having error later on:
    if (currentKernel >= nfunc) currentKernel = nfunc-1;
    
    return true;
  }

  /* ************************************************************ */
  //  build rotation kernel
  //
  /* ************************************************************ */
  /* ************************************************************ */
  public void buildKernel() {
    int i, fnum;
    dataCenter.rotaKernel = new double[nfunc][ngrid];
    double pi=Math.PI;

    long degree;
    int nlines = ngrid;

    for (fnum = 0; fnum < nfunc; fnum++){
      degree = dataCenter.lK[fnum];
      long L2 = (long)degree * (degree + 1);
      dataCenter.rotaKernel[fnum][0] = 0;
      double cf=dataCenter.Enl[fnum]*dataCenter.StellarMass*dataCenter.EnlCoefficient[fnum]/(dataCenter.StellarRadius*dataCenter.StellarRadius);
      if (L2 != 0) {
        for (i = 1; i < nlines; i++) {
          //  if (dataCenter.grid[i] <= 1.)
          dataCenter.rotaKernel[fnum][i] = (dataCenter.vDisp[fnum][i] * dataCenter.vDisp[fnum][i] +
            dataCenter.hDisp[fnum][i] * dataCenter.hDisp[fnum][i] -
            2 * dataCenter.vDisp[fnum][i] * dataCenter.hDisp[fnum][i] / Math.sqrt(L2) - dataCenter.hDisp[fnum][i] * dataCenter.hDisp[fnum][i] / L2)
              / (cf * dataCenter.grid[i]); //
          //  else
          //    dataCenter.rotaKernel[fnum][i] = 0.;
        }
      }
      else
      {
        for (i = 1; i < nlines; i++)
        {
    /*      if (dataCenter.grid[i] <= 1.)
            dataCenter.rotaKernel[fnum][i] = (dataCenter.vDisp[fnum][i] * dataCenter.vDisp[fnum][i])
                / (cf * dataCenter.grid[i]); //
          else */
            dataCenter.rotaKernel[fnum][i] = 0.;
        }
      }
    }
  }


  /* ______________________________________________________ */


  public void buildKernel_Kc2rho() {
    int i, j, fnum;
    int nlines = ngrid;
    dataCenter.c2rhoKernel = new double[nfunc][ngrid];

    double[] ksir = new double[nlines];
    double[] dv;
    double[] xxx = new double[nlines];
    long L2;
    double L;
    double Integral;

    double[] eigV = new double[nlines];
    double[] eigH = new double[nlines];
    double a;
    double pi = Math.PI;

    for (j = 0; j < nfunc; j++)
    {

      L2 = (long)dataCenter.lK[j] * (dataCenter.lK[j] + 1);
      L = Math.sqrt((double)L2);
      // Firsr, let's write proper eigenfunctions....

      eigV[0] = 0;
      eigH[0] = L * eigV[0];

      if (L2 != 0)
        for (i = 1; i < nlines; i++)
        {
          a = dataCenter.z1max[j]* Math.sqrt(dataCenter.StellarMass / (4 * pi  * dataCenter.StellarRadius * dataCenter.RhoRef * dataCenter.rho[i])); //* dataCenter.grid[i]
          eigV[i] = dataCenter.vDisp[j][i] * a * Math.sqrt(dataCenter.grid[i]);
          eigH[i] = dataCenter.hDisp[j][i] * a * Math.sqrt(dataCenter.grid[i])/ L; 
        } 
      else
        for (i = 1; i < nlines; i++)
        {
          a = dataCenter.z1max[j]* Math.sqrt(dataCenter.StellarMass / (4 * pi  * dataCenter.StellarRadius * dataCenter.RhoRef * dataCenter.rho[i])); //* dataCenter.grid[i]
          eigV[i] = dataCenter.vDisp[j][i] * a * Math.sqrt(dataCenter.grid[i]);
          eigH[i] = 0;
        }

      dv = Numerics.differentiate(dataCenter.grid, eigV);


      for (i = 0; i < nlines; i++)
      {
        ksir[i] = 2 * eigV[i]  + dataCenter.grid[i] * dv[i]- L2 * eigH[i]; 
      }

      Integral = dataCenter.z1max[j] * dataCenter.z1max[j] * dataCenter.Enl[j]*dataCenter.StellarMass*dataCenter.StellarMass*dataCenter.EnlCoefficient[j]/(4*pi);

      for (i = 0; i < nlines; i++)
      {
        dataCenter.c2rhoKernel[j][i] = dataCenter.StellarRadius * dataCenter.Pref * dataCenter.rho[i] * dataCenter.c2[i] * ksir[i] * ksir[i] / 
          (2 * Integral * 4 * pi * pi * dataCenter.freq[j] * dataCenter.freq[j]);
      }
    }
  }

  /* ************************************************************ */
  //  build rotation kernel
  //
  //

  /* ************************************************************ */
  public void buildKernel_Krhoc2()
  {
    int i, j, fnum;
    int nlines = ngrid;
    dataCenter.rhoc2Kernel = new double[nfunc][ngrid];

    double[] m = new double[nlines];
    double[] ksir = new double[nlines];
    double[] dv;
    double[] drho = Numerics.differentiate(dataCenter.grid, dataCenter.rho);
    double[] Int1 = new double[nlines];
    double[] Int2 = new double[nlines];
    double[] Int3 = new double[nlines];
    double[] sl = new double[nlines];

    double Integral;
    double L2;
    double L;

    double G = dataCenter.G;
    int degree;

    double[] eigV = new double[nlines];
    double[] eigH = new double[nlines];
    double a;
    double pi = Math.PI;

    
     double surface=1.0;
      int indSurface=0;
      double err=Math.abs(surface - dataCenter.grid[0]);
      for (i=0;i<nlines;i++) if (Math.abs(surface - dataCenter.grid[i])< err) 
      {
        indSurface=i;
        err=Math.abs(surface - dataCenter.grid[i]);
      }

    for (fnum = 0; fnum < nfunc; fnum++)
    {
      degree = dataCenter.lK[fnum];
      L2 = (double)dataCenter.lK[fnum] * (dataCenter.lK[fnum] + 1);

      L = Math.sqrt((double)L2);
      // First, let's write proper eigenfunctions....


      eigV[0] = 0;
      eigH[0] = L * eigV[0];

      if (L2 != 0)
        for (i = 1; i < nlines; i++)
        {
          a = dataCenter.z1max[fnum]* Math.sqrt(dataCenter.StellarMass / (4 * pi  * dataCenter.StellarRadius * dataCenter.RhoRef * dataCenter.rho[i])); //* dataCenter.grid[i]
          eigV[i] = dataCenter.vDisp[fnum][i] * a * Math.sqrt(dataCenter.grid[i]);
          eigH[i] = dataCenter.hDisp[fnum][i] * a * Math.sqrt(dataCenter.grid[i])/ L; 
        } 
      else
        for (i = 1; i < nlines; i++)
        {
          a = dataCenter.z1max[fnum]* Math.sqrt(dataCenter.StellarMass / (4 * pi  * dataCenter.StellarRadius * dataCenter.RhoRef * dataCenter.rho[i])); //* dataCenter.grid[i]
          eigV[i] = dataCenter.vDisp[fnum][i] * a * Math.sqrt(dataCenter.grid[i]);
          eigH[i] = 0; 

        }


      // differentiate vertical displacement
      dv = Numerics.differentiate(dataCenter.grid, eigV);
    
      // define grid^degree array (needed fpr integration
      for (i = 0; i < nlines; i++)
      {
        sl[i] = 1;
        for (j = 0; j < dataCenter.lK[fnum]; j++) sl[i] = sl[i] * dataCenter.grid[i];
      }

      // calculate \xsi * r
      for (i = 0; i < nlines; i++)
      {
        ksir[i] = dataCenter.grid[i] * dv[i] + 2 * eigV[i] - L2 * eigH[i];
      }


      // calculate Int2 and m-array: 
      //  m(r)=\int_0^r (4\pi \rho r^2) dr

      m[0] = 0;
      m[1] = .5 * dataCenter.StellarRadius * dataCenter.StellarRadius * (dataCenter.grid[1] - dataCenter.grid[0]) * (4 * pi * dataCenter.grid[1] * dataCenter.grid[1] *
        dataCenter.RhoRef * dataCenter.rho[1]
        - 4 * pi * dataCenter.grid[0] * dataCenter.grid[0] * dataCenter.RhoRef * dataCenter.rho[0]);

      Int2[0] = 0;




// CONSTRUCT INTEGRATION MATRIX:
      double[] mat = Numerics.findIntegrationMatrix(dataCenter.grid);

// ANTI DERIVATIVE METHOD

      {
        double[] r2e, rint2e;
        r2e=new double[nlines];
        rint2e=new double[nlines];
        for (j = 0; j < nlines; j++)
        {
          r2e[j]        = 4 * pi * dataCenter.StellarRadius * dataCenter.StellarRadius * dataCenter.grid[j] * dataCenter.grid[j] * dataCenter.RhoRef * dataCenter.rho[j];
          rint2e[j] = dataCenter.StellarRadius * dataCenter.StellarRadius * //dataCenter.StellarRadius *
            (dataCenter.RhoRef * dataCenter.rho[j] * ksir[j] + eigV[j] * dataCenter.RhoRef * drho[j] * dataCenter.grid[j]) 
          * sl[j] * dataCenter.grid[j];                        
        }
        double[] am= Numerics.applyIntegrationMatrix(mat,r2e);
        double[] aInt= Numerics.applyIntegrationMatrix(mat,rint2e);

        for (i=1;i<nlines;i++)
        {
          m[i]=am[i]-am[0];
          Int2[i]= aInt[i]-aInt[0];
        }
      }




      //    set up Int1 and Int 3 arrays
      Int1[nlines - 1] = 0; Int3[nlines - 1] = 0;
      Int3[0] = 0;
      Int1[0] = 0;

// ANTI DERIVATIVE METHOD

      {
        double[] r2e, rint2e;
        r2e=new double[nlines];
        rint2e=new double[nlines];
        for (j = 1; j < nlines; j++)
        {
          r2e[j] = (2 * eigV[j] * ksir[j] / dataCenter.grid[j] + eigV[j] * eigV[j] * dataCenter.RhoRef * drho[j] / (dataCenter.RhoRef * dataCenter.rho[j]))
                      * dataCenter.RhoRef * dataCenter.rho[j];

          //Int3
          rint2e[j] = 
                 (dataCenter.RhoRef * dataCenter.rho[j] * ksir[j] 
                + eigV[j] * dataCenter.RhoRef * drho[j] * dataCenter.grid[j]) / sl[j];
        }

        j=0;
        if (degree == 0) rint2e[j] = 0.;
        else rint2e[j] = dataCenter.RhoRef * drho[j]; //dataCenter.StellarRadius * dataCenter.StellarRadius * 

        r2e[j] = (eigV[j] * eigV[j] * dataCenter.RhoRef * drho[j] / (dataCenter.RhoRef * dataCenter.rho[j]))
                          * dataCenter.RhoRef * dataCenter.rho[j];
        
        double[] am= Numerics.applyIntegrationMatrix(mat,r2e);
        double[] aInt= Numerics.applyIntegrationMatrix(mat,rint2e);

        for (i=0;i<nlines;i++)
        {
          Int1[i]=am[indSurface]-am[i]; //[nlines-1]
          Int3[i]=aInt[indSurface]-aInt[i]; //[nlines-1]
        }
      }

      // and now the kernel...
        

      Integral = dataCenter.z1max[fnum] * dataCenter.z1max[fnum] * dataCenter.Enl[fnum]*dataCenter.StellarMass*dataCenter.StellarMass*dataCenter.EnlCoefficient[fnum]/(4*pi); //

      for (i = 1; i < nlines; i++)
      {
        dataCenter.rhoc2Kernel[fnum][i] = 0;

        dataCenter.rhoc2Kernel[fnum][i] = dataCenter.Pref * dataCenter.rho[i] * dataCenter.c2[i] * ksir[i] * ksir[i];
        dataCenter.rhoc2Kernel[fnum][i] = dataCenter.rhoc2Kernel[fnum][i] - 4 * pi * pi * dataCenter.RhoRef * dataCenter.rho[i] * dataCenter.freq[fnum] * dataCenter.freq[fnum] *
          (eigV[i] * eigV[i] + L2 * eigH[i] * eigH[i]) * dataCenter.StellarRadius * dataCenter.StellarRadius * dataCenter.grid[i] * dataCenter.grid[i]; 
        dataCenter.rhoc2Kernel[fnum][i] = dataCenter.rhoc2Kernel[fnum][i] 
        + 2 * G * m[i] * dataCenter.RhoRef * dataCenter.rho[i] * eigV[i] * dv[i] 
        + G * (
        //      m[i] * dataCenter.RhoRef * drho[i] +
        4 * pi * dataCenter.StellarRadius * dataCenter.StellarRadius * dataCenter.grid[i] * dataCenter.grid[i] * dataCenter.RhoRef * dataCenter.rho[i]) *
            eigV[i] * eigV[i];

        dataCenter.rhoc2Kernel[fnum][i] = dataCenter.rhoc2Kernel[fnum][i] - 
                4 * pi * G * dataCenter.StellarRadius * dataCenter.StellarRadius * dataCenter.RhoRef * dataCenter.rho[i] * dataCenter.grid[i] * dataCenter.grid[i] * Int1[i];

        
        dataCenter.rhoc2Kernel[fnum][i] = dataCenter.rhoc2Kernel[fnum][i] 
                - 8 * pi * G * dataCenter.RhoRef * dataCenter.rho[i] / (2 * degree + 1) *
        ((degree + 1) / sl[i] * (eigV[i] - degree * eigH[i]) * Int2[i] 
          - degree * sl[i] * dataCenter.grid[i] * (eigV[i] + (degree + 1) * eigH[i]) * dataCenter.StellarRadius * dataCenter.StellarRadius * 
                Int3[i]*4 * pi * pi * dataCenter.freq[fnum] * dataCenter.freq[fnum]);
 
        dataCenter.rhoc2Kernel[fnum][i] = dataCenter.rhoc2Kernel[fnum][i] - G * m[i] * dataCenter.RhoRef * dataCenter.rho[i] *
       (2 * eigV[i] * ksir[i]/dataCenter.grid[i]);
        //       + eigV[i] * eigV[i] * dataCenter.RhoRef * drho[i] / (dataCenter.RhoRef * dataCenter.rho[i]));

        dataCenter.rhoc2Kernel[fnum][i] = dataCenter.StellarRadius * dataCenter.rhoc2Kernel[fnum][i]/  
                (2 * Integral * 4 * pi * pi * dataCenter.freq[fnum] * dataCenter.freq[fnum]); //* dataCenter.z1max[fnum] * dataCenter.z1max[fnum]);
      }
    }
  } 
}
