/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008
*
****************************************************************************/
package inversion_applet;

import org.netlib.blas.Dgemm;
import org.netlib.blas.Dgbmv;

public class Numerics {

	/* ************************************************************ */
	//	numerically differentiate 
	//					function f using grid r
	//
	/* ************************************************************ */
	public static double[] differentiate(double[] r, double[] f)
	{
		int i;
		int n = r.length;
		int m = f.length;
		double[] result;
		if (n != m)
		{
			result = new double[1];
			result[0] = -1;

		}
		else
		{
			result = new double[n];
			result[0] = (f[1] - f[0]) / (r[1] - r[0]);
			for (i = 1; i < n - 1; i++) result[i] = (f[i + 1] - f[i - 1]) / (r[i + 1] - r[i - 1]);
			result[n - 1] = (f[n - 1] - f[n - 2]) / (r[n - 1] - r[n - 2]);
		}
		return result;

	}

	/* ************************************************************ */
	//	numerically integrate
	//					function f using grid x
	//			(emulated int_tabulated function in IDL)
	/* ************************************************************ */
	public static double integrateNewtonCotes5(double[] x, double[] f)
	{
		int n = x.length;
		int i, n2;
		double dx, res;
		double[] xi, fi;

		n2 = 2 * n;
		while (n2 % 4 != 0) n2++;

		//System.out.println("New array length:"+n2);
		xi = new double[n2+1];
		dx = (x[n - 1] - x[0]) / (n2);

		for (i = 0; i <= n2; i++) xi[i] = x[0] + dx * i;
		fi = cubicSplineInterpol(x, f, xi);

		res = 0;
		for (i = 0; i < n2; i += 4)
		{
			res += 2 * dx * (7 * (fi[i] + fi[i + 4]) +
					32 * (fi[i + 1] + fi[i + 3]) + 12 * fi[i + 2]) / 45.;
		}
		return res;
	}


	/* ************************************************************ */
	// NATURAL CUBIC SPLINE INTERPOLATION
	//	used in numeric integration
	//					function f using grid x,
	//					newx - new grid for remapping
	//
	/* ************************************************************ */
	public static double[] cubicSplineInterpol(double[] x, double[] f, double newx[])
	// x - x-coordinate, (sorted)
	// f - function, (sorted)
	// n+1 - # elements in x & f
	// newx - new x-grid, with the same border points
	{

		int n2 = newx.length;
		int n = x.length - 1;

		double[] cf = new double[n];
		double[] h = new double[n];
		double[] res = new double[n2];
		int i, j;
		res[0] = f[0]; res[n2 - 1] = f[n];
		//		n = 10;
		//		double dx=1.;
		//		for (i = 0; i < n; i++) { x[i] = i * dx; f[i] = x[i] * x[i] * x[i]; }
		for (i = 0; i < n; i++) h[i] = x[i + 1] - x[i];
		cf = cubicSplineCf(x, f, h);

		j = 0;
		double di, di1;
		for (i = 1; i < n2 - 1; i++)
		{
			while (j < n + 1 & newx[i] > x[j]) j++;
			di1 = newx[i] - x[j - 1];
			di = x[j] - newx[i];
			res[i] = (cf[j] * di1 * di1 * di1 + cf[j - 1] * di * di * di) / 6 / h[j - 1];
			res[i] = res[i] + (f[j] / h[j - 1] - cf[j] * h[j - 1] / 6) * di1;
			res[i] = res[i] + (f[j - 1] / h[j - 1] - cf[j - 1] * h[j - 1] / 6) * di;
		}

		return res;
	}


	/* ************************************************************ */
	// NATURAL CUBIC SPLINE INTERPOLATION
	//	evaluated interpolation coefficients
	//					function f using grid x,
	//					newx - new grid for remapping
	//
	/* ************************************************************ */
	private static double[] cubicSplineCf(double[] x, double[] f, double h[])
	// x - x-coordinate, (sorted)
	// f - function, (sorted)
	// n+1 - # elements in x & f
	{
		int i;
		int n = x.length - 1;
		//System.out.println("n=" + n+" hlength="+h.length);
		double[] y = new double[n]; // 0 elements not used but kept for symmetry
		double[] z = new double[n+1];

		double[] a1 = new double[n+1];
		double[] a2 = new double[n+1];
		double[] r = new double[n+1];
		double[] y2 = new double[n+1];
		z[0] = z[n] = y2[0] = y2[n] = 0;



		for (i = 1; i < n; i++) y[i] = 6 * ((f[i + 1] - f[i]) / h[i] - (f[i] - f[i - 1]) / h[i - 1]);

		a1[1] = 2 * (h[0] + h[1]);
		y2[1] = y[1];
		a2[1] = h[1];
		r[1] = 1;
		for (i = 2; i < n; i++)
		{
			r[i] = h[i - 1] / a1[i - 1];
			a1[i] = 2 * (h[i - 1] + h[i]) - a2[i - 1] * r[i];
			a2[i] = h[i];
			y2[i] = y[i] - y2[i - 1] * r[i];
		}

		for (i = n - 1; i > 0; i--)
		{
			z[i] = (y2[i] - a2[i] * z[i + 1]) / a1[i];
		}




		return z;

	}

  /********************************************************
  * Initialise Chebyshev spectral transformation real2spec:
  ********************************************************/
  private static double[] initChebReal2Spec(int nr) {
    double[] real2spec = new double[nr*nr];
    double[] temp = new double[2*nr];
    
    for(int i=0; i<2*nr-2; i++) {
      temp[i] = Math.cos((Math.PI*i)/(nr-1));
    }

    double aux = 2.0/(double)(nr-1);
    for (int i=0; i<nr; i++){
      for (int j=0; j<nr; j++){
        real2spec[i+nr*j] = aux*temp[(i*j)%(2*nr-2)];
      }
    }
    
    for(int i=0; i<nr; i++) {
      real2spec[0+nr*i]   *=0.5;
      real2spec[nr-1+nr*i]*=0.5;
      real2spec[i+nr*0]   *=0.5;
      real2spec[i+nr*(nr-1)]*=0.5;
    }
    
    return real2spec;
  }

  /********************************************************
  * Initialise Chebyshev spectral transformation spec2real:
  ********************************************************/
  private static double[] initChebSpec2Real(int nr) {
    double[] spec2real = new double[nr*nr];
    double[] temp = new double[2*nr];
    
    for(int i=0; i<2*nr-2; i++) {
      temp[i] = Math.cos((Math.PI*i)/(nr-1));
    }

    double aux = 2.0/(double)(nr-1);
    for (int i=0; i<nr; i++){
      for (int j=0; j<nr; j++){
        spec2real[i+nr*j] = temp[(i*j)%(2*nr-2)];
      }
    }

    return spec2real;
  }

  /********************************************************
  * Initialise Chebyshev-based 1st order derivation matrix:
  ********************************************************/
  public static double[] initDerive(int nr) {
    double[] derive    = new double[nr*nr];
    double[] derive_spec = new double[nr*nr];
    double[] derive_temp = new double[nr*nr];
    double[] real2spec = initChebReal2Spec(nr);
    double[] spec2real = initChebSpec2Real(nr);
    
    for(int i=0; i<nr; i++) {
      for(int j=i+1; j<nr; j+=2) {
        derive_spec[i+nr*j] = -(double)(4*j);
      }
    }
    
    for(int i=0; i<nr; i++){derive_spec[0+nr*i]*=0.5;}

    // Use BLAS matrix multiplication
    // (add an offset of 0 to all array arguments)
    Dgemm.dgemm("N","N",nr,nr,nr,1.0,derive_spec,0,nr,real2spec,0,nr,
                0.0,derive_temp,0,nr);
    Dgemm.dgemm("N","N",nr,nr,nr,1.0,spec2real,0,nr,derive_temp,0,nr,
                0.0,derive,0,nr);
    return derive;
  }

  /********************************************************
  * Initialise Chebyshev-based 2nd order derivation matrix:
  ********************************************************/
  public static double[] initDDerive(int nr) {
    double[] derive    = initDerive(nr);
    double[] dderive   = new double[nr*nr];
    Dgemm.dgemm("N","N",nr,nr,nr,1.0,derive,0,nr,derive,0,nr,
                0.0,dderive,0,nr);
    return dderive;
  }

  /**************************************************************************
  *  This subroutine sets up a derivation matrix for functions expressed on
  *  the Gauss-Lobatto Chebyshev collocation grid.
  ***************************************************************************
  * Variables:
  *
  * rspan    = the span of the underlying grid (rspan = grid[nr-1]-grid[0])
  * nr       = the number of points in the grid (the grid is assumed to be the
               Gauss-Lobatto Chebyshev collocation grid)
  * der_max  = highest order derivative
  ***************************************************************************
  * The output: derive[0:der_max][nr*nr] (where nr = grid.length)
  *
  * derive = a two dimensional matrix which contains the derivation matrices
  *          in general (non-banded) form.  The first coordinate corresponds
  *          to the derivation order.  The second coordinate gives the
  *          position (i+nr*j) in the matrix  (where i=row, j=column).
  **************************************************************************/
  
  public static double[][] initDeriveCheb(double rspan, int nr, int der_max) {
  
    // check parameters:
    if (der_max < 1)
      throw new IllegalArgumentException("Please set der_max > 0 in initDeriveCheb");
    if (nr < 2)
      throw new IllegalArgumentException("Please set nr > 1 in initDeriveCheb");

    // set up derivation matrix:
    double[][] derive    = new double[der_max+1][nr*nr];
    derive[1] = initDerive(nr);
    for(int i=0; i< nr; i++) {
      derive[0][i+nr*i] = 1.0;
      for (int j=0; j<nr; j++) derive[1][i+nr*j]/=rspan;
    }
    for(int i=2; i<= der_max; i++)
      Dgemm.dgemm("N","N",nr,nr,nr,1.0,derive[1],0,nr,derive[i-1],0,nr,
                  0.0,derive[i],0,nr);
    return derive;
  }

  /********************************************************
  * Initialise integration weights for a Chebyshev 
  * Gauss-Lobatto collocation grid (i.e. using
  * spectral methods):
  ********************************************************/
  public static double[] initWeightsCheb(int n_grid) {
    double[] weights = new double [n_grid];
    double[] real2spec = initChebReal2Spec(n_grid);
    for(int i=0; i<n_grid; i++) {
      weights[i] = 0.0;
      for (int j=0; j<n_grid; j+=2) {
        weights[i]+=real2spec[j+n_grid*i]/(double)(1-j*j);
      }
    }
    return weights;
  }

  /**************************************************************************
  *  This subroutine sets up a banded derivation matrices (which is stored in
  *  banded form) using nth order finite difference schemes.  The
  *  derivatives are calculated with respect to an (almost) arbitrary grid.
  *  
  *  It calculates the derivation coefficients based on Lagrange interpolation
  *  polynomials (cf. Fornberg, 1988).
  ***************************************************************************
  * Variables:
  *
  * grid     = underlying on which the derivatives are calculated (grid(1:nr))
  *            Warning: this grid must not contain two identical points.
  * order    = order of the scheme, divided by 2.  Note:  (2.order >= der_max)
  * der_max  = highest order derivative
  ***************************************************************************
  * The output: derive[0:der_max][nr*nr] (where nr = grid.length)
  *
  * derive = a two dimensional matrix which contains the derivation matrices
  *          in general (non-banded) form.  The first coordinate corresponds
  *          to the derivation order.  The second coordinate gives the
  *          position (i+nr*j) in the matrix  (where i=row, j=column).
  **************************************************************************/
  public static double[][] initDeriveFDBand(double[] grid, int order, int der_max) {

    int start, finish;
    double prdct;

    // check parameters:
    if (order < der_max) 
      throw new IllegalArgumentException("Please set order >= der_max in initDeriveFD");
    if (order < 0) 
      throw new IllegalArgumentException("Please set order > 0 in initDeriveFD");
    if (der_max < 1)
      throw new IllegalArgumentException("Please set der_max > 0 in initDeriveFD");

    int nr = grid.length;
    
    // This is rather confusing but derive[der][i][j] = derive[der][i-j+order+(2*order+1)*j]
    // This is a banded matrix stored in banded form, which itself is stored as a vector,
    // according to its columns.
    double[][] derive = new double[der_max+1][nr*(2*order+1)];
    
    double[] a = new double[1+der_max];
    double[] mu = new double[2*order+1];

    for(int i=0; i<nr; i++) derive[0][order+(2*order+1)*i] = 1.0;

    for(int i=0; i<nr; i++) {

      start  = -order;
      finish = order;

      // special treatment do the endpoints
      // Note: this leads to results which are less precise on the
      // edges, but preserves the banded form of the derivation matrix.
      if ((i+start) < 0) start = -i;
      if ((i+finish) > (nr-1)) finish = nr - 1 - i;

      for(int j=start; j<=finish; j++) mu[j+order]=grid[j+i]-grid[i];
      
      for(int j=start; j<=finish; j++) {

        // Initialise Lagrange polynomial
        a[0] = 1.0;
        for(int l=1; l <= der_max; l++) a[l] = 0.0;
        
        prdct = 1.0;
        for(int k=start; k<j; k++)     prdct*=(mu[j+order]-mu[k+order]);
        for(int k=j+1; k<=finish; k++) prdct*=(mu[j+order]-mu[k+order]);
        a[0]/=prdct;

        // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
        for(int k=start; k<j; k++) {
          for(int l=der_max; l>0; l--) a[l] = -mu[k+order]*a[l] + a[l-1];
          a[0]*=(-mu[k+order]);
        }
        for(int k=j+1; k<=finish; k++) {
          for(int l=der_max; l>0; l--) a[l] = -mu[k+order]*a[l] + a[l-1];
          a[0]*=(-mu[k+order]);
        }
        
        // The coeffecients a(l) of the Lagrange polynomial are the
        // derivation coefficients divided by l!.
        prdct = 1.0;
        for(int l=1; l<= der_max; l++) {
          prdct*=l;

          // Reminder: derive[der][i][j] = derive[der][i-j+order+(2*order+1)*j]
          // Here we want derive[l][i][i+j]
          derive[l][order-j+(2*order+1)*(i+j)]=a[l]*prdct;
        }
      }
    }

    return derive;
  }

  /**************************************************************************
  *  This subroutine sets up a banded derivation matrices (which is stored in
  *  general form) using nth order finite difference schemes.  The
  *  derivatives are calculated with respect to an (almost) arbitrary grid.
  *  
  *  It calculates the derivation coefficients based on Lagrange interpolation
  *  polynomials (cf. Fornberg, 1988).
  ***************************************************************************
  * Variables:
  *
  * grid     = underlying on which the derivatives are calculated (grid(1:nr))
  *            Warning: this grid must not contain two identical points.
  * order    = order of the scheme, divided by 2.  Note:  (2.order >= der_max)
  * der_max  = highest order derivative
  ***************************************************************************
  * The output: derive[0:der_max][nr*nr] (where nr = grid.length)
  *
  * derive = a two dimensional matrix which contains the derivation matrices
  *          in general (non-banded) form.  The first coordinate corresponds
  *          to the derivation order.  The second coordinate gives the
  *          position (i+nr*j) in the matrix  (where i=row, j=column).
  **************************************************************************/
  public static double[][] initDeriveFD(double[] grid, int order, int der_max) {

    int start, finish;
    double prdct;

    // check parameters:
    if (order < der_max) 
      throw new IllegalArgumentException("Please set order >= der_max in initDeriveFD");
    if (order < 0) 
      throw new IllegalArgumentException("Please set order > 0 in initDeriveFD");
    if (der_max < 1)
      throw new IllegalArgumentException("Please set der_max > 0 in initDeriveFD");

    int nr = grid.length;
    double[][] derive = new double[der_max+1][nr*nr];
    
    double[] a = new double[1+der_max];
    double[] mu = new double[2*order+1];

    for(int i=0; i<nr; i++) derive[0][i+nr*i] = 1.0;

    for(int i=0; i<nr; i++) {

      start  = -order;
      finish = order;

      // special treatment do the endpoints
      // Note: this leads to results which are less precise on the
      // edges, but preserves the banded form of the derivation matrix.
      if ((i+start) < 0) start = -i;
      if ((i+finish) > (nr-1)) finish = nr - 1 - i;

      for(int j=start; j<=finish; j++) mu[j+order]=grid[j+i]-grid[i];
      
      for(int j=start; j<=finish; j++) {

        // Initialise Lagrange polynomial
        a[0] = 1.0;
        for(int l=1; l <= der_max; l++) a[l] = 0.0;
        
        prdct = 1.0;
        for(int k=start; k<j; k++)     prdct*=(mu[j+order]-mu[k+order]);
        for(int k=j+1; k<=finish; k++) prdct*=(mu[j+order]-mu[k+order]);
        a[0]/=prdct;

        // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
        for(int k=start; k<j; k++) {
          for(int l=der_max; l>0; l--) a[l] = -mu[k+order]*a[l] + a[l-1];
          a[0]*=(-mu[k+order]);
        }
        for(int k=j+1; k<=finish; k++) {
          for(int l=der_max; l>0; l--) a[l] = -mu[k+order]*a[l] + a[l-1];
          a[0]*=(-mu[k+order]);
        }
        
        // The coeffecients a(l) of the Lagrange polynomial are the
        // derivation coefficients divided by l!.
        prdct = 1.0;
        for(int l=1; l<= der_max; l++) {
          prdct*=l;
          derive[l][i+nr*(i+j)]=a[l]*prdct;
        }
      }
    }

    return derive;
  }

  /******************************************************************************** 
  *  This subroutine sets up an integration matrix in banded form for an arbitrary
  *  grid.  This matrix can then be applied to a function using the 
  *  applyIntegrationMatrix method in order to produce the anti-derivative of the
  *  function.  The conficients are based on the integrals of Lagrange
  *  interpolation polynomials calculated over a sliding window which spans
  *  [i-order, i+order+1].
  ******************************************************************************* 
  * description of variables:
  *
  * grid[]    = grid on which the integration is carried out
  * order     = positive int which gives half the size of the window
  *
  * Output:
  *
  * matrix[]  = matrix which enables one to calculate the anti-derivative of a
  *             function.
  ******************************************************************************/
  public static double[] findIntegrationMatrix(double[] grid) {
    return findIntegrationMatrix(grid, 1);
  }

  public static double[] findIntegrationMatrix(double[] grid, int order) {

    if (grid == null) return null;
    int ngrid = grid.length;
    if (ngrid == 0) return new double[0];
    // check parameters:
    if (order < 0)
      throw new IllegalArgumentException("Please set order >= 0");

    int ld = 2*order+2;
    double[] matrix = new double[ngrid*ld];
    
    double[] mu = new double[2*order+2];
    double[] a = new double[2*order+3];

    int start, finish;
    double  prdct, my_sum;

    for(int i=0; i < (ngrid-1); i++) {

      start  = -order;
      finish = order+1;

      // special treatment do the endpoints
      // Note: this leads to results which are less precise on the edges
      if ((i+start) < 0) start = -i;
      if ((i+finish) > (ngrid-1)) finish = ngrid - 1 - i;

      for(int j=start; j <= finish; j++) mu[j+order] = grid[j+i] - grid[i];

      for(int j=start; j <= finish; j++) {

        // Initialise Lagrange polynomial
        for(int l=1; l<= 2*(order+1); l++) a[l]=0.0;
        
        prdct = 1.0;
        for(int k=start; k < j; k++) prdct*=(mu[j+order]-mu[k+order]);
        for(int k=j+1; k <= finish; k++) prdct*=(mu[j+order]-mu[k+order]);
        a[0]=1.0/prdct;

        // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
        for(int k=start; k < j; k++) {
          for(int l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
          a[0] = -mu[k+order]*a[0];
        }
        for(int k=j+1; k <= finish; k++) {
          for(int l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
          a[0] = -mu[k+order]*a[0];
        }

        // Integrate the Lagrange polynomials, and add the results
        // as weights at the appropriate points.

        for(int l=0; l <= (2*order+2); l++) 
          matrix[1+order-j+ld*(i+j)]+=
            a[l]*Math.pow(mu[1+order],(double)(l+1))/(double)(l+1);
      }
    }
    
    return matrix;
  }

  /******************************************************************************** 
  *  This applies an integration matrix to a function to find the anti-derivative of
  *  the function.  See below for a description of the integration matrix.
  ******************************************************************************* 
  * description of variables:
  *
  * matrix[]  = integration matrix as supplied by findIntegrationMatrix
  * f[]       = function for which we want an anti-derivative
  * offset    = starting point for the anti-derivative
  * order     = positive int which gives half the size of the window
  *
  * Output:
  *
  * result[]  = contains the anti-derivative
  ******************************************************************************/
  public static double[] applyIntegrationMatrix(double[] matrix, double[] f) {
    return applyIntegrationMatrix(matrix,f,0.0,1);
  }

  public static double[] applyIntegrationMatrix(double[] matrix, double[] f, int order) {
    return applyIntegrationMatrix(matrix,f,0.0,order);
  }

  public static double[] applyIntegrationMatrix(double[] matrix, double[] f, double offset) {
    return applyIntegrationMatrix(matrix,f,offset,1);
  }

  public static double[] applyIntegrationMatrix(double[] matrix, double[] f, double offset, int order) {
    if ((f == null)||(matrix == null)) return null;
    int ngrid = f.length;
    if (ngrid == 0) return new double[0];

    // check parameters:
    if (order < 0)
      throw new IllegalArgumentException("Please set order >= 0");
    if (matrix.length < ((2*order+2)*ngrid))
      throw new IllegalArgumentException("The dimension of matrix is incorrect");

    double[] result = new double[ngrid];
    Dgbmv.dgbmv("N",ngrid,ngrid,order+1,order,1.0,matrix,0,2*order+2,f,0,1,0.0,result,0,1);
    result[0] = offset;
    for(int i=1; i<ngrid; i++) result[i]+=result[i-1];
    return result;
  }
  

/******************************************************************************** 
*  This subroutine sets up integration weights for an arbitrary grid.  It does 
*  this by integrating Lagrange interpolation polynomials calculated over a
*  sliding window which spans [i-order, i+order+1].
******************************************************************************* 
* description of variables:
*
* grid[]    = grid on which the integration is carried out
* order     = positive int which gives half the size of the window
*
* Output:
*
* weights[] = contains the integration weights.  This is actually an ngrid x ngrid
*             matrix stored as a vector, to be used with Dgemm.
******************************************************************************/
    public static double[] findIntegrationWeights(double[] grid) {
      return findIntegrationWeights(grid, 1);
    }

    public static double[] findIntegrationWeights(double[] grid, int order) {

      if (grid == null) return null;
      int ngrid = grid.length;
      if (ngrid == 0) return new double[0];
      double[] weights = new double[ngrid];
      
      // check parameters:
      if (order < 0)
        throw new IllegalArgumentException("Please set order >= order in Integral.findWeigths");

      double[] mu = new double[2*order+2];
      double[] a = new double[2*order+3];

      int start, finish;
      double  prdct, my_sum;

      for(int i=0; i < (ngrid-1); i++) {

        start  = -order;
        finish = order+1;

        // special treatment do the endpoints
        // Note: this leads to results which are less precise on the edges
        if ((i+start) < 0) start = -i;
        if ((i+finish) > (ngrid-1)) finish = ngrid - 1 - i;

        for(int j=start; j <= finish; j++) mu[j+order] = grid[j+i] - grid[i];

        for(int j=start; j <= finish; j++) {

          // Initialise Lagrange polynomial
          for(int l=1; l<= 2*(order+1); l++) a[l]=0.0;
          
          prdct = 1.0;
          for(int k=start; k < j; k++) prdct*=(mu[j+order]-mu[k+order]);
          for(int k=j+1; k <= finish; k++) prdct*=(mu[j+order]-mu[k+order]);
          a[0]=1.0/prdct;

          // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
          for(int k=start; k < j; k++) {
            for(int l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }
          for(int k=j+1; k <= finish; k++) {
            for(int l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }

          // Integrate the Lagrange polynomials, and add the results
          // as weights at the appropriate points.

          for(int l=0; l <= (2*order+2); l++) 
            weights[i+j]+=a[l]*Math.pow(mu[1+order],(double)(l+1))/(double)(l+1);
        }
      }
      
      return weights;
    }

/******************************************************************************** 
*  This subroutine finds an antiderivatie of a function defined over an
*  arbitrary grid.  It does this by integrating Lagrange interpolation
*  polynomials calculated over a sliding window which spans [i-order,
*  i+order+1].
*********************************************************************************
* description of variables:
*
* grid[]    = grid on which the integration is carried out
* order     = positive int which gives half the size of the window
*
* Output:
*
* weights[] = contains the integration weights.  This is actually an ngrid x ngrid
*             matrix stored as a vector, to be used with Dgemm.
******************************************************************************/
    public static double[] findAntiDerivative(double[] grid, double[] f) {
      return findAntiDerivative(grid, f, 0.0, 1);
    }

    public static double[] findAntiDerivative(double[] grid, double[] f, double offset) {
      return findAntiDerivative(grid, f, offset, 1);
    }

    public static double[] findAntiDerivative(double[] grid, double[] f, int order) {
      return findAntiDerivative(grid, f, 0.0, order);
    }

    public static double[] findAntiDerivative(double[] grid, double[] f, double offset, int order) {

      if (grid == null) return null;
      if (f == null) return null;
      int ngrid = grid.length;
      int nf = f.length;
      ngrid = (ngrid>nf?nf:ngrid);
      if (ngrid == 0) return new double[0];
      double[] antiDerivative = new double[ngrid];
      
      // check parameters:
      if (order < 0)
        throw new IllegalArgumentException("Please set order >= order in antiDerivative.findWeigths");

      double[] mu = new double[2*order+2];
      double[] a = new double[2*order+3];

      int start, finish;
      double  prdct, my_sum;
      antiDerivative[0] = offset;

      for(int i=0; i < (ngrid-1); i++) {
        antiDerivative[i+1] = antiDerivative[i];

        start  = -order;
        finish = order+1;

        // special treatment do the endpoints
        // Note: this leads to results which are less precise on the edges
        if ((i+start) < 0) start = -i;
        if ((i+finish) > (ngrid-1)) finish = ngrid - 1 - i;

        for(int j=start; j <= finish; j++) mu[j+order] = grid[j+i] - grid[i];

        for(int j=start; j <= finish; j++) {

          // Initialise Lagrange polynomial
          for(int l=1; l<= 2*(order+1); l++) a[l]=0.0;
          
          prdct = 1.0;
          for(int k=start; k < j; k++) prdct*=(mu[j+order]-mu[k+order]);
          for(int k=j+1; k <= finish; k++) prdct*=(mu[j+order]-mu[k+order]);
          a[0]=1.0/prdct;

          // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
          for(int k=start; k < j; k++) {
            for(int l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }
          for(int k=j+1; k <= finish; k++) {
            for(int l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }

          // Integrate the Lagrange polynomials, and add the results
          // as weights at the appropriate points.

          for(int l=0; l <= (2*order+2); l++)
            antiDerivative[i+1]+=f[i+j]*a[l]*Math.pow(mu[1+order],(double)(l+1))/(double)(l+1);
        }
      }
      
      return antiDerivative;
    }

/******************************************************************************** 
*  This subroutine finds the integral of a function defined over an
*  arbitrary grid.  It does this by integrating Lagrange interpolation
*  polynomials calculated over a sliding window which spans [i-order,
*  i+order+1].
*********************************************************************************
* description of variables:
*
* grid[]    = grid on which the integration is carried out
* order     = positive int which gives half the size of the window
*
* Output:
*
* weights[] = contains the integration weights.  This is actually an ngrid x ngrid
*             matrix stored as a vector, to be used with Dgemm.
******************************************************************************/
    public static double findIntegral(double[] grid, double[] f) {
      return findIntegral(grid, f, 1);
    }

    public static double findIntegral(double[] grid, double[] f, int order) {

      if (grid == null) return 0.0;
      if (f == null) return 0.0;
      int ngrid = grid.length;
      int nf = f.length;
      ngrid = (ngrid>nf?nf:ngrid);
      if (ngrid == 0) return 0.0;
      
      // check parameters:
      if (order < 0)
        throw new IllegalArgumentException("Please set order >= order in Integral.findWeigths");

      double[] mu = new double[2*order+2];
      double[] a = new double[2*order+3];

      int start, finish;
      double  prdct, my_sum;
      double integral = 0.0;

      for(int i=0; i < (ngrid-1); i++) {

        start  = -order;
        finish = order+1;

        // special treatment do the endpoints
        // Note: this leads to results which are less precise on the edges
        if ((i+start) < 0) start = -i;
        if ((i+finish) > (ngrid-1)) finish = ngrid - 1 - i;

        for(int j=start; j <= finish; j++) mu[j+order] = grid[j+i] - grid[i];

        for(int j=start; j <= finish; j++) {

          // Initialise Lagrange polynomial
          for(int l=1; l<= 2*(order+1); l++) a[l]=0.0;
          
          prdct = 1.0;
          for(int k=start; k < j; k++) prdct*=(mu[j+order]-mu[k+order]);
          for(int k=j+1; k <= finish; k++) prdct*=(mu[j+order]-mu[k+order]);
          a[0]=1.0/prdct;

          // Calculate Lagrange polynomial, by calculating product of (x-mu(k))
          for(int k=start; k < j; k++) {
            for(int l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }
          for(int k=j+1; k <= finish; k++) {
            for(int l=2*order+2; l > 0; l--) a[l] = -mu[k+order]*a[l]+a[l-1];
            a[0] = -mu[k+order]*a[0];
          }

          // Integrate the Lagrange polynomials, and add the results
          // as weights at the appropriate points.

          for(int l=0; l <= (2*order+2); l++)
            integral+=f[i+j]*a[l]*Math.pow(mu[1+order],(double)(l+1))/(double)(l+1);
        }
      }
      
      return integral;
    }

  ////////////////////////////////////////////////////////////////////////////
  /** This method gives the value of the Legendre polynomials
  *   of degrees 0 to n-1, at the point x.  This is done using
  *   an efficient recursion formula. */
  ////////////////////////////////////////////////////////////////////////////
  public static double[] Legendre(double x, int n) {
    if (n < 1) return null;
    double[] result = new double[n];
    result[0] = 1.0;
    if (n > 1) result[1] = x;
    for(int i=2; i<n; i++)
      result[i] = ((2.0*i-1.0)*x*result[i-1]-(i-1.0)*result[i-2])/(double)i;
    return result;
  }

}
