/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008
*
****************************************************************************/
package graph;

import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.awt.Color;
import java.awt.BasicStroke;
import java.awt.Stroke;
import java.awt.Graphics2D;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.Font;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.awt.geom.AffineTransform;
import java.awt.font.LineMetrics;
import java.awt.font.FontRenderContext;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import java.lang.Cloneable;

/////////////////////////////////////////////////////////////////////////////
/**
*
* The purpose of this class is to draw plots in Java.
*
*/
/////////////////////////////////////////////////////////////////////////////

public class Plot extends JComponent implements Cloneable
{
  // Arrays containing data sets to be plotted:
  // (x[0][*],y[0][*]) is the first data set to be plotted.
  private ArrayList<double[]> x = new ArrayList<double[]>();
  private ArrayList<double[]> y = new ArrayList<double[]>();
  
  // Integer containing the number of data sets:
  private int nDataSets = 0;

  // Array which says type of stroke to use for plotting
  // the different data sets. See section "Various line patterns"
  // below for acceptable values and their meanings.
  private ArrayList<Integer> linestyleArray = new ArrayList<Integer>();

  // Array which says what symbols to use for plotting
  // the different data sets. See section "Various symbols"
  // below for acceptable values and their meanings.
  private ArrayList<Integer> psymArray = new ArrayList<Integer>();

  // Array which contains the different colors used for the different
  // data sets:
  private ArrayList<Color> colorArray = new ArrayList<Color>();
  
  // Margins around the plot area.  These
  // need to be big enough to allow tick
  // marks, annotations and various titles.
  private float leftMinMargin= 50F;
  private float leftMargin   = 50F;
  private float rightMargin  = 30F;
  private float topMargin    = 50F;
  private float bottomMargin = 50F;
  
  // These submargins are for the x and y axis titles:
  private float leftSubMargin   = 30F;
  private float bottomSubMargin = 30F;

  // These are the lengths of the tick marks
  private float xLongTick  = 8F;
  private float xShortTick = 4F;
  private float yLongTick  = 8F;
  private float yShortTick = 4F;
  
  // These are variables in which to store information
  // about the ticks and their labels;
  private int nticks;
  private double[] ticks;
  private boolean[] isMajor;
  private int nDecimal;
  private boolean withExp;

  // The x and y bounds.
  private double xmin=0, xmax=1, ymin=0, ymax=1;

  // Various titles:
  private String xtitle = "x axis";
  private String ytitle = "y axis";
  protected String title = "A plot";

  // Various fonts:
  private Font annotationFont = new Font("Sans Serif", Font.PLAIN, 12);
  private Font exponentFont = annotationFont.deriveFont(9F);
  private Font xyTitleFont = annotationFont.deriveFont(18F);
  private Font titleFont   = annotationFont.deriveFont(Font.BOLD,22F);
  private Font timesFont   = annotationFont.deriveFont(10F);

  // These strings are used when plotting numbers using scientific notation
  private String timesString="x";
  private String tenString="10";

  // These keep track of the size of the plot area:
  private float hspan, vspan;

  // These are the coordinates of the four corners
  // of the plot area.  They are calculated using
  // the margins given above and the bounds of the
  // component.  By declaring xleft as a method, it
  // automatically adjusts to any changes in the left
  // margin.
  private float xleft() {return leftMargin;}
  private float xright() {return hspan-rightMargin;}
  private float yup() {return topMargin;}
  private float ydown() {return vspan-bottomMargin;}

  /////////////////////////////////////////////////////////////////////////////
  // Various line patterns:
  // The conventions used here are similar to IDL.
  public static final int noLine = -1;
  public static final int solid = 0;      // the default
  public static final int dotted = 1;
  public static final int dashed = 2;
  public static final int dotDash = 3;
  public static final int dotDotDash = 4;
  public static final int longDash = 5;
  
  private static final Stroke[] linestyle = new Stroke[]
    {
      // Stroke pattern corresponding to a solid line
      new BasicStroke(2F),
      // Stroke pattern corresponding to a dotted line 
      new BasicStroke(2F,                  // Width
          BasicStroke.CAP_BUTT,            // End cap
          BasicStroke.JOIN_MITER,          // Join style
          3.0F,                            // Miter limit
          new float[] {2F,2F},             // Dash pattern
          0.0F),                           // Dash phase
      // Stroke pattern corresponding to a dashed line 
      new BasicStroke(2F,                  // Width
          BasicStroke.CAP_BUTT,            // End cap
          BasicStroke.JOIN_MITER,          // Join style
          3.0F,                            // Miter limit
          new float[] {6F,6F},             // Dash pattern
          0.0F),                           // Dash phase
      // Stroke pattern corresponding to a dot-dash line
      new BasicStroke(2F,                  // Width
          BasicStroke.CAP_BUTT,            // End cap
          BasicStroke.JOIN_MITER,          // Join style
          3.0F,                            // Miter limit
          new float[] {6F,6F,2F,6F},       // Dash pattern
          0.0F),                           // Dash phase
      // Stroke pattern corresponding to a dot-dot-dash line
      new BasicStroke(2F,                  // Width
          BasicStroke.CAP_BUTT,            // End cap
          BasicStroke.JOIN_MITER,          // Join style
          3.0F,                            // Miter limit
          new float[] {6F,6F,2F,2F,2F,6F}, // Dash pattern
          0.0F),                           // Dash phase
      // Stroke pattern corresponding to a line with long dashes
      new BasicStroke(2F,                  // Width
          BasicStroke.CAP_BUTT,            // End cap
          BasicStroke.JOIN_MITER,          // Join style
          3.0F,                            // Miter limit
          new float[] {12F,12F},           // Dash pattern
          0.0F)                            // Dash phase
    };

  /////////////////////////////////////////////////////////////////////////////
  // Various symbols:
  // The conventions used here are similar to IDL.
  public static final int noSymbol = 0;      // the default
  public static final int plus = 1;
  public static final int asterisk = 2;
  public static final int singleDot = 3;
  public static final int diamond = 4;
  public static final int triangle = 5;
  public static final int square = 6;
  public static final int times = 7;
  
  // This part codes the different segments as a series
  // of line segments (x0,y0,x1,y1,x2,y2 ...) where:
  // (x0,y0)-(x1,y1) = first segment
  // (x2,y2)-(x3,y3) = second segment
  // ...
  // Note the y axis is inverted (this affects the triangle).
  private static final float[][] psym = new float[][]
  {
    null,
    new float[] {0F,5F,0F,-5F,-5F,0F,5F,0F},
    new float[] {0F,5F,0F,-5F,-5F,0F,5F,0F,-5F,-5F,5F,5F,-5F,5F,5F,-5F},
    new float[] {-1F,1F,1F,1F,1F,1F,1F,-1F,1F,-1F,-1F,-1F,-1F,-1F,-1F,1F},
    new float[] {-5F,0F,0F,5F,0F,5F,5F,0F,5F,0F,0F,-5F,0F,-5F,-5F,0F},
    new float[] {0F,-6F,5.2F,3F,5.2F,3F,-5.2F,3F,-5.2F,3F,0F,-6F},    
    new float[] {-5F,5F,5F,5F,5F,5F,5F,-5F,5F,-5F,-5F,-5F,-5F,-5F,-5F,5F},
    new float[] {-5F,-5F,5F,5F,-5F,5F,5F,-5F}
  };

  // The stroke pattern used for creating the symbols:
  private static final Stroke psymStroke = new BasicStroke(1F);

  /////////////////////////////////////////////////////////////////////////////
  /**
   * A constructor.  If the arrays do not have the same lengths,
   * the longer one is truncated.
   *
   * x      = array of x values
   * y      = array of y values
   * ls     = integer giving the line style
   * ps     = integer giving the choice of symbols
   * aColor = a color
   */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(double[] x, double[] y, int ls, int ps, Color aColor) {

    // Check input parameters:
    if ((x == null) || (y == null)) return;

    int xsize = x.length;
    int ysize = y.length;

    if ((ls < -1)||(ls >5)) ls = 0;

    if ((ps < 0)||(ps > 7)) ps = 0;

    // input (x,y) arrays, linestyle ls and symbol style ps
    int length = (xsize>ysize)?ysize:xsize;
    
    nDataSets = 1;
    this.x.add(new double[length]);
    this.y.add(new double[length]);
    this.linestyleArray.add(ls);
    this.psymArray.add(ps);
    this.colorArray.add(aColor);

    System.arraycopy(x,0,this.x.get(this.x.size()-1),0,length);
    System.arraycopy(y,0,this.y.get(this.y.size()-1),0,length);
    
  }

  /////////////////////////////////////////////////////////////////////////////
  /** Another constructor.  Input x[], y[], linestyle and choice of symbols. */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(double[] x, double[] y, int ls, int ps) {this(x,y,ls,ps,Color.black);}

  /////////////////////////////////////////////////////////////////////////////
  /** Another constructor.  Input x[], y[], linestyle and choice of color. */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(double[] x, double[] y, int ls, Color aColor) {this(x,y,ls,0,aColor);}

  /////////////////////////////////////////////////////////////////////////////
  /** Another constructor.  Input x[], y[] and linestyle.  */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(double[] x, double[] y, int ls) {this(x,y,ls,0,Color.black);}

  /////////////////////////////////////////////////////////////////////////////
  /** Another constructor.  Input x[], y[].  */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(double[] x, double[] y) {this(x,y,0,0,Color.black);}
  
  /////////////////////////////////////////////////////////////////////////////
  /** Another constructor. Input title, xtitle, ytitle and number of data sets.*/
  /////////////////////////////////////////////////////////////////////////////
  public Plot(String title, String xtitle, String ytitle, int nDataSets) {
    setTitles(title,xtitle,ytitle);

    if (nDataSets < 0) {
      this.nDataSets = 0;
      return;
    }

    this.nDataSets = nDataSets;
    for(int i=0; i<nDataSets; i++) {
      x.add(null);
      y.add(null);
      linestyleArray.add(0);
      psymArray.add(0);
      colorArray.add(Color.black);  // It is important to put a true color
                                    // rather than null because of the way
                                    // replaceDataSet is implemented.  The
                                    // same applies to linestyleArray and
                                    // psymArray.
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** Another constructor.  Input title, xtitle and ytitle.  */
  /////////////////////////////////////////////////////////////////////////////
  public Plot(String title, String xtitle, String ytitle) {
    setTitles(title,xtitle,ytitle);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the titles to be modified */
  /////////////////////////////////////////////////////////////////////////////
  public void setTitles(String title, String xtitle, String ytitle) {
    this.title = title;
    this.xtitle = xtitle;
    this.ytitle = ytitle;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle and choice of symboles. */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls, int ps, Color aColor) {

    // Check parameters:
    if ((x == null) || (y == null)) return;

    int xsize = x.length;
    int ysize = y.length;

    if ((ls < -1)||(ls >5)) ls = 0;

    if ((ps < 0)||(ps > 7)) ps = 0;

    // Add data set:
    int length = (xsize>ysize)?ysize:xsize;
    this.x.add(new double[length]);
    this.y.add(new double[length]);
    System.arraycopy(x,0,this.x.get(this.x.size()-1),0,length);
    System.arraycopy(y,0,this.y.get(this.y.size()-1),0,length);
    linestyleArray.add(ls);
    psymArray.add(ps);
    colorArray.add(aColor);
    nDataSets++;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle and choice of 
  *   symbols. */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls, int ps) {
    addDataSet(x,y,ls,ps,Color.black);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle and choice of 
  *   color. */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls, Color aColor) {
    addDataSet(x,y,ls,0,aColor);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y) with a specified linestyle. */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y, int ls) {
    addDataSet(x,y,ls,0,Color.black);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /** This adds a data set (x,y). */
  /////////////////////////////////////////////////////////////////////////////
  public void addDataSet(double[] x, double[] y) {
    addDataSet(x,y,0,0,Color.black);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This removes data set number ds from arrays x and y, where:
   *    0 <= ds < nDataSets
   */
  /////////////////////////////////////////////////////////////////////////////
  public void removeDataSet(int ds) {
    if ((ds >= nDataSets)||(ds < 0)) return;
   
    x.remove(ds);
    y.remove(ds);
    linestyleArray.remove(ds);
    psymArray.remove(ds);
    colorArray.remove(ds);
    nDataSets--;
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y, linestyle ls
   * and symbol choice ps.
   *
   * ds     = data set that will be replaced (0 <= ds < nDataSets)
   * x      = new array of x values
   * y      = new array of y values
   * ls     = new linestyle
   * ps     = new choice of symbols
   * aColor = new choice of color
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls, int ps,
                             Color aColor) {

    // check parameters:
    if ((ds >= nDataSets)||(ds < 0)) return;
    
    if ((x == null) || (y == null)) return;

    int xsize = x.length;
    int ysize = y.length;

    if ((ls < -1)||(ls >5)) ls = linestyleArray.get(ds);

    if ((ps < 0)||(ps > 7)) ps = psymArray.get(ds);

    // replace data set:
    int length = (xsize>ysize)?ysize:xsize;
    this.x.set(ds,new double[length]);
    this.y.set(ds,new double[length]);
    System.arraycopy(x,0,this.x.get(ds),0,length);
    System.arraycopy(y,0,this.y.get(ds),0,length);
    linestyleArray.set(ds,ls);
    psymArray.set(ds,ps);
    colorArray.set(ds,aColor);
  }
  
  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y, linestyle ls
   * and symbol choice ps.
   *
   * ds  = data set that will be replaced (0 <= ds < nDataSets)
   * x   = new array of x values
   * y   = new array of y values
   * ls  = new linestyle
   * ps  = new choice of symbols
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls, int ps) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    replaceDataSet(ds,x,y,ls,ps,colorArray.get(ds));
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y, linestyle ls
   * and symbol choice ps.
   *
   * ds     = data set that will be replaced (0 <= ds < nDataSets)
   * x      = new array of x values
   * y      = new array of y values
   * ls     = new linestyle
   * aColor = new choice of symbols
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls, Color aColor) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    replaceDataSet(ds,x,y,ls,psymArray.get(ds),aColor);
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y, linestyle ls
   * and symbol choice ps.
   *
   * ds  = data set that will be replaced (0 <= ds < nDataSets)
   * x   = new array of x values
   * y   = new array of y values
   * ls  = new linestyle
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y, int ls) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    replaceDataSet(ds,x,y,ls,psymArray.get(ds),colorArray.get(ds));
  }

  /////////////////////////////////////////////////////////////////////////////
  /**
   * This replaces data set number ds using arrays x and y, linestyle ls
   * and symbol choice ps.
   *
   * ds  = data set that will be replaced (0 <= ds < nDataSets)
   * x   = new array of x values
   * y   = new array of y values
   */
  /////////////////////////////////////////////////////////////////////////////
  public void replaceDataSet(int ds, double[] x, double[] y) {
    if ((ds >= nDataSets)||(ds < 0)) return;
    replaceDataSet(ds,x,y,linestyleArray.get(ds),psymArray.get(ds),colorArray.get(ds));
  }
 
  /////////////////////////////////////////////////////////////////////////////
  /** This finds appropriate values for xmin, xmax, ymin and ymax. */
  /////////////////////////////////////////////////////////////////////////////
  public void findRange() {
    int i;
    if (nDataSets == 0) return;
    
    for(i=0; i<nDataSets; i++) {
      if ((x.get(i) != null)&&(x.get(i).length > 0)) break;
    }

    if (i == nDataSets) return;

    xmin = x.get(i)[0];
    xmax = x.get(i)[0];
    ymin = y.get(i)[0];
    ymax = y.get(i)[0];
    
    for(i = 0; i < nDataSets; i++) {
      if (x.get(i) != null) {
        for(int j = 0; j < x.get(i).length; j++) {
          if (xmin > x.get(i)[j]) xmin = x.get(i)[j];
          if (xmax < x.get(i)[j]) xmax = x.get(i)[j];
          if (ymin > y.get(i)[j]) ymin = y.get(i)[j];
          if (ymax < y.get(i)[j]) ymax = y.get(i)[j];
        }
      }
    }
   
    // This adds a margin above and below the plotted functions.
    double dy = ymax - ymin;
    ymax+=(dy*0.1);
    ymin-=(dy*0.1);

    // if ymin = ymax, then add space below and above:
    if (ymin==ymax) {ymin-=0.5; ymax+=0.5;}

    // if xmin = xmax, then add space to the left and right:
    if (xmin==xmax) {xmin-=0.5; xmax+=0.5;}
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This finds appropriate values for xmin, xmax, ymin and ymax for the
  *   data sets selected by the array "selection". */
  /////////////////////////////////////////////////////////////////////////////
  public void findRange(int[] selection) {
    int ii;
    int i = 0;
    int ns = selection.length;
    if (nDataSets == 0) return;
    
    for(ii=0; ii<ns; ii++) {
      i = selection[ii];
      if ((i >= 0)&&(i < nDataSets)) 
        if ((x.get(i) != null)&&(x.get(i).length > 0)) break;
    }
    if (ii == ns) return;

    xmin = x.get(i)[0];
    xmax = x.get(i)[0];
    ymin = y.get(i)[0];
    ymax = y.get(i)[0];
    
    for(ii = 0; ii < ns; ii++) {
      i = selection[ii];
      if ((i >= 0)&&(i < nDataSets)) {
        if (x.get(i) != null) {
          for(int j = 0; j < x.get(i).length; j++) {
            if (xmin > x.get(i)[j]) xmin = x.get(i)[j];
            if (xmax < x.get(i)[j]) xmax = x.get(i)[j];
            if (ymin > y.get(i)[j]) ymin = y.get(i)[j];
            if (ymax < y.get(i)[j]) ymax = y.get(i)[j];
          }
        }
      }
    }
   
    // This adds a margin above and below the plotted functions.
    double dy = ymax - ymin;
    ymax+=(dy*0.1);
    ymin-=(dy*0.1);

    // if ymin = ymax, then add space below and above:
    if (ymin==ymax) {ymin-=0.5; ymax+=0.5;}

    // if xmin = xmax, then add space to the left and right:
    if (xmin==xmax) {xmin-=0.5; xmax+=0.5;}
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set xmin */
  /////////////////////////////////////////////////////////////////////////////
  public void setXmin(double xmin) {this.xmin = xmin;}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set xmax */
  /////////////////////////////////////////////////////////////////////////////
  public void setXmax(double xmax) {this.xmax = xmax;}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set ymin */
  /////////////////////////////////////////////////////////////////////////////
  public void setYmin(double ymin) {this.ymin = ymin;}

  /////////////////////////////////////////////////////////////////////////////
  /** This allows the user to set ymax */
  /////////////////////////////////////////////////////////////////////////////
  public void setYmax(double ymax) {this.ymax = ymax;}

  /////////////////////////////////////////////////////////////////////////////
  /** This zooms to the specified coordinates.  IMPORTANT: these coordinates
  *   are given in terms of the graphics window rather than in terms of
  *   the coordinate system associated with the data sets. */
  /////////////////////////////////////////////////////////////////////////////
  protected void zoom(int xleft, int xright, int ybottom, int ytop) {
    double xxmin, xxmax, yymin, yymax;

    // Important: xValue and yValue depend on xmin, xmax, ymin and ymax.
    // Therefore, these should be modified only after calling xValue
    // and yValue.

    xxmin = xValue(xleft);
    xxmax = xValue(xright);
    yymin = yValue(ybottom);
    yymax = yValue(ytop);
    
    xmin = xxmin;
    xmax = xxmax;
    ymin = yymin;
    ymax = yymax;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This creates the plot: */
  /////////////////////////////////////////////////////////////////////////////
  // g = graphics frame in which to create the plot.
  /////////////////////////////////////////////////////////////////////////////
  public void paint(Graphics g) {
    Graphics2D g2D = (Graphics2D)g;
    Rectangle r = this.getBounds();

    hspan = (float) r.getWidth();
    vspan = (float) r.getHeight();
    
    // I think that r.x, r.y are the offset from the
    // complete frame, and should therefore be removed:
    r.x = 0;
    r.y = 0;
    g2D.setColor(Color.white);
    g2D.fill(r);
    
    // Make a nice frame around the plot:
    r.height--; // this is to see the lower part of the frame
    r.width--;  // this is to see the right part of the frame
    g2D.setColor(Color.black);
    g2D.draw(r);

    int i;
    for(i=0; i < nDataSets; i++) {
      if ((x.get(i) != null)&&(x.get(i).length > 0)) break;
    }

    // Plot tick marks and labels:
    plotYTicks(g2D);
    plotXTicks(g2D);

    // Create a rectangle which corresponds to the axes of the plot:
    Rectangle2D rect = new Rectangle2D.Float(xleft(),yup(),xright()-xleft(),ydown()-yup());

    // Do the rest of the plot:
    plotFrame(g2D,rect);
    g2D.clip(rect);
    for(i=0; i < nDataSets; i++) {
      if (x.get(i) != null) {
        g2D.setColor(colorArray.get(i));
        plotXY(x.get(i),y.get(i),linestyleArray.get(i),g2D);
        plotXYpsym(x.get(i),y.get(i),psymArray.get(i),g2D);
      }
    }
    g2D.setClip(null);
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This draws a rectangle which will shows which region is being zoomed to.*/
  /////////////////////////////////////////////////////////////////////////////
  protected void drawRect(int x0, int y0, int x1, int y1) {

    // get current graphics context:
    Graphics2D g2D = (Graphics2D) this.getGraphics();
    
    GeneralPath f = new GeneralPath();
    f.moveTo(x0,y0);
    f.lineTo(x0,y1);
    f.lineTo(x1,y1);
    f.lineTo(x1,y0);
    f.lineTo(x0,y0);
    
    g2D.setXORMode(Color.white);
    g2D.setStroke(new BasicStroke(1F));
    g2D.draw(f);
  }
    
  
  /////////////////////////////////////////////////////////////////////////////
  /** This writes an IDL file which can reproduce the plot */
  /////////////////////////////////////////////////////////////////////////////
  protected void writeIDLfile(File f, JComponent c) {
    PrintWriter out;
    boolean skipData = false;
    try {
      out = new PrintWriter(new FileWriter(f));
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Unable to write to "+f.getName(),
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    int maxLength = 0;
    for(int i = 0; i < nDataSets; i++) {
      if (x.get(i) != null) {
        if (x.get(i).length > maxLength) maxLength = x.get(i).length;
      }
    }

    if (maxLength == 0) {
      out.println("; Warning: no data - plot not possible.");
      out.close();
      return;
    }

    out.println("; To run me type \"idl "+f.getName()+"\"");
    out.println(";------------------------------------------------------");
    out.println("; This part contains different variables which can be customised");
    out.println("!p.charsize = 1.5 ; change this to change the font size");
    out.println();
    out.println("filename = 'idl.eps' ; this is the name of the output file");
    out.println("title = '!17"+title+"' ; the title of the plot");
    out.println("xtitle = '!17"+xtitle+"' ; the title of the x axis");
    out.println("ytitle = '!17"+ytitle+"' ; the title of the y axis");
    out.println();
    out.println(";------------------------------------------------------");
    out.println("; This part contains various settings for the eps file");
    out.println("; Comment this part out if you want to visualise the plot");
    out.println("; directly from IDL.");
    out.println();
    out.println("set_plot,'ps'");
    out.println("device,file=filename,xsize=15,ysize=12,xoffset=0,yoffset=0,$");
    out.println("       /color,encapsulated=1,Bits_per_Pixel=8");
    out.println("print,'Writing in file ',filename");
    out.println(";------------------------------------------------------");
    out.println();
    out.println("nDataSets = "+nDataSets);
    out.println("xmin = "+xmin);
    out.println("xmax = "+xmax);
    out.println("ymin = "+ymin);
    out.println("ymax = "+ymax);
    out.println("xrange = [xmin, xmax]");
    out.println("yrange = [ymin, ymax]");
    out.println("x = dblarr("+maxLength+","+nDataSets+")");
    out.println("y = dblarr("+maxLength+","+nDataSets+")");
    out.println("x(*,*) = 0");
    out.println("y(*,*) = 0");
    for(int i=0; i<nDataSets; i++) {
      if (x.get(i) != null) {
        for (int j=0; j<x.get(i).length; j++) {
          if (Double.isNaN(x.get(i)[j])||Double.isNaN(y.get(i)[j])) {
            skipData = true;
          } else {
            out.printf("x(%d,%d)=%19.12e%n",j,i,x.get(i)[j]);
            out.printf("y(%d,%d)=%19.12e%n",j,i,y.get(i)[j]);
          }
        }
      }
    }
    out.println();
    out.println("; the plot:");
    out.println("plot,xrange,yrange,/nodata,xrange=xrange,yrange=yrange,$");
    out.println("     xstyle=1,ystyle=1,title=title,$");
    out.println("     xtitle=xtitle,ytitle=ytitle");
    out.println();
    out.println("; This next part defines the different colors used in");
    out.println("; the plot by creating a custom color table:");
    out.println("red =   intarr("+nDataSets+")");
    out.println("green = intarr("+nDataSets+")");
    out.println("blue =  intarr("+nDataSets+")");
    for(int i=0; i<nDataSets; i++) {
      out.printf("red(%d) = %d%n",i,colorArray.get(i).getRed());
      out.printf("green(%d) = %d%n",i,colorArray.get(i).getGreen());
      out.printf("blue(%d) = %d%n",i,colorArray.get(i).getBlue());
    }
    out.println("tvlct,red,green,blue");
    out.println();
    out.println("; The style used for plotting the different lines can");
    out.println("; be modified by changing the value of linestyle.");
    out.println("; The choice of symbols can be modified by changing");
    out.println("; the value of psym.");
    for(int i=0; i<nDataSets; i++) {
      if ((linestyleArray.get(i) != -1)&&(x.get(i) != null)&&(x.get(i).length>0))
        out.printf("oplot,x(0:%d,%d),y(0:%d,%d),linestyle=%d,color=%d,thick=2.0%n",
            x.get(i).length-1,i,x.get(i).length-1,i,linestyleArray.get(i),i);
      if ((psymArray.get(i) != 0)&&(x.get(i) != null)&&(x.get(i).length>0))
        out.printf("oplot,x(0:%d,%d),y(0:%d,%d),psym=%d,color=%d%n",
            x.get(i).length-1,i,x.get(i).length-1,i,psymArray.get(i),i);
    }
    out.println("exit");
    out.close();
    
    if (skipData) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the data was NaN\nand was therefore skipped",
        "Error",
        JOptionPane.WARNING_MESSAGE
      );
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This writes an TXT file which contains the data from the plot. */
  /////////////////////////////////////////////////////////////////////////////
  protected void writeTXTfile(File f, JComponent c) {
    PrintWriter out;
    boolean skipData = false;
    try {
      out = new PrintWriter(new FileWriter(f));
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Unable to write to "+f.getName(),
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      return;
    }

    int maxLength = 0;
    for(int i = 0; i < nDataSets; i++) {
      if (x.get(i) != null) {
        if (x.get(i).length > maxLength) maxLength = x.get(i).length;
      }
    }

    if (maxLength == 0) {
      out.println("Warning: no data.");
      out.close();
      return;
    }

    for(int i=0; i<nDataSets; i++) {
      if (x.get(i) != null) {
        for (int j=0; j<x.get(i).length; j++) {
          if (Double.isNaN(x.get(i)[j])||Double.isNaN(y.get(i)[j])) {
            skipData = true;
          } else {
            out.printf("%19.12e %19.12e%n",x.get(i)[j],y.get(i)[j]);
          }
        }
      }
      out.println("&");
    }
    out.close();
    
    if (skipData) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the data was NaN\nand was therefore skipped",
        "Error",
        JOptionPane.WARNING_MESSAGE
      );
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots the box around the plot and adds tick marks
  // and anotation.
  /////////////////////////////////////////////////////////////////////////////
  private void plotFrame(Graphics2D g2D, Rectangle2D rect) {
    float width, height;

    g2D.setStroke(linestyle[Plot.solid]);
    g2D.draw(rect);

    // Insert title to x-axis:
    g2D.setFont(xyTitleFont);
    FontRenderContext frc = g2D.getFontRenderContext();
    Rectangle2D bounds = xyTitleFont.getStringBounds(xtitle,frc);
    width = (float) bounds.getWidth(); 
    height = getStringHeight(xtitle,xyTitleFont,frc);
    g2D.drawString(xtitle,(xleft()+xright()-width)/2F,vspan-(bottomSubMargin-height)/2F);

    // Insert title to y-axis (including rotating the text font):
    bounds = xyTitleFont.getStringBounds(ytitle,frc);
    width = (float) bounds.getWidth();
    height = getStringHeight(ytitle,xyTitleFont,frc);
    AffineTransform t = g2D.getTransform();
    g2D.rotate(-Math.PI/2.0);
    g2D.drawString(ytitle,-(ydown()+yup()+width)/2F,(leftSubMargin+height)/2F);
    g2D.setTransform(t);

    // Insert top title:
    g2D.setFont(titleFont);
    frc = g2D.getFontRenderContext();
    bounds = titleFont.getStringBounds(title,frc);
    width = (float) bounds.getWidth();
    height = getStringHeight(title,titleFont,frc);
    g2D.drawString(title,(xleft()+xright()-width)/2F,(height+bottomMargin)/2F);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This calculates the tick arrays for either axis.
  /////////////////////////////////////////////////////////////////////////////
  // z1, z2 = bounds for the domain
  /////////////////////////////////////////////////////////////////////////////
  private void findTickArrays(double z1, double z2) {
    if (z1 == z2) {z1-=0.5; z2+=0.5;} // just is case ...
    double zspan = Math.abs(z2-z1);
    double zexp = Math.floor(Math.log10(zspan));
    double r = zspan/Math.pow(10.0,zexp);
    
    // dz  = the large increment (i.e. between major ticks)
    // ddz = the small increment (i.e. between minor ticks)
    // n   = the ratio dx/ddx
    double dz, ddz; int n;

    // this is quite similar to what is done in IDL:
         if (r <= 1.5) {dz = 0.2; ddz = 0.05; n = 4; }
    else if (r <= 3.5) {dz = 0.5; ddz = 0.1;  n = 5; }
    else if (r <= 7.0) {dz = 1.0; ddz = 0.1;  n = 10;}
    else               {dz = 2.0; ddz = 0.5;  n = 4; }

    dz  = dz*Math.pow(10.0,zexp);
    ddz = ddz*Math.pow(10.0,zexp);
    
    // In case "xmax" is smaller than "xmin":
    double zmax, zmin;
    if (z1>z2) {zmax=z1; zmin=z2;}
          else {zmax=z2; zmin=z1;}
              
    // Determine where the ticks start, using the small
    // increment ddx:
    double nstart = Math.ceil(zmin/ddz);
    double zstart = nstart*ddz;

    // Normally I should add 1.0, but this is to avoid truncation
    // errors:
    nticks = (int)(Math.floor(zmax/ddz) - nstart + 1.1);
    
    // Allocate space for the tick arrays:
    ticks = new double[nticks];
    isMajor = new boolean[nticks];
    
    // i is just a counter to let me know when to put a
    // large tick.
    int i = (int)(((long)nstart)%(long)n);
    double zz = zstart;
    
    for (int j = 0; j < nticks; j++) {
      if (i == 0) {
        // we have a major tick
        isMajor[j] = true;
        ticks[j] = zz;
        // Perhaps this is unecessary, but it is there to make sure I
        // get 0 exactly rather than some stupid value like 10e-320:
        if (Math.abs(zz) < dz/2.0) {ticks[j] = 0.0;}
      } else {
        // we have a minor tick
        isMajor[j] = false;
        ticks[j] = zz;
      }
      i++;
      if (i==n) {i=0;}
      zz+=ddz;
    }
    
    
    /////////////////////////////////////
    // This next part is tricky:
    //
    // Determining the number of decimal
    // points and whether or not to keep
    // the exponent.
    /////////////////////////////////////
    
    // a new definition for zmax
    zmax = (Math.abs(z1)>Math.abs(z2)?Math.abs(z1):Math.abs(z2));

    // This gives the number of significant digits - 1(the 0.1 is
    // to avoid truncation errors):
    int nDigits = (int)(Math.floor(Math.log10(zmax))
                - Math.floor(Math.log10(dz)) + 0.1);
    
    int maxExp = (int)(Math.floor(Math.log10(zmax)));

    // This part is really complicated, but the main idea
    // is to try to find the shortest representation of
    // the numbers.
    if (maxExp >= 0) {
      if (maxExp > (nDigits+3)) {
        withExp = true;
        nDecimal = nDigits;
      } else {
        withExp = false;
        nDecimal = nDigits-maxExp;
        if (nDecimal < 0) {nDecimal = 0;}
      }
    } else {
      if (maxExp < -3) {
        withExp = true;
        nDecimal = nDigits;
      } else {
        withExp = false;
        nDecimal = nDigits-maxExp;
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots the tick marks (major and minor) on the x-axis
  /////////////////////////////////////////////////////////////////////////////
  private void plotYTicks(Graphics2D g2D) {
    findTickArrays(ymin,ymax);
    
    // find the maximum width for the y-axis tick labels
    float yLabelWidth = 0;
    float a_width;
    for(int j=0; j < nticks; j++) {
      if (isMajor[j]) {
        a_width = findNumberWidth(ticks[j],g2D);
        yLabelWidth = (a_width>yLabelWidth?a_width:yLabelWidth);
      }
    }

    // adjust left margin accordingly:    
    leftMargin = Math.max(34F + yLabelWidth,leftMinMargin);

    // A few preliminary declarations and preparations:
    GeneralPath f = new GeneralPath();
    double yvalue;
    float yy, vAlign;

    for (int j=0; j < nticks; j++) {
      yvalue = ticks[j];
      yy = yCoor(yvalue);
      if (isMajor[j]) {
        f.moveTo(xleft(),yy);
        f.lineTo(xleft()+yLongTick,yy);
        f.moveTo(xright(),yy);
        f.lineTo(xright()-yLongTick,yy);
        if (Math.abs(yy-ydown()) < 3F) {vAlign = 0F;} else {vAlign = 0.5F;}
        plotNumber(yvalue,nDecimal,withExp,xleft()-4F,yy,1F,vAlign,g2D);
      } else {
        f.moveTo(xleft(),yy);
        f.lineTo(xleft()+yShortTick,yy);
        f.moveTo(xright(),yy);
        f.lineTo(xright()-yShortTick,yy);
      }
    }
    g2D.setStroke(linestyle[Plot.solid]);
    g2D.draw(f);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots the tick marks (major and minor) on the x-axis
  /////////////////////////////////////////////////////////////////////////////
  private void plotXTicks(Graphics2D g2D) {
    findTickArrays(xmin,xmax);
    
    // A few preliminary declarations and preparations:
    GeneralPath f = new GeneralPath();
    double xvalue;
    float xx;

    for (int j=0; j < nticks; j++) {
      xvalue = ticks[j];
      xx = xCoor(xvalue);
      if (isMajor[j]) {
        f.moveTo(xx,ydown());
        f.lineTo(xx,ydown()-xLongTick);
        f.moveTo(xx,yup());
        f.lineTo(xx,yup()+xLongTick);
        plotNumber(xvalue,nDecimal,withExp,xx,ydown()+17F,0.5F,0F,g2D);
      } else {
        f.moveTo(xx,ydown());
        f.lineTo(xx,ydown()-xShortTick);
        f.moveTo(xx,yup());
        f.lineTo(xx,yup()+xShortTick);
      }
    }
    g2D.setStroke(linestyle[Plot.solid]);
    g2D.draw(f);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots a number using scientific notation, and centers it around
  // a particular position.
  /////////////////////////////////////////////////////////////////////////////
  // z       = the number to be plotted
  // g2D     = graphics onto which to plot the number
  /////////////////////////////////////////////////////////////////////////////
  private float findNumberWidth(double z, Graphics2D g2D) {
    FontRenderContext frc = g2D.getFontRenderContext();
    Rectangle2D bounds;
    float  width;
    String labelCoef, labelExp;
    String sformat;

    if (z == 0) {
      bounds = annotationFont.getStringBounds("0",frc);
      width = (float) bounds.getWidth(); 
      return width;
    }
    
    if (withExp) {
      ////////////////////////////////////
      // The number uses an exponent
      ////////////////////////////////////
      double zexp,r;
      
      // separate number into mantissa (r) and exponent (zexp):
      zexp = Math.floor(Math.log10(Math.abs(z)));
      r = z/Math.pow(10.0,zexp);

      sformat = "%."+nDecimal+"f";
      labelCoef = String.format(sformat,r);
      labelExp = String.format("%.0f",zexp);

      // get dimensions of different components:
      bounds = annotationFont.getStringBounds(labelCoef,frc);
      width = (float) bounds.getWidth(); 

      bounds = timesFont.getStringBounds(timesString,frc);
      width += (float) bounds.getWidth();

      bounds = annotationFont.getStringBounds(tenString,frc);
      width += (float) bounds.getWidth();
    
      bounds = exponentFont.getStringBounds(labelExp,frc);
      width += (float) bounds.getWidth();

      // return the result:
      return width-1F;

    } else {
      //////////////////////////////////////
      // The number is without an exponent
      //////////////////////////////////////
      sformat = "%."+nDecimal+"f";
      labelCoef = String.format(sformat,z);

      // get dimensions of different components:
      bounds = annotationFont.getStringBounds(labelCoef,frc);
      width = (float) bounds.getWidth(); 
      return width;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots a number using scientific notation at a particular location
  // (xxpos, yypos), with an horizontal and vertical alignment determined by
  // (hpos, ypos).
  /////////////////////////////////////////////////////////////////////////////
  // z       = the number to be plotted
  // n       = the number of digits after the decimal point
  // wthExp  = if true, use exponent, otherwise, omit exponent
  // xxpos   = x position around which to center the number
  // yypos   = y position around which to center the number
  // hAlign  = this specifies how to align the number horizontally:
  //             0f = flushleft, 1f = flushright
  // vAlign  = this specifies how to align the number vertically:
  //             0f = [baseline = yypos], 1f = [topline = yypos]
  // g2D     = graphics onto which to plot the number
  /////////////////////////////////////////////////////////////////////////////
  private void plotNumber(double z, int n, boolean wthExp,
                          float xxpos, float yypos, float hAlign, float vAlign,
                          Graphics2D g2D) {
    FontRenderContext frc = g2D.getFontRenderContext();
    Rectangle2D bounds;
    LineMetrics metrics;
    float  widthCoef, widthExp, widthTimes, widthTen, width, height, xx, yy;
    String labelCoef, labelExp;
    String sformat;

    // Treat zero separatly.  It will have 0 decimal places and
    // no exponent.
    if (z==0) {
      labelCoef = "0";

      // get dimensions of different components:
      bounds = annotationFont.getStringBounds(labelCoef,frc);
      widthCoef = (float) bounds.getWidth(); 
      height = getStringHeight(labelCoef, annotationFont, frc);

      // find starting position:
      xx = xxpos - hAlign*widthCoef;
      yy = yypos + vAlign*height;

      // plot the number:
      g2D.setFont(annotationFont);
      g2D.drawString(labelCoef,xx,yy);
      return;
    }
      
    if (wthExp) {
      ////////////////////////////////////
      // Plot the number using an exponent
      ////////////////////////////////////
      double zexp,r;
      
      // separate number into mantissa (r) and exponent (zexp):
      zexp = Math.floor(Math.log10(Math.abs(z)));
      r = z/Math.pow(10.0,zexp);

      sformat = "%."+n+"f";
      labelCoef = String.format(sformat,r);
      labelExp = String.format("%.0f",zexp);

      // get dimensions of different components:
      bounds = annotationFont.getStringBounds(labelCoef,frc);
      widthCoef = (float) bounds.getWidth(); 

      bounds = timesFont.getStringBounds(timesString,frc);
      widthTimes = (float) bounds.getWidth();

      bounds = annotationFont.getStringBounds(tenString,frc);
      widthTen = (float) bounds.getWidth();
    
      // I'm assuming that the exponent determines the maximum
      // height:
      bounds = exponentFont.getStringBounds(labelExp,frc);
      widthExp = (float) bounds.getWidth();
      metrics = exponentFont.getLineMetrics(labelExp,frc);
      height = 5F + getStringHeight(labelExp,exponentFont,frc);

      // find starting position:
      width = widthCoef+widthTimes+widthTen+widthExp-1F;
      
      // this is to avoid having the number go off the right edge:
      if ((xxpos + (1F-hAlign)*width) > (hspan-2F)) {
        xx = hspan - width - 2F;
      } else {
        xx = xxpos - hAlign*width;
      }
      yy = yypos + vAlign*height;

      // plot the number:
      g2D.setFont(annotationFont);
      g2D.drawString(labelCoef,xx,yy);
      g2D.drawString(tenString,xx+widthCoef+widthTimes,yy);
      g2D.setFont(timesFont);
      g2D.drawString(timesString,xx+widthCoef+1,yy-1);
      g2D.setFont(exponentFont);
      g2D.drawString(labelExp,xx+widthCoef+widthTimes+widthTen-1F,yy-5F);
    } else {
      //////////////////////////////////////
      // Plot the number without an exponent
      //////////////////////////////////////
      sformat = "%."+n+"f";
      labelCoef = String.format(sformat,z);

      // get dimensions of different components:
      bounds = annotationFont.getStringBounds(labelCoef,frc);
      widthCoef = (float) bounds.getWidth(); 
      height = getStringHeight(labelCoef, annotationFont, frc);

      // find starting position:
      width = widthCoef;
      // this is to avoid having the number go off the right edge:
      if ((xxpos + (1F-hAlign)*width) > (hspan-2F)) {
        xx = hspan - width - 2F;
      } else {
        xx = xxpos - hAlign*width;
      }
      yy = yypos + vAlign*height;

      // plot the number:
      g2D.setFont(annotationFont);
      g2D.drawString(labelCoef,xx,yy);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  // This function calculates the height of a string:
  /////////////////////////////////////////////////////////////////////////////
  private float getStringHeight(String s, Font f, FontRenderContext frc) {
    LineMetrics metrics = f.getLineMetrics(s, frc);
    float ascent = metrics.getAscent();         // from baseline to top of text
    float descent = metrics.getDescent();       // from baseline to bottom of text
    //float leading = metrics.getLeading();     // additional space used in
                                                // between lines of text
    //float lineheight = metrics.getHeight();   // The total height of our text
                                                // = ascent + descent + leading

    // This may look incorrect, but it seems that in most cases, the ascent
    // actually overshoots the top of the text by an amount that's close to the
    // descent, hence the formula. Furthermore, I only want to center the ascent,
    // which is more aesthetic.
    // Final remark: the leading usually seems to be 0, so I do not use it.
    return ascent-descent;
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots a data set using the linestyle aLinestyle:
  /////////////////////////////////////////////////////////////////////////////
  // There is probably a simpler way of doing this, but I do not know it.
  // This function not only plots the function but clips the GeneralPath f
  // to a circular region which encompases the strict plotting area.  This
  // is to avoid spending a long time plotting line segments significantly
  // outside the plot area.
  /////////////////////////////////////////////////////////////////////////////

  private void plotXY(double[] x, double[] y, int aLinestyle, Graphics2D g2D) {

    if (aLinestyle == -1) {return;}

    GeneralPath f = new GeneralPath();
    boolean isInside1, isInside2;
    float x1, y1, x2, y2, xTemp, yTemp, tt, dt;
    float xcenter, ycenter, radius2;

    // Rather than simply plotting and clipping, it is
    // necessary to clip beforehand to avoid wasting a
    // lot of time plotting outside the plot area.

    // the center of the plot area:
    xcenter = 0.5F*(xleft()+xright());
    ycenter = 0.5F*(yup()+ydown());

    // set radius2 slightly larger than half the diagonal squared
    radius2 = 0.3F*norm2(xright()-xleft(),ydown()-yup());
  
    x2 = xCoor(x[0]);
    y2 = yCoor(y[0]); 
    isInside2 = isInside(x2,y2,xcenter,ycenter,radius2);
    if (isInside2) {f.moveTo(x2,y2);}

    for (int i = 1; i < x.length; i++) {

      x1 = x2;
      y1 = y2;
      isInside1 = isInside2;

      x2 = xCoor(x[i]);
      y2 = yCoor(y[i]);
      isInside2 = isInside(x2,y2,xcenter,ycenter,radius2);

      // Both points inside:
      //////////////////////
      if (isInside1&&isInside2) {
        f.lineTo(x2,y2);
        continue;
      }

      // check to see if both points are identitical:
      if (0F == norm2(x2-x1,y2-y1)) {continue;}

      // First point inside, second point outside:
      ////////////////////////////////////////////
      if (isInside1) {
        // an approximate upper bound for tt:
        tt = 2F * (float)Math.sqrt(radius2/norm2(x2-x1,y2-y1));
        xTemp = x1 + tt*(x2-x1);
        yTemp = y1 + tt*(y2-y1);
        f.lineTo(xTemp,yTemp);
        continue;
      }
      
      // First point outside, second point inside:
      ////////////////////////////////////////////
      if (isInside2) {
        // an approximate upper bound for tt:
        tt = 2F * (float)Math.sqrt(radius2/norm2(x2-x1,y2-y1));
        xTemp = x2 + tt*(x1-x2);
        yTemp = y2 + tt*(y1-y2);
        f.moveTo(xTemp,yTemp);
        f.lineTo(x2,y2);
        continue;
      }

      // Both points outside, the most complicated case:
      //////////////////////////////////////////////////
      
      // check to see if the segment (x1,y1)-(x2,y2) intersects the circular
      // plot region:
      tt = ((xcenter-x1)*(x2-x1)+(ycenter-y1)*(y2-y1))/norm2(x2-x1,y2-y1);
      // no intersetion so go to next iteration:
      if ((tt < 0F)||(tt > 1F)) {continue;} 
      // the segment might intersect the plot region:
      dt = (float)Math.sqrt(radius2/norm2(x2-x1,y2-y1));
      xTemp = x1 + (tt-dt)*(x2-x1);
      yTemp = y1 + (tt-dt)*(y2-y1);
      f.moveTo(xTemp,yTemp);
      xTemp = x1 + (tt+dt)*(x2-x1);
      yTemp = y1 + (tt+dt)*(y2-y1);
      f.lineTo(xTemp,yTemp);
    }
         
    g2D.setStroke(linestyle[aLinestyle]);
    g2D.draw(f);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This plots a data set using the symbols designated by aPsym.
  /////////////////////////////////////////////////////////////////////////////
  private void plotXYpsym(double[] x, double[] y, int aPsym, Graphics2D g2D) {

    if (aPsym == 0) {return;}

    GeneralPath f = new GeneralPath();
    float xx, yy;

    for (int i = 0; i < x.length; i++) {
      xx = xCoor(x[i]);
      yy = yCoor(y[i]);
      for (int j = 0; j < psym[aPsym].length; j+=4) {
        f.moveTo(xx+psym[aPsym][j],yy+psym[aPsym][j+1]);
        f.lineTo(xx+psym[aPsym][j+2],yy+psym[aPsym][j+3]);
      }
    }
    
    g2D.setStroke(Plot.psymStroke);
    g2D.draw(f);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This function returns "true" when a specific point lies inside the plot
  // area (which is a circle centered on (xcenter,ycenter) and of radius
  // sqrt(radius2)), and "false" otherwise.
  /////////////////////////////////////////////////////////////////////////////
  private static boolean isInside(float xx, float yy, float xcenter,
                                  float ycenter, float radius2) {
    float distance2 = norm2(xx-xcenter,yy-ycenter);
    return (distance2 <= radius2);
  }

  /////////////////////////////////////////////////////////////////////////////
  // This function returns the squared norm of vector (xx,yy).
  /////////////////////////////////////////////////////////////////////////////
  private static float norm2(float xx, float yy) {
    return xx*xx+yy*yy;
  }

  /////////////////////////////////////////////////////////////////////////////
  // This converts x values to x coordinates:
  /////////////////////////////////////////////////////////////////////////////
  private float xCoor(double xvalue) {
    return (float)(xleft() + (xright()-xleft())*(xvalue-xmin)/(xmax-xmin));
  }

  /////////////////////////////////////////////////////////////////////////////
  // This converts y values to y coordinates:
  /////////////////////////////////////////////////////////////////////////////
  private float yCoor(double yvalue) {
    return (float)(ydown() + (yup()-ydown())*(yvalue-ymin)/(ymax-ymin));
  }

  /////////////////////////////////////////////////////////////////////////////
  // This converts x coordinates to x values:
  /////////////////////////////////////////////////////////////////////////////
  private double xValue(int xcoor) {
    return  xmin+(xmax-xmin)*((float)xcoor-xleft())/(xright()-xleft());
  }

  /////////////////////////////////////////////////////////////////////////////
  // This converts y coordinates to y values:
  /////////////////////////////////////////////////////////////////////////////
  private double yValue(int ycoor) {
    return ymin+(ymax-ymin)*((float)ycoor-ydown())/(yup()-ydown());
  }

  /////////////////////////////////////////////////////////////////////////////
  /** Set all arrays contained in x and y to null, but keep the same number of
  *   data sets. */
  /////////////////////////////////////////////////////////////////////////////
  public void nullify() {
    for(int i=0; i<x.size(); i++) {
      x.set(i,null);
      y.set(i,null);
    }
    // change the bounds
    xmin=0; xmax=1; ymin=0; ymax=1;
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This nullifies a specific data set in a plot by setting its arrays 
  *   contained in graphicsArea.x and graphicsArea.y to null */
  /////////////////////////////////////////////////////////////////////////////
  public void nullify(int i) {
    if ((i >= 0)&&(i < x.size())) {
      x.set(i,null);
      y.set(i,null);
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  /** This clones the plot */
  /////////////////////////////////////////////////////////////////////////////
  public Plot clone() {
    Plot newPlot = new Plot(title,xtitle,ytitle);
    for(int i=0; i<nDataSets; i++) {
      double[] x, y;
      if (this.x.get(i) == null) {
        x = null;
      } else {
        x = this.x.get(i).clone();
      }
      if (this.y.get(i) == null) {
        y = null;
      } else {
        y = this.y.get(i).clone();
      }
      Color aColor = new Color(colorArray.get(i).getRGB());
      newPlot.addDataSet(x,y,linestyleArray.get(i),psymArray.get(i),aColor);
      newPlot.setXmin(xmin);
      newPlot.setXmax(xmax);
      newPlot.setYmin(ymin);
      newPlot.setYmax(ymax);
      
    }
    return newPlot;
  }
}
