/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008
*
****************************************************************************/
package inversion_applet;

public class Polytrope {

  /** the normalised enthalpy (h[0] = 1): */
  public double[] h;
  
  /** the radial derivative of the normalised enthalpy: */
  public double[] hp;
  
  /** the noramlised grid (r[last_point] = 1): */
  public double[] r;
  
  /** the radius, as given by the Lane-Emden equation */
  public double Xi;
  
  /** the ratio of the central density to the average density: */
  public double Alpha;
  
  // the polytropic index (this is set on private to avoid
  // unacceptable values):
  private double N;


  ////////////////////////////////////////////////////////////////
  /** this returns the polytropic index */
  ////////////////////////////////////////////////////////////////
  public double getN() {return N;}

  ////////////////////////////////////////////////////////////////
  /** this sets the polytropic index */
  ////////////////////////////////////////////////////////////////
  public void setN(double N) {
    if (N > 4.9) 
      throw new IllegalArgumentException("Please choose a lower polytropic index (N <= 4.9)");
    if (N < 0.0)
      throw new IllegalArgumentException("Please choose a non-negative polytropic index");
    this.N = N;
  }

  /***************************************************************
  *  The main constructor.  This is where the polytrope is
  *  generated
  ***************************************************************/
  public Polytrope(int nstep, double N) {
    this.setN(N);
    double k1, k1p, k2, k2p, k3, k3p, k4, k4p;
    double rprev = 0.0;
    double rnext = 18.0/(5.0-N); // This provides a first estimate
                                 // of the radius but is in fact an
                                 // overestimate.
    double step = rnext/(double)(nstep-1);
    
    r  = new double[nstep];
    h  = new double[nstep];
    hp = new double[nstep];
    h[0] = 1.0;
    
    double eps = 1e-10;
    while(Math.abs(rprev-rnext) > eps) {
      int nr = 0;
      while((h[nr] > 0.0) && (nr < (nstep-1))) {

        // Runge-Kutta 4 method applied to polytropic equation:
        k1  = step*hp[nr];
        k1p = step*fp(r[nr],h[nr],hp[nr],N);

        k2  = step*(hp[nr]+k1p/2.0);
        k2p = step*fp(r[nr]+step/2.0,h[nr]+k1/2.0,hp[nr]+k1p/2.0,N);

        k3  = step*(hp[nr]+k2p/2.0);
        k3p = step*fp(r[nr]+step/2.0,h[nr]+k2/2.0,hp[nr]+k2p/2.0,N);

        k4  = step*(hp[nr]+k3p);
        k4p = step*fp(r[nr]+step,h[nr]+k3,hp[nr]+k3p,N);

        r[nr+1]  = r[nr]  + step;
        h[nr+1]  = h[nr]  + (k1 + 2.0*k2 + 2.0*k3 + k4)/6.0;
        hp[nr+1] = hp[nr] + (k1p+ 2.0*k2p+ 2.0*k3p+ k4p)/6.0;
        nr++;
      }
      
      // must interpolate/extrapolate to obtain last point
      double alpha  = h[nr]/(h[nr]-h[nr-1]);
      r[nr]  = alpha*r[nr-1]  + (1.0-alpha)*r[nr];
      h[nr]  = alpha*h[nr-1]  + (1.0-alpha)*h[nr];
      hp[nr] = alpha*h[nr-1]  + (1.0-alpha)*hp[nr];
      step   = r[nr]/(double)(nstep-1);
      
      rprev = rnext;
      rnext = r[nr];
    }
    Xi = r[nstep-1];
    for(int i=0; i<nstep; i++) r[i] = r[i]/r[nstep-1];
    Alpha = findAlpha();
  }

  ////////////////////////////////////////////////////////////////
  /** Another constructor.  N = 0, number of grid points = 1001 */
  ////////////////////////////////////////////////////////////////
  public Polytrope()          {this(1001,0.0);}

  ////////////////////////////////////////////////////////////////
  /** Another constructor.  Number of grid points = 1001 */
  ////////////////////////////////////////////////////////////////
  public Polytrope(double N)  {this(1001,N);}

  ////////////////////////////////////////////////////////////////
  /** Another constructor.  N = 0 */
  ////////////////////////////////////////////////////////////////
  public Polytrope(int nstep) {this(nstep,0.0);}

  ////////////////////////////////////////////////////////////////
  /** Another constructor */
  ////////////////////////////////////////////////////////////////
  public Polytrope(double N, int nstep) {this(nstep,N);}
  
  ////////////////////////////////////////////////////////////////
  // A function that is used when constructing the polytrope
  ////////////////////////////////////////////////////////////////
  private static double fp(double r, double h, double hp, double N) {
    double result;
    if (Math.abs(r) < 1e-10) result = - Math.pow(Math.abs(h),N)/3.0;
    else result = -2.0*hp/r - Math.pow(Math.abs(h),N);
    return result;
  }
 
  ////////////////////////////////////////////////////////////////
  // This calculates Alpha, once the polytrope has been generated
  ////////////////////////////////////////////////////////////////
  private double findAlpha() {
    double sum = 0.0;
    int nstep = r.length;

    // integrate interior:
    for(int i=3; i<nstep-3; i++)
      sum+=(Math.pow(Math.abs(h[i]),N)*r[i]*r[i]);
   
    // treat ends separately: 
    sum+=(9.0/24.0*Math.pow(Math.abs(h[0]),N)*r[0]*r[0]);
    sum+=(9.0/24.0*Math.pow(Math.abs(h[nstep-1]),N)*r[nstep-1]*r[nstep-1]);
    sum+=(7.0/6.0*Math.pow(Math.abs(h[1]),N)*r[1]*r[1]);
    sum+=(7.0/6.0*Math.pow(Math.abs(h[nstep-2]),N)*r[nstep-2]*r[nstep-2]);
    sum+=(23.0/24.0*Math.pow(Math.abs(h[2]),N)*r[2]*r[2]);
    sum+=(23.0/24.0*Math.pow(Math.abs(h[nstep-3]),N)*r[nstep-3]*r[nstep-3]);

    return (double)(nstep-1)/(3.0*sum);
  }
      
  ////////////////////////////////////////////////////////////////
  // This is to test polytrope on its own. (note: the package
  // command needs to be commented to run this on its own).
  ////////////////////////////////////////////////////////////////
  public static void main(String[] argv) {
    Polytrope toto;
    double[] NN = {1.0, 1.33, 1.5, 1.67, 1.9, 2.0, 2.5, 3.0, 3.5, 4.0};

    for(int i=0; i<NN.length; i++) {
      toto = new Polytrope(1001,NN[i]);
      System.out.printf("%.2f %.3f %.5f %n",NN[i],toto.Xi, toto.Alpha);
    }
  }
}
