/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008
*
****************************************************************************/
package inversion_applet;

import inversion_applet.DataCenter;
import inversion_applet.Numerics;
import org.netlib.blas.Dgemv;
import org.netlib.lapack.Dsytrf;
import org.netlib.lapack.Dsytrs;
import org.netlib.lapack.Ilaenv;
import java.util.Random;
import java.util.ArrayList;
import java.util.Vector;
import javax.swing.JOptionPane;
import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JTable;

/*************************************************************************
**
**    RotationInversion
**                                              Version 1.0   Mai 2008
**
*************************************************************************/

public class RotationInversion { 

  private DataCenter dataCenter;
  protected boolean canInvert;   // after initialisation, if the provided information
                                 // is insufficient or erroneous, it is not possible
                                 // to do an inversion, in which case, canInvert = false
                                 // Otherwise, canInvert = true
  protected int n_inputKernels;  // the number of kernels used in the inversion
  public double[] averagingKernelsRLS;  // Kernels obtained after a RLS inversion
  public double[] averagingKernelsSOLA; // Kernels obtained after a SOLA inversion
                                 // 1st coordinate: grid position
                                 // 2nd coordinate: target position
  public double[] f_invertedRLS; // function given by the RLS inversion (f_inverted[n_grid])
  public double[] RLS_errorBars; // this gives the uncertainty in the inversion, based on
                                 // the error bars of the frequency shifts
  public double[] f_invertedSOLA;// function given by the SOLA inversion (f_inverted[n_grid])
  public double[] SOLA_errorBars;// this gives the uncertainty in the inversion, based on
                                 // the error bars of the frequency shifts
  
  // various intermediate matrices used in the inverstions:
  private double[] fitDataMatRLS;        // matrix used for fitting the data in a RLS inversion
  private double[] regularisationMatRLS; // matrix used for regularisation in a RLS inversion
  private double[] kernelMatRLS;         // matrix used for calculating output kernels
  private double[] fitDataMatSOLA;       // matrix used for fitting the data in a SOLA inversion
  private double[] regularisationMatSOLA;// matrix used for regularisation in a SOLA inversion
  private double[] targetKernelMatSOLA;  // matrix used for calculating the target kernels
  private org.netlib.util.intW info = new org.netlib.util.intW(2); // necessary for lapack
  private ArrayList<Integer> intersection = new ArrayList<Integer>();
  private ArrayList<Double>  dataValues   = new ArrayList<Double>();
  private ArrayList<Double>  dataErrors   = new ArrayList<Double>();
  private double[][] rotaKernel;
  private int lwork;  // an integer used in the matrix solvers
  private double[] weights;
  private double[] derive;
  
  /********************************************************
  *  A constructor which does the initialisation
  ********************************************************/
  public RotationInversion(DataCenter dataCenter) {
    this.dataCenter = dataCenter;

    // At the beginning, no data is loaded, so no inversion
    // can be done.  Therefore the inversion is up to date.
    dataCenter.rotationInversionUpToDate = true;
    this.canInvert = false;

    // This determines the blocksize to be used in the symmetric
    // matrix solvers. I will just put some average matrix size
    // rather than an exact number.
    this.lwork = Ilaenv.ilaenv(1,"DSYTRF","U",100,100,-1,-1);
  }

  ////////////////////////////////////////////////////////////
  /** This calls various initialisation methods */
  ////////////////////////////////////////////////////////////
  public void init(JComponent c) {
    canInvert = true;
    dataCenter.rotationInversionUpToDate = true;
    n_inputKernels = 0;
    initIntersection(c);
    if (canInvert) {
      initGridInversionCheb();
      initRotaKernel();
      initNoise();
      initRLS();
      initSOLA();
      initDeltaSOLA();
    }
  }

  ////////////////////////////////////////////////////////////
  // This finds the intersection between the set of kernels
  // and the set of frequency shift data and stores it in
  // an ArrayList called intersection
  ////////////////////////////////////////////////////////////
  private void initIntersection(JComponent c) {

    boolean unusedRS = false;
    boolean dataWithZeroError = false;
    boolean incompleteRS = false;
    boolean incorrectRS = false;
    boolean duplicateRS = false;
    boolean incorrectK = false;
    boolean duplicateK = false;

    Vector<Vector<Object>> unusedData = new Vector<Vector<Object>>();

    // clear previous data:
    intersection.clear();
    dataValues.clear();
    dataErrors.clear();

    // make sure rotational splittings are loaded:
    if (!dataCenter.RSloaded) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No frequency shift data.  Unable to proceed",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      n_inputKernels = 0;
      canInvert = false;
      return;
    }

    // make sure kernels are loaded:
    if (!dataCenter.kernelsLoaded) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No kernels loaded.  Unable to proceed",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      n_inputKernels = 0;
      canInvert = false;
      return;
    }

    int nnK  = dataCenter.nK.length;
    int nnRS = dataCenter.nRS.size();
    int[] indexK  = dataCenter.sortKernels();
    int[] indexRS = dataCenter.sortRS();

    // find intersection on sorted arrays
    int iK = 0;
    int iRS = 0;
    int cmp;
    while ((iRS < nnRS)&&(iK < nnK)) {
      
      // test the frequency data to make sure that
      // none of the fields are set to null:
      if ((dataCenter.nRS.get(indexRS[iRS]) == null)||
          (dataCenter.lRS.get(indexRS[iRS]) == null)||
          (dataCenter.RS.get(indexRS[iRS]) == null)||
          (dataCenter.ErrorRS.get(indexRS[iRS]) == null)) {
        incompleteRS = true;
        iRS++;
        continue;
      }

      // make sure that all the error bars are strictly
      // greater than zero:
      if (dataCenter.ErrorRS.get(indexRS[iRS]) <= 0) {
        dataWithZeroError = true;
        iRS++;
        continue;
      }

      // make sure that n and l have possible values
      // (i.e. l>1) the rotational splittings:
      if (dataCenter.lRS.get(indexRS[iRS])<1) {
        incorrectRS = true;
        iRS++;
        continue;
      }

      // check for duplicate rotational splittings:
      if ((iRS>0)&&equalsRS(indexRS[iRS],indexRS[iRS-1])) {
        duplicateRS = true;
        iRS++;
        continue;
      }

      // make sure that n and l have possible values
      // (i.e. l>0 for n>0 and l>1 for n<=0) for
      // kernel data:
      if (dataCenter.nK[indexK[iK]]<=0) {
        if (dataCenter.lK[indexK[iK]]<1) {
          incorrectK = true;
          iK++;
          continue;
        }
      } else {
        if (dataCenter.lK[indexK[iK]]<0) {
          incorrectK = true;
          iK++;
          continue;
        }
      }

      // check for duplicate kernel data:
      if ((iK>0)&&
          (dataCenter.nK[indexK[iK-1]]==dataCenter.nK[indexK[iK]]) &&
          (dataCenter.lK[indexK[iK-1]]==dataCenter.lK[indexK[iK]])) {
        duplicateK = true;
        iK++;
        continue;
      }
     
      // do the comparison:
      cmp = compareRSKernel(indexRS[iRS],indexK[iK]);
      
      // if K[iK] is smaller, increase iK:
      if (cmp > 0) {
        iK++;
        continue;
      }
      
      // If RS[iRS] is smaller, increase iRS:
      if (cmp < 0) {
        Vector<Object> temp = new Vector<Object>();
        temp.add(new Integer(dataCenter.nRS.get(indexRS[iRS])));
        temp.add(new Integer(dataCenter.lRS.get(indexRS[iRS])));
        temp.add(new Double(dataCenter.RS.get(indexRS[iRS])));
        temp.add(new Double(dataCenter.ErrorRS.get(indexRS[iRS])));
        unusedData.add(temp);
        iRS++;
        unusedRS = true;
        continue;
      }
      
      // RS[iRS] == K[iK]: add this to intersection:
      intersection.add(indexK[iK]);
      dataValues.add(dataCenter.RS.get(indexRS[iRS]));
      dataErrors.add(dataCenter.ErrorRS.get(indexRS[iRS]));
      iRS++;
      iK++;
    }

    if (intersection.size() == 0) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No common modes (n,\u2113) between the kernel data\n"+
        "set and the rotational splittings.  Unable to proceed.",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      canInvert = false;
      n_inputKernels = 0;
      return;
    }

    if (dataWithZeroError) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the rotational splittings have\n"+
        "a 0 error bar, which is not allowed.\n"+
        "This data has consequently been ignored.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
      
    if (duplicateRS) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "The rotational splittings contains modes with \n"+
        "duplicate (n,\u2113) which have been ignored.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    if (incompleteRS) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the rotational splittings were\n"+
        "incomplete and were therefore not used.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
    
    if (incorrectRS) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the rotational splittings seemed to be erroneous:\n"+
        "  (\u2113 < 1).\n"+
        "This data has consequently not been not used.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
    
    if (duplicateK) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "The kernel data contains modes with \n"+
        "duplicate (n,\u2113) which has been ignored.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    if (incorrectK) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the kernel data seemed to be erroneous:\n"+
        "  (\u2113 < 1 for n \u2264 0) or (\u2113 < 0 for n > 0).\n"+
        "This data has consequently not been not used.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    if (unusedRS||(iRS < nnRS)) {
      String message = "Some of the rotational splittings were not used.\n"+
                       "Loading the appropriate kernels would allow\n"+
                       "use of this data.";
      String[] options = {"View skipped data", "OK"};
      int result = JOptionPane.showOptionDialog(
        c,                            // the parent that the dialog blocks
        message,                      // the dialog message array
        "Warning",                    // the title of the dialog window
        JOptionPane.DEFAULT_OPTION,   // option type
        JOptionPane.QUESTION_MESSAGE, // message type
        null,                         // optional icon, use null to use the default icon
        options,                      // options string array, will be made into buttons
        options[1]                    // option that should be made into a default button
      );
      if (result == 0) {
        for(int i=iRS; i<nnRS; i++) {
          Vector<Object> temp = new Vector<Object>();
          temp.add(new Integer(dataCenter.nRS.get(indexRS[iRS])));
          temp.add(new Integer(dataCenter.lRS.get(indexRS[iRS])));
          temp.add(new Double(dataCenter.RS.get(indexRS[iRS])));
          temp.add(new Double(dataCenter.ErrorRS.get(indexRS[iRS])));
          unusedData.add(temp);
        }

        Vector<String> columns = new Vector<String>();
        columns.add("n");
        columns.add("\u2113");
        columns.add("R.S.");
        columns.add("Error(R.S.)");

        JTable aTable = new JTable(unusedData,columns) {
          public boolean isCellEditable(int row, int col) {return false;}
        };
        JScrollPane scroller = new JScrollPane();
        scroller.getViewport().add(aTable);
        Object[] contents = new Object[2];
        contents[0] = "List of rotational splittings which\n"+
                      "are not included in the inversion.";
        contents[1] = scroller;
        JOptionPane.showMessageDialog(
          c, // parent component
          contents,
          "List of unused data",
          JOptionPane.INFORMATION_MESSAGE
        );
      }
    }
    n_inputKernels = intersection.size();
  }

  ////////////////////////////////////////////////////////////
  // Returns true if the two rotational splittings have the
  // same (n,l) and false otherwise.  Returns false if
  // n or l of eithers rotational splitting is null.
  ////////////////////////////////////////////////////////////
  private boolean equalsRS(int i1, int i2) {
    if (dataCenter.nRS.get(i1) == null) return false;
    if (dataCenter.lRS.get(i1) == null) return false;
    if (dataCenter.nRS.get(i2) == null) return false;
    if (dataCenter.lRS.get(i2) == null) return false;
    return
      (dataCenter.nRS.get(i2).compareTo(dataCenter.nRS.get(i1))==0)&&
      (dataCenter.lRS.get(i2).compareTo(dataCenter.lRS.get(i1))==0);
  }

  ////////////////////////////////////////////////////////////
  // Compare (n,l) from a frequancy-shift data (indexed by iRS) and 
  // from a Kernel (indexed by iK).
  // if RS(iRS) < K(iK),  returns negative int
  // if RS(iRS) == K(iK), return 0
  // if RS(iRS) > K(iK),  return positive int
  ////////////////////////////////////////////////////////////
  private int compareRSKernel(int iRS, int iK) {
    int compareL = dataCenter.lRS.get(iRS).compareTo(Integer.valueOf(dataCenter.lK[iK]));
    if (compareL == 0) {
      return dataCenter.nRS.get(iRS).compareTo(Integer.valueOf(dataCenter.nK[iK]));
    } else {
      return compareL;
    }
  }

  ////////////////////////////////////////////////////////////
  // This initialises a grid to be used in the inversion.
  // For a given number of grid points, it is not as
  // accurate as initGridInversionCheb() below.
  ////////////////////////////////////////////////////////////
  private void initGridInversion() {
    int n_grid = dataCenter.n_gridR;

    // obtain the sound velocity:
    double[] c = new double[dataCenter.c2.length];
    for(int i=0; i<c.length; i++) c[i] = Math.sqrt(dataCenter.c2[i]);

    // this is to avoid division by zero later on:
    double max_c = 0.0;
    for(int i=0; i<c.length; i++) if (max_c < c[i]) max_c = c[i];
    for(int i=0; i<c.length; i++) if ((c[i]/max_c) < 1e-6) c[i] = max_c*1e-6;

    // find the inverse sound velocity:
    for(int i=0; i<c.length; i++) c[i] = 1.0/Math.sqrt(c[i]);

    double[] CC = Numerics.findAntiDerivative(dataCenter.gridModel,c);
    double cInt = CC[CC.length-1]-CC[0];

    dataCenter.gridRInversion = new double[n_grid];
    dataCenter.gridRInversion[0] = dataCenter.gridModel[0];
    dataCenter.gridRInversion[n_grid-1] =
       dataCenter.gridModel[dataCenter.gridModel.length-1];

    int j = 0;
    double cTarget, lambda;
    for(int i=1; i<n_grid-1; i++) {
      cTarget = cInt*(double)i/(double)(n_grid-1);  // a target value
      while(CC[j]<cTarget) j++;
      lambda = (CC[j]-cTarget)/(CC[j]-CC[j-1]);
      dataCenter.gridRInversion[i] = dataCenter.gridModel[j-1]*lambda 
                                  + dataCenter.gridModel[j]*(1.0-lambda);
      j--;
    }
    
    // initialise some integration weights:
    weights = Numerics.findIntegrationWeights(dataCenter.gridRInversion);
    // initialise a derivation matrix:
    double[][] derive_temp = Numerics.initDeriveFD(dataCenter.gridRInversion, 2, 2);
    derive = derive_temp[2];
  }

  ////////////////////////////////////////////////////////////
  // This initialises a Chebyshev collocation grid and spectral
  // integration weights to be used in the inversion.  For a
  // given number of grid points, it is more accurate than
  // initGridInversion() above.
  ////////////////////////////////////////////////////////////
  private void initGridInversionCheb() {
    int n_grid = dataCenter.n_gridR;
    double rspan = dataCenter.gridModel[dataCenter.gridModel.length-1]-
                   dataCenter.gridModel[0];

    dataCenter.gridRInversion = new double[n_grid];
    for(int i=0; i<n_grid;i++)
      dataCenter.gridRInversion[i] = 
         0.5*(1.0-Math.cos(Math.PI*(double)i/(double)(n_grid-1)))
        *rspan+dataCenter.gridModel[0];

    // initialise some integration weights:
    weights = Numerics.initWeightsCheb(n_grid);
    for(int i=0; i< n_grid; i++) weights[i]*=rspan;

    // initialise a derivation matrix:
    derive = Numerics.initDDerive(n_grid);
    rspan = 1/(rspan*rspan);
    for(int i=0; i<(n_grid*n_grid); i++) derive[i]*=rspan;
  }

  ////////////////////////////////////////////////////////////
  // interpolate all rotational kernels before doing inversions:
  ////////////////////////////////////////////////////////////
  private void initRotaKernel() {
    rotaKernel = new double[n_inputKernels][];
    for(int l=0; l<n_inputKernels; l++)
      rotaKernel[l] = Numerics.cubicSplineInterpol(dataCenter.grid,
                    dataCenter.rotaKernel[intersection.get(l)],
                    dataCenter.gridRInversion);
  }

  /********************************************************
  * Initialise the RNoise array with a Gauss distribution:
  * Mean = 0.0; std. deviation= 1.0.
  ********************************************************/
  public void initNoise() {
    Random generator = new Random();
    dataCenter.RNoise = new double[intersection.size()];
    for(int l=0; l < intersection.size(); l++) {
      dataCenter.RNoise[l] = generator.nextGaussian();
    }
  }

  /********************************************************
  * Initialise various matrices used in a RLS inversion.
  ********************************************************/
  private void initRLS() {
    int n_grid = dataCenter.gridRInversion.length;

    // initialise matrix used for fitting the data:
    fitDataMatRLS = new double[n_grid*n_grid];
    for(int i=0; i<n_grid; i++){
      for(int j=i; j<n_grid; j++){
        fitDataMatRLS[i+n_grid*j]=0.0;
        for(int l=0; l<n_inputKernels; l++){
          fitDataMatRLS[i+n_grid*j]+=weights[i]*rotaKernel[l][i]
                                    *weights[j]*rotaKernel[l][j]
                                    /(dataErrors.get(l)*dataErrors.get(l));
        }
      }
    }

    // initialise the regularisation matrix:
    regularisationMatRLS = new double[n_grid*n_grid];
    for(int i=0; i<n_grid; i++){
      for(int j=i; j<n_grid; j++){
        regularisationMatRLS[i+n_grid*j]=0.0;
        for(int k=0; k<n_grid; k++){
          regularisationMatRLS[i+n_grid*j]+=(weights[k]*derive[k+n_grid*i]*derive[k+n_grid*j]);
        }
      }
    }

    // initialise the matrix used for calculating the output kernels:
    kernelMatRLS = new double[n_grid*n_inputKernels];
    for(int l=0; l<n_inputKernels; l++){
      for(int i=0; i<n_grid; i++)
        kernelMatRLS[i+n_grid*l] = weights[i]*rotaKernel[l][i]
                                   /(dataErrors.get(l)*dataErrors.get(l));
    }

  }

  /********************************************************
  * Do the inversion, using RLS
  ********************************************************/
  protected void RLSinversion() {  
    int n_grid = dataCenter.gridRInversion.length;
    int n_grid2 = n_grid*n_grid;
    int ipiv[] = new int[n_grid];
    
    double[] mat = fitDataMatRLS.clone();
    
    double Lambda = 0.0;
    for(int l=0; l<n_inputKernels; l++) Lambda+=1.0/(dataErrors.get(l)*dataErrors.get(l));
    Lambda=dataCenter.RLambda*Lambda/(double)n_inputKernels;
    
    for(int i=0; i<n_grid2; i++)
      mat[i]+=Lambda*regularisationMatRLS[i];
    double[] mat_bis = kernelMatRLS.clone();
    
    double[] work = new double[lwork*n_grid];
    Dsytrf.dsytrf("U",n_grid,mat,0,n_grid,ipiv,0,work,0,lwork*n_grid,info);
    Dsytrs.dsytrs("U",n_grid,n_inputKernels,mat,0,n_grid,ipiv,0,mat_bis,0,n_grid,info);

    // find the inverted function:
    f_invertedRLS = new double[n_grid];
    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.RNoiseSelected) {
      for(int i=0; i<n_inputKernels; i++) {
        dataPlusNoise[i] = dataValues.get(i)+dataCenter.RNoiseLevel*dataCenter.RNoise[i];
      }
    } else {
      for(int i=0; i<n_inputKernels; i++) {
        dataPlusNoise[i] = dataValues.get(i);
      }
    }
    Dgemv.dgemv("N",n_grid,n_inputKernels,1.0,mat_bis,0,n_grid,
                dataPlusNoise,0,1,0.0,f_invertedRLS,0,1);

    // find the error bars:
    RLS_errorBars = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      RLS_errorBars[i] = 0.0;
      for(int l=0; l<n_inputKernels; l++) {
        RLS_errorBars[i]+=Math.pow(mat_bis[i+n_grid*l]*dataErrors.get(l),2.0);
      }
      RLS_errorBars[i] = Math.sqrt(RLS_errorBars[i]);
    }
    
    // Averaging kernels obtained from inversion
    // 1st coordinate: grid position
    // 2nd coordinate: target position
    averagingKernelsRLS = new double[n_grid*n_grid];
    for(int i=0; i<n_grid ;i++){
      for(int j=0; j<n_grid;j++){
        averagingKernelsRLS[i+n_grid*j]=0.0;
        for(int k=0; k<n_inputKernels;k++){
          averagingKernelsRLS[i+n_grid*j]+=(rotaKernel[k][i]*mat_bis[j+n_grid*k]);
        }
      }
    }
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion.
  ********************************************************/
  private void initSOLA() {
    int n_grid = dataCenter.gridRInversion.length;

    // initialise regularisation "matrix":
    double mu = 0.0;
    for(int l=0; l<n_inputKernels; l++) mu+=dataErrors.get(l)*dataErrors.get(l);
    mu = n_inputKernels/mu;
    
    regularisationMatSOLA = new double[n_inputKernels];
    for(int l=0; l<n_inputKernels; l++){
      regularisationMatSOLA[l]=mu*dataErrors.get(l)*dataErrors.get(l);
    }

    // initialise matrix used for fitting the data:
    fitDataMatSOLA = new double[(n_inputKernels+1)*(n_inputKernels+1)];
    for(int l=0; l<n_inputKernels; l++){
      for(int ll=l; ll<n_inputKernels; ll++){
        fitDataMatSOLA[l+(n_inputKernels+1)*ll]=0.0;
        for(int i=0; i<n_grid; i++){
          fitDataMatSOLA[l+(n_inputKernels+1)*ll]+=weights[i]
               *rotaKernel[l][i]*rotaKernel[ll][i];
        }
      }
    }
    
    double sum;
    for(int l=0; l<n_inputKernels; l++){
      sum = 0.0;
      for(int i=0; i<n_grid; i++) sum+=(weights[i]*rotaKernel[l][i]);
      fitDataMatSOLA[l+(n_inputKernels+1)*n_inputKernels] = sum*0.5;
    }
    fitDataMatSOLA[n_inputKernels+(n_inputKernels+1)*n_inputKernels]=0.0;
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion.
  ********************************************************/
  protected void initDeltaSOLA() {
    int n_grid = dataCenter.gridRInversion.length;

    // obtain the sound velocity:
    double[] c2 = Numerics.cubicSplineInterpol(dataCenter.gridModel,dataCenter.c2,
                                               dataCenter.gridRInversion);

    // this is to avoid division by zero later on:
    double max_c2 = 0.0;
    for(int i=0; i<c2.length; i++) if (max_c2 < c2[i]) max_c2 = c2[i];
    for(int i=0; i<c2.length; i++) if ((c2[i]/max_c2) < 1e-6) c2[i] = max_c2*1e-3;

    // initialise the matrix with the target kernels:
    double Norm;
    double FunctionValue;
    targetKernelMatSOLA = new double[(n_inputKernels+1)*n_grid];
    for(int i=0; i<n_grid; i++) {
      Norm = 0.0;
      for(int j=0; j<n_grid; j++) {
        FunctionValue = Math.exp(-Math.pow((dataCenter.gridRInversion[j]-dataCenter.gridRInversion[i]),2.0)/
                                (c2[i]*dataCenter.RDelta*dataCenter.RDelta))*weights[j];
        Norm+=FunctionValue;
        for(int l=0; l<n_inputKernels; l++)
          targetKernelMatSOLA[l+(n_inputKernels+1)*i]+=(FunctionValue*rotaKernel[l][j]);
      }
      Norm = 1.0/Norm; 
      for(int l=0; l<n_inputKernels; l++)
        targetKernelMatSOLA[l+(n_inputKernels+1)*i]*=Norm;
      targetKernelMatSOLA[n_inputKernels+(n_inputKernels+1)*i]=0.5;
    }
  }

  /********************************************************
  * Do the inversion, using SOLA
  ********************************************************/
  protected void SOLAinversion() {
    int n_grid = dataCenter.gridRInversion.length;
    int ipiv[] = new int[n_inputKernels+1];

    double[] mat = fitDataMatSOLA.clone();
    double tanTheta = Math.tan(dataCenter.RTheta);
    for(int l=0; l<n_inputKernels; l++)
      mat[(n_inputKernels+2)*l]+=tanTheta*regularisationMatSOLA[l];

    double[] mat_bis = targetKernelMatSOLA.clone();

    double[] work = new double[lwork*(n_inputKernels+1)];
    Dsytrf.dsytrf("U",n_inputKernels+1,mat,0,n_inputKernels+1,ipiv,0,work,0,lwork*(n_inputKernels+1),info);
    Dsytrs.dsytrs("U",n_inputKernels+1,n_grid,mat,0,n_inputKernels+1,ipiv,0,mat_bis,0,n_inputKernels+1,info);

    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.RNoiseSelected) {
      for(int i=0; i<n_inputKernels; i++) {
        dataPlusNoise[i] = dataValues.get(i)+dataCenter.RNoiseLevel*dataCenter.RNoise[i];
      }
    } else {
      for(int i=0; i<n_inputKernels; i++) {
        dataPlusNoise[i] = dataValues.get(i);
      }
    }
    
    // find the inteverted function:
    f_invertedSOLA = new double[n_grid];
    Dgemv.dgemv("T",n_inputKernels,n_grid,1.0,mat_bis,0,n_inputKernels+1,
                dataPlusNoise,0,1,0.0,f_invertedSOLA,0,1);

    // find the error bars:
    SOLA_errorBars = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      SOLA_errorBars[i] = 0.0;
      for(int l=0; l<n_inputKernels; l++) {
        SOLA_errorBars[i]+=Math.pow(mat_bis[l+(n_inputKernels+1)*i]*dataErrors.get(l),2.0);
      }
      SOLA_errorBars[i]=Math.sqrt(SOLA_errorBars[i]);
    }
    
    // Averaging kernels obtained from the inversion:
    // 1st coordinate: grid position
    // 2nd coordinate: target position
    averagingKernelsSOLA = new double[n_grid*n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int j=0; j<n_grid; j++) {
        averagingKernelsSOLA[i+n_grid*j]=0;
        for(int k=0; k<n_inputKernels; k++) {
          averagingKernelsSOLA[i+n_grid*j]+=(rotaKernel[k][i]*mat_bis[k+(n_inputKernels+1)*j]);
        }
      }
    }
  }
}
