/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008
*
****************************************************************************/
package inversion_applet;

import inversion_applet.DataCenter;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.util.StringTokenizer;

public class RotationalSplittingsDataModel extends AbstractTableModel {

  // the data with the frequency shifts is in dataCenter;
  private DataCenter dataCenter;
 
  // These contain the parent components, which intervenes
  // when using a JPaneOption object.
  private JComponent parentComponent;

  
  // the names of the columns:
  private String[] columns = {"n","\u2113","Rotational Splitting",
                              "Error(R.S.)"};

  // the classes of the columns:
  private Class[] columnClasses = new Class[] 
    {Integer.class, Integer.class, Double.class, Double.class};
 
  /////////////////////////////////////////////////////////
  // a constructor:
  /////////////////////////////////////////////////////////
  protected RotationalSplittingsDataModel(DataCenter dataCenter,
                               JComponent parentComponent) {
    this.dataCenter = dataCenter;
    this.parentComponent = parentComponent;
  }
  
  /////////////////////////////////////////////////////////
  // clear all the data:
  /////////////////////////////////////////////////////////
  protected void clear() {
    dataCenter.nRS.clear();
    dataCenter.lRS.clear();
    dataCenter.RS.clear();
    dataCenter.ErrorRS.clear();
    dataCenter.indexRS.clear();
    dataCenter.rotationInversionUpToDate = false;
    dataCenter.RSloaded = false;
  }

  /////////////////////////////////////////////////////////
  // load a file:
  /////////////////////////////////////////////////////////
  protected void loadFile(String filename){
    BufferedReader f;
    String line;
    Boolean skipLine = false;
    int n = dataCenter.nRS.size();
    int nRS, lRS;
    double RS, ErrorRS;
    try {
      f = new BufferedReader(new FileReader(filename));
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
                    parentComponent,  // the parent component
                    "Unable to open file \""+filename+"\"",
                    "File not found",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }
    try {
      while (true) {
        line = f.readLine();     // read a line of text
        if (line == null) break; // test for the end of the buffer
        line = trim(line);       // remove comments
        StringTokenizer t = new StringTokenizer(line," ");
        // only read lines with EXACTLY 4 numbers
        if (t.countTokens() == 4) {
          nRS = Integer.parseInt(t.nextToken());
          lRS = Integer.parseInt(t.nextToken());
          RS = Double.parseDouble(t.nextToken());
          ErrorRS = Double.parseDouble(t.nextToken());
          if (lRS > 0) {
            dataCenter.nRS.add(nRS);
            dataCenter.lRS.add(lRS);
            dataCenter.RS.add(RS);
            dataCenter.ErrorRS.add(ErrorRS);
            dataCenter.indexRS.add(new Integer(n));
            n++;
          } else {
            skipLine = true;
          }
        } else {
          if (t.countTokens() != 0) skipLine = true;
        }
      }
      f.close();
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
                    parentComponent,  // the parent component
                    "Format not recognised",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }
    if (skipLine) {
      JOptionPane.showMessageDialog(
                    parentComponent,  // the parent component
                    "Some lines with either the wrong\n"+
                    "number of entries or with \u2113 < 1\n"+
                    "were skipped.",
                    "Warning",
                    JOptionPane.WARNING_MESSAGE
                  );
    }
    dataCenter.rotationInversionUpToDate = false;
    if (n > 0) dataCenter.RSloaded = true;
  }

  /////////////////////////////////////////////////////////
  // This remove comments (i.e. anything following a "#")
  // from a string and is used by the loadFile method:
  /////////////////////////////////////////////////////////
  private static String trim(String str) {
    int len = str.length();
    int i = 0;
    while ((i < len)&&(str.charAt(i) != '#')) i++;
    return str.substring(0,i);
  }
      
  /////////////////////////////////////////////////////////
  // write a file:
  /////////////////////////////////////////////////////////
  protected void writeFile(String filename) {
    File f = new File(filename);
    PrintWriter out;
    try {
      out = new PrintWriter(new FileWriter(f));
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
        parentComponent,
        "Unable to write to file \""+filename+"\"",
        "Failure",
        JOptionPane.WARNING_MESSAGE
      );
      return;
    }
    out.println("# n l Rotational_splittings Error");
    for(int i = 0; i < dataCenter.nRS.size(); i++) {
      out.printf("%3d %3d %19.12e %19.12e%n",
                 dataCenter.nRS.get(dataCenter.indexRS.get(i).intValue()), 
                 dataCenter.lRS.get(dataCenter.indexRS.get(i).intValue()),
                 dataCenter.RS.get(dataCenter.indexRS.get(i).intValue()),
                 dataCenter.ErrorRS.get(dataCenter.indexRS.get(i).intValue()));
    }
    out.close();
  }


  /////////////////////////////////////////////////////////
  // the number of columns:
  /////////////////////////////////////////////////////////
  public int getColumnCount() {return 4;}
  
  /////////////////////////////////////////////////////////
  // the number of rows:
  /////////////////////////////////////////////////////////
  public int getRowCount() {return dataCenter.nRS.size();}
 
  /////////////////////////////////////////////////////////
  // return the value of a cell:
  /////////////////////////////////////////////////////////
  public Object getValueAt(int row, int col) {
    if (col == 0) return dataCenter.nRS.get(dataCenter.indexRS.get(row).intValue());
    if (col == 1) return dataCenter.lRS.get(dataCenter.indexRS.get(row).intValue());
    if (col == 2) return dataCenter.RS.get(dataCenter.indexRS.get(row).intValue());
    return dataCenter.ErrorRS.get(dataCenter.indexRS.get(row).intValue());
  }

  /////////////////////////////////////////////////////////
  // return the name of a column:
  /////////////////////////////////////////////////////////
  public String getColumnName(int column) {return columns[column];}
  
  /////////////////////////////////////////////////////////
  // return the class of a column:
  /////////////////////////////////////////////////////////
  public Class getColumnClass(int c) {return columnClasses[c];}
  
  /////////////////////////////////////////////////////////
  // this makes all of the cells editable:
  /////////////////////////////////////////////////////////
  public boolean isCellEditable(int row, int col) {return true;}

  /////////////////////////////////////////////////////////
  /** set the value of a particular cell: */
  /////////////////////////////////////////////////////////
  public void setValueAt(Object obj, int row, int col) {
    if (col == 0) dataCenter.nRS.set(dataCenter.indexRS.get(row).intValue(),(Integer)obj);
    if (col == 1) dataCenter.lRS.set(dataCenter.indexRS.get(row).intValue(),(Integer)obj);
    if (col == 2) dataCenter.RS.set(dataCenter.indexRS.get(row).intValue(),(Double)obj);
    if (col == 3) dataCenter.ErrorRS.set(dataCenter.indexRS.get(row).intValue(),(Double)obj);
    dataCenter.rotationInversionUpToDate = false;
  }
  
  /////////////////////////////////////////////////////////
  // add a blank row:
  /////////////////////////////////////////////////////////
  public void addBlankRow() {
    dataCenter.indexRS.add(dataCenter.nRS.size());
    dataCenter.nRS.add(null);
    dataCenter.lRS.add(null);
    dataCenter.RS.add(null);
    dataCenter.ErrorRS.add(null);
    dataCenter.rotationInversionUpToDate = false;
    dataCenter.RSloaded = true;
  }

  /////////////////////////////////////////////////////////
  /** remove a row: */
  /////////////////////////////////////////////////////////
  public void removeRow(int i) {
    if (i >= dataCenter.nRS.size()) return;
    int ii = dataCenter.indexRS.get(i).intValue();
    int jj;
    dataCenter.nRS.remove(ii);
    dataCenter.lRS.remove(ii);
    dataCenter.RS.remove(ii);
    dataCenter.ErrorRS.remove(ii);
    dataCenter.indexRS.remove(i);
    for(int j=0; j<dataCenter.indexRS.size(); j++) {
      jj = dataCenter.indexRS.get(j).intValue();
      if (jj > ii) dataCenter.indexRS.set(j,jj-1);
    }
    dataCenter.rotationInversionUpToDate = false;
  }

  /////////////////////////////////////////////////////////
  /** sort the table: */
  /////////////////////////////////////////////////////////
  public void sort() {
    int[] new_indexRS = dataCenter.sortRS();
    for(int i=0; i<new_indexRS.length; i++)
      dataCenter.indexRS.set(i,new_indexRS[i]);
  }

  /////////////////////////////////////////////////////////
  // Calculate theoretical rotational splittings
  /////////////////////////////////////////////////////////
  public void calculateRSfromRotaTarget(JComponent c){
    // I'll assume that if targetRota != null, than gridTargetRota != null
    if (!dataCenter.targetRotaLoaded) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "No target \u03A9 function loaded.  Unable to proceed.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    if (!dataCenter.kernelsLoaded) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "No kernels loaded.  Unable to proceed.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    // check to see if the span covered by the target grid is
    // large enough:
    if ((dataCenter.gridTargetRota[0]>dataCenter.grid[0])||
        (dataCenter.gridTargetRota[dataCenter.gridTargetRota.length-1]<
         dataCenter.grid[dataCenter.grid.length-1])) {
      JOptionPane.showMessageDialog(
                    c,  // the parent component
                    "Unable to calculate rotational splittings from\n"+
                    "target \u03A9 function: the grid does not span\n"+
                    "a large enough domain.  Please enter a new target\n"+
                    "function defined on a larger domain.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }
         
    // interpolate rotaTarget onto the same grid as the kernels
    double[] rotaTarget = Numerics.cubicSplineInterpol(
                          dataCenter.gridTargetRota,
                          dataCenter.targetRota,
                          dataCenter.grid);
    double[] weights = Numerics.findIntegrationWeights(dataCenter.grid);
    for (int j=0; j<rotaTarget.length; j++) weights[j]*=rotaTarget[j];
    clear();
    double sum;
    int l;
    int ii = 0;
    for(int i=0; i<dataCenter.nK.length; i++) {
      l = dataCenter.lK[i];
      if (l != 0) {
        // calculate rotational splitting for m = 1 (prograde convention):
        sum = 0.0;
        for(int j=0; j<rotaTarget.length; j++)
          sum+=weights[j]*dataCenter.rotaKernel[i][j];
        if (!Double.isNaN(sum)) {
          dataCenter.nRS.add(dataCenter.nK[i]);
          dataCenter.lRS.add(l);
          dataCenter.ErrorRS.add(1.0);
          dataCenter.RS.add(sum);
          dataCenter.indexRS.add(ii);
          ii++;
        }
      }
    }
    if (ii > 0) dataCenter.RSloaded = true;
    dataCenter.rotationInversionUpToDate = false;
  }
}
