/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008
*
****************************************************************************/
package inversion_applet;

import inversion_applet.DataCenter;
import inversion_applet.Numerics;
import org.netlib.lapack.Dsytrf;
import org.netlib.lapack.Dsytrs;
import org.netlib.lapack.Ilaenv;
import java.util.Random;
import java.util.ArrayList;
import java.util.Vector;
import javax.swing.JOptionPane;
import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JTable;

/*************************************************************************
**
**    StructuralInversion
**                                              Version 1.0   Mai 2008
**
*************************************************************************/

public class StructuralInversion { 

  private DataCenter dataCenter;
  protected boolean canInvert;   // after initialisation, if the provided information
                                 // is insufficient or erroneous, it is not possible
                                 // to do an inversion, in which case, canInvert = false
                                 // Otherwise, canInvert = true
  protected int n_inputKernels;  // the number of kernels used in the inversion

  // various inversion results
  public double[] averagingKernels1RLS;  // Kernels obtained after a RLS inversion
  public double[] averagingKernels1SOLA; // Kernels obtained after a SOLA inversion
  public double[] averagingKernels2RLS;  // Kernels obtained after a RLS inversion
  public double[] averagingKernels2SOLA; // Kernels obtained after a SOLA inversion
  public double[] crossTermKernels1RLS;  // Kernels obtained after a RLS inversion
  public double[] crossTermKernels1SOLA; // Kernels obtained after a SOLA inversion
  public double[] crossTermKernels2RLS;  // Kernels obtained after a RLS inversion
  public double[] crossTermKernels2SOLA; // Kernels obtained after a SOLA inversion
                                         // 1st coordinate: grid position
                                         // 2nd coordinate: target position
  public double[] f1_invertedRLS; // function given by the RLS inversion (f_inverted[n_grid])
  public double[] f2_invertedRLS; // function given by the RLS inversion (f_inverted[n_grid])
  public double[] RLS_errorBars1; // this gives the uncertainty in the inversion, based on
                                  // the error bars of the frequency shifts
  public double[] RLS_errorBars2; // this gives the uncertainty in the inversion, based on
                                  // the error bars of the frequency shifts
  public double[] f1_invertedSOLA;// function given by the SOLA inversion (f_inverted[n_grid])
  public double[] f2_invertedSOLA;// function given by the SOLA inversion (f_inverted[n_grid])
  public double[] SOLA_errorBars1;// this gives the uncertainty in the inversion, based on
                                  // the error bars of the frequency shifts
  public double[] SOLA_errorBars2;// this gives the uncertainty in the inversion, based on
                                  // the error bars of the frequency shifts
  
  // various intermediate matrices used in the inverstions:
  private double[] fitDataMatRLS;        // matrix used for fitting the data in a RLS inversion
  private double[] regularisationMatRLS; // matrix used for regularisation in a RLS inversion
  private double[] kernelMatRLS;         // matrix used for calculating output kernels
  private double[] kernelMat1SOLA;       // matrix used for fitting the data in a SOLA inversion
  private double[] kernelMat2SOLA;       // matrix used for fitting the data in a SOLA inversion
  private double[] supplMat1SOLA;        // matrix with term relating to supplementary conditions
  private double[] supplMat2SOLA;        // matrix with term relating to supplementary conditions
  private double[] regularisationMatSOLA;// matrix used for regularisation in a SOLA inversion
  private double[] targetKernelMat1SOLA; // matrix used for calculating the target kernels
  private double[] targetKernelMat2SOLA; // matrix used for calculating the target kernels
  private org.netlib.util.intW info = new org.netlib.util.intW(2); // necessary for lapack
  private ArrayList<Integer> intersection = new ArrayList<Integer>();
  private ArrayList<Double>  dataValues   = new ArrayList<Double>();
  private ArrayList<Double>  dataErrors   = new ArrayList<Double>();
  private double[][] struct1Kernel;
  private double[][] struct2Kernel;
  private double[][] surfPoly;
  private int lwork;  // an integer used in the matrix solvers
  private double[] weights;
  private double[] derive;
  

  /********************************************************
  *  A constructor which does the initialisation
  ********************************************************/
  public StructuralInversion(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
     
    // At the beginning, no data is loaded, so no inversion
    // can be done.  Therefore the inversion is up to date.
    dataCenter.structInversionUpToDate = true;
    this.canInvert = false;
    
    // This determines the blocksize to be used in the symmetric 
    // matrix solvers. I will just put some average matrix size
    // rather than an exact number.
    this.lwork = Ilaenv.ilaenv(1,"DSYTRF","U",200,200,-1,-1);
  }

  ////////////////////////////////////////////////////////////
  /** This calls various initialisation methods */
  ////////////////////////////////////////////////////////////
  public void init(JComponent c) {
    canInvert = true;
    dataCenter.structInversionUpToDate = true;
    n_inputKernels = 0;
    initIntersection(c);
    if (canInvert) {
      initGridInversionCheb();
      initStructKernels();
      initNoise();
      initSurfPoly();
      initRLS();
      initSOLA();
      initDelta1SOLA();
      initDelta2SOLA();
    }
  }

  ////////////////////////////////////////////////////////////
  // This finds the intersection between the set of kernels
  // and the set of frequency shift data and stores it in
  // an ArrayList called intersection
  ////////////////////////////////////////////////////////////
  private void initIntersection(JComponent c) {

    boolean unusedDF = false;
    boolean dataWithZeroError = false;
    boolean incompleteDF = false;
    boolean incorrectDF = false;
    boolean duplicateDF = false;
    boolean incorrectK = false;
    boolean duplicateK = false;

    Vector<Vector<Object>> unusedData = new Vector<Vector<Object>>();

    // clear previous data:
    intersection.clear();
    dataValues.clear();
    dataErrors.clear();

    // make sure frequency shift data is loaded:
    if (!dataCenter.DFloaded) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No frequency shift data.  Unable to proceed",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      canInvert = false;
      n_inputKernels = 0;
      return;
    }

    
    // make sure kernel data is loaded:
    if (!dataCenter.kernelsLoaded) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No kernels loaded.  Unable to proceed",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      canInvert = false;
      n_inputKernels = 0;
      return;
    }

    // set up sort indices
    int nnDF = dataCenter.nDF.size();
    int nnK  = dataCenter.nK.length;
    int[] indexDF = dataCenter.sortDF();
    int[] indexK  = dataCenter.sortKernels();

    // find intersection on sorted arrays
    int iK = 0;
    int iDF = 0;
    int cmp;
    while ((iDF < nnDF)&&(iK < nnK)) {
      
      // test the frequency data to make sure that
      // none of the fields are set to null:
      if ((dataCenter.nDF.get(indexDF[iDF]) == null)||
          (dataCenter.lDF.get(indexDF[iDF]) == null)||
          (dataCenter.DF.get(indexDF[iDF]) == null)||
          (dataCenter.ErrorDF.get(indexDF[iDF]) == null)) {
        incompleteDF = true;
        iDF++;
        continue;
      }
      
      // test to make sure that the error bars are
      // not zero:
      if (dataCenter.ErrorDF.get(indexDF[iDF]) <= 0) {
        dataWithZeroError = true;
        iDF++;
        continue;
      }

      // make sure that n and l have possible values
      // (i.e. l>0 for n>0 and l>1 for n<=0) for
      // frequency shift data:
      if (dataCenter.nDF.get(indexDF[iDF])<=0) {
        if (dataCenter.lDF.get(indexDF[iDF])<1) {
          incorrectDF = true;
          iDF++;
          continue;
        }
      } else {
        if (dataCenter.lDF.get(indexDF[iDF])<0) {
          incorrectDF = true;
          iDF++;
          continue;
        }
      }

      // check for duplicate frequency shift data:
      if ((iDF>0)&&equalsDF(indexDF[iDF],indexDF[iDF-1])) {
        duplicateDF = true;
        iDF++;
        continue;
      }
     
      // make sure that n and l have possible values
      // (i.e. l>0 for n>0 and l>1 for n<=0) for
      // kernel data:
      if (dataCenter.nK[indexK[iK]]<=0) {
        if (dataCenter.lK[indexK[iK]]<1) {
          incorrectK = true;
          iK++;
          continue;
        }
      } else {
        if (dataCenter.lK[indexK[iK]]<0) {
          incorrectK = true;
          iK++;
          continue;
        }
      }

      // check for duplicate kernel data:
      if ((iK>0)&&
          (dataCenter.nK[indexK[iK-1]]==dataCenter.nK[indexK[iK]]) &&
          (dataCenter.lK[indexK[iK-1]]==dataCenter.lK[indexK[iK]])) {
        duplicateK = true;
        iK++;
        continue;
      }
     
      // do the comparison:
      cmp = compareDFKernel(indexDF[iDF],indexK[iK]);
      
      // if K[iK] is smaller, increase iK:
      if (cmp > 0) {
        iK++;
        continue;
      }
      
      // If DF[iDF] is smaller, increase iDF:
      if (cmp < 0) {
        Vector<Object> temp = new Vector<Object>();
        temp.add(new Integer(dataCenter.nDF.get(indexDF[iDF])));
        temp.add(new Integer(dataCenter.lDF.get(indexDF[iDF])));
        temp.add(new Double(dataCenter.DF.get(indexDF[iDF])));
        temp.add(new Double(dataCenter.ErrorDF.get(indexDF[iDF])));
        unusedData.add(temp);
        unusedDF = true;
        iDF++;
        continue;
      }
      
      // DF[iDF] == K[iK]: add this to intersection:
      intersection.add(indexK[iK]);
      dataValues.add(dataCenter.DF.get(indexDF[iDF]));
      dataErrors.add(dataCenter.ErrorDF.get(indexDF[iDF]));
      iDF++;
      iK++;
    }

    if (intersection.size() == 0) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "No common modes (n,\u2113) between the kernel data\n"+
        "set and the frequency shift data.  Unable to proceed.",
        "Error",
        JOptionPane.ERROR_MESSAGE
      );
      canInvert = false;
      n_inputKernels = 0;
      return;
    }

    if (incompleteDF) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the frequency shift data is\n"+
        "incomplete and was therefore not used.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
    
    if (dataWithZeroError) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the frequency shift data has\n"+
        "a 0 error bar, which is not allowed.\n"+
        "This data has consequently been ignored.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
      
    if (duplicateDF) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "The frequency shift data contains modes with \n"+
        "duplicate (n,\u2113) which has been ignored.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    if (incorrectDF) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the frequency shift data seems to be erroneous:\n"+
        "  (\u2113 < 1 for n < 1) or (\u2113 < 0 for n > 0).\n"+
        "This data has consequently not been not used.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
    
    if (duplicateK) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "The kernel data contains modes with \n"+
        "duplicate (n,\u2113) which has been ignored.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }

    if (incorrectK) {
      JOptionPane.showMessageDialog(
        c, // parent component
        "Some of the kernel data seems to be erroneous:\n"+
        "  (\u2113 < 1 for n n \u2264 0) or (\u2113 < 0 for n > 0).\n"+
        "This data has consequently not been not used.",
        "Warning",
        JOptionPane.WARNING_MESSAGE
      );
    }
    
    if (unusedDF||(iDF < nnDF)) {
      String message = "Some of the frequency shift data was not used.\n"+
                       "Loading the appropriate kernels would allow\n"+
                       "use of this data.";
      String[] options = {"View skipped data", "OK"};
      int result = JOptionPane.showOptionDialog(
        c,                            // the parent that the dialog blocks
        message,                      // the dialog message array
        "Warning",                    // the title of the dialog window
        JOptionPane.DEFAULT_OPTION,   // option type
        JOptionPane.QUESTION_MESSAGE, // message type
        null,                         // optional icon, use null to use the default icon
        options,                      // options string array, will be made into buttons
        options[1]                    // option that should be made into a default button
      );
      if (result == 0) {
        for(int i=iDF; i<nnDF; i++) {
          Vector<Object> temp = new Vector<Object>();
          temp.add(new Integer(dataCenter.nDF.get(indexDF[iDF])));
          temp.add(new Integer(dataCenter.lDF.get(indexDF[iDF])));
          temp.add(new Double(dataCenter.DF.get(indexDF[iDF])));
          temp.add(new Double(dataCenter.ErrorDF.get(indexDF[iDF])));
          unusedData.add(temp);
        }

        Vector<String> columns = new Vector<String>();
        columns.add("n");
        columns.add("\u2113");
        columns.add("\u0394f/f");
        columns.add("Error(\u0394f/f)");

        JTable aTable = new JTable(unusedData,columns) {
          public boolean isCellEditable(int row, int col) {return false;}
        };
        JScrollPane scroller = new JScrollPane();
        scroller.getViewport().add(aTable);
        Object[] contents = new Object[2];
        contents[0] = "List of frequency shift data which\n"+
                      "was not included in the inversion.";
        contents[1] = scroller;
        JOptionPane.showMessageDialog(
          c, // parent component
          contents,
          "List of unused data",
          JOptionPane.INFORMATION_MESSAGE
        );
      }
    }
    n_inputKernels = intersection.size();
  }

  ////////////////////////////////////////////////////////////
  // Returns true if the two frequency shifts data have the
  // same (n,l) and false otherwise.  Returns false if
  // n or l of eithers rotational splitting is null.
  ////////////////////////////////////////////////////////////
  private boolean equalsDF(int i1, int i2) {
    if (dataCenter.nDF.get(i1) == null) return false;
    if (dataCenter.lDF.get(i1) == null) return false;
    if (dataCenter.nDF.get(i2) == null) return false;
    if (dataCenter.lDF.get(i2) == null) return false;
    return
      (dataCenter.nDF.get(i2).compareTo(dataCenter.nDF.get(i1))==0)&&
      (dataCenter.lDF.get(i2).compareTo(dataCenter.lDF.get(i1))==0);
  }

  ////////////////////////////////////////////////////////////
  // Compare (n,l) from a frequancy-shift data (indexed by iDF) and 
  // from a Kernel (indexed by iK).
  // if DF(iDF) < K(iK),  returns negative int
  // if DF(iDF) == K(iK), return 0
  // if DF(iDF) > K(iK),  return positive int
  ////////////////////////////////////////////////////////////
  private int compareDFKernel(int iDF, int iK) {
    int compareL = dataCenter.lDF.get(iDF).compareTo(Integer.valueOf(dataCenter.lK[iK]));
    if (compareL == 0) {
      return dataCenter.nDF.get(iDF).compareTo(Integer.valueOf(dataCenter.nK[iK]));
    } else {
      return compareL;
    }
  }

  ////////////////////////////////////////////////////////////
  // This initialises a grid to be used in the inversion.
  // For a given number of grid points, it is not as
  // accurate as initGridInversionCheb() below.
  ////////////////////////////////////////////////////////////
  private void initGridInversion() {
    int n_grid = dataCenter.n_gridS;

    // obtain the sound velocity:
    double[] c = new double[dataCenter.c2.length];
    for(int i=0; i<c.length; i++) c[i] = Math.sqrt(dataCenter.c2[i]);

    // this is to avoid division by zero later on:
    double max_c = 0.0;
    for(int i=0; i<c.length; i++) if (max_c < c[i]) max_c = c[i];
    for(int i=0; i<c.length; i++) if ((c[i]/max_c) < 1e-6) c[i] = max_c*1e-6;

    // find the inverse sound velocity:
    for(int i=0; i<c.length; i++) c[i] = 1.0/Math.sqrt(c[i]);

    double[] CC = Numerics.findAntiDerivative(dataCenter.gridModel,c);
    double cInt = CC[CC.length-1]-CC[0];

    dataCenter.gridSInversion = new double[n_grid];
    dataCenter.gridSInversion[0] = dataCenter.gridModel[0];
    dataCenter.gridSInversion[n_grid-1] =
       dataCenter.gridModel[dataCenter.gridModel.length-1];

    int j = 0;
    double cTarget, lambda;
    for(int i=1; i<n_grid-1; i++) {
      cTarget = cInt*(double)i/(double)(n_grid-1);  // a target value
      while(CC[j]<cTarget) j++;
      lambda = (CC[j]-cTarget)/(CC[j]-CC[j-1]);
      dataCenter.gridSInversion[i] = dataCenter.gridModel[j-1]*lambda 
                                  + dataCenter.gridModel[j]*(1.0-lambda);
      j--;
    }
    
    // initialise some integration weights:
    weights = Numerics.findIntegrationWeights(dataCenter.gridSInversion);
    // initialise a derivation matrix:
    double[][] derive_temp = Numerics.initDeriveFD(dataCenter.gridSInversion, 2, 2);
    derive = derive_temp[2];
  }

  ////////////////////////////////////////////////////////////
  // This initialises a Chebyshev collocation grid and spectral
  // integration weights to be used in the inversion.  For a
  // given number of grid points, it is more accurate than
  // initGridInversion() above.
  ////////////////////////////////////////////////////////////
  private void initGridInversionCheb() {
    int n_grid = dataCenter.n_gridS;
    double rspan = dataCenter.gridModel[dataCenter.gridModel.length-1]-
                   dataCenter.gridModel[0];

    dataCenter.gridSInversion = new double[n_grid];
    for(int i=0; i<n_grid;i++)
      dataCenter.gridSInversion[i] = 
         0.5*(1.0-Math.cos(Math.PI*(double)i/(double)(n_grid-1)))
        *rspan+dataCenter.gridModel[0];

    // initialise some integration weights:
    weights = Numerics.initWeightsCheb(n_grid);
    for(int i=0; i< n_grid; i++) weights[i]*=rspan;
    
    // initialise a derivation matrix:
    derive = Numerics.initDDerive(n_grid);
    rspan = 1/(rspan*rspan);
    for(int i=0; i<(n_grid*n_grid); i++) derive[i]*=rspan;
  }

  ////////////////////////////////////////////////////////////
  // interpolate all rotational kernels before doing inversions:
  ////////////////////////////////////////////////////////////
  private void initStructKernels() {
    struct1Kernel = new double[n_inputKernels][];
    struct2Kernel = new double[n_inputKernels][];
    for(int l=0; l<n_inputKernels; l++) {
      struct1Kernel[l] = Numerics.cubicSplineInterpol(dataCenter.grid,
                    dataCenter.struct1Kernel[intersection.get(l)],
                    dataCenter.gridSInversion);
      struct2Kernel[l] = Numerics.cubicSplineInterpol(dataCenter.grid,
                    dataCenter.struct2Kernel[intersection.get(l)],
                    dataCenter.gridSInversion);
    }
  }

  /********************************************************
  * Initialise the Noise array with a Gauss distribution:
  * Mean = 0.0; std. deviation= 1.0.
  ********************************************************/
  public void initNoise() {
    Random generator = new Random();
    dataCenter.SNoise = new double[intersection.size()];
    for(int l=0; l < intersection.size(); l++) {
      dataCenter.SNoise[l] = generator.nextGaussian();
    }
  }

  /********************************************************
  * This initialises the polynomial expansion used to
  * model the surface terms.
  ********************************************************/
  public void initSurfPoly() {
    surfPoly = new double[n_inputKernels][];
    
    for(int l=0; l<n_inputKernels; l++) {
      surfPoly[l] = Numerics.Legendre(
                      dataCenter.freqDimless[intersection.get(l)],
                      dataCenter.n_surf
                    );
      for(int n=0; n<dataCenter.n_surf; n++)
        surfPoly[l][n]/=dataCenter.Enl[intersection.get(l)];
    }
  }

  /********************************************************
  * Initialise various matrices used in a RLS inversion.
  ********************************************************/
  private void initRLS() {
    int n_grid = dataCenter.n_gridS;
    int matSize = 2*n_grid+dataCenter.n_surf;
    double[] rho = null;
    
    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      matSize++;
    if (dataCenter.constantMass==3)
      matSize+=2;
    if (dataCenter.constantMass!=0)
      rho = Numerics.cubicSplineInterpol(dataCenter.gridModel,
                     dataCenter.rho,dataCenter.gridSInversion);

    // initialise matrix used for fitting the data:
    fitDataMatRLS = new double[matSize*matSize];
    for(int i=0; i<n_grid; i++){
      for(int j=i; j<n_grid; j++){
        for(int l=0; l<n_inputKernels; l++){
          fitDataMatRLS[i+matSize*j]+=
              weights[i]*struct1Kernel[l][i]
             *weights[j]*struct1Kernel[l][j]
             /(dataErrors.get(l)*dataErrors.get(l));
          fitDataMatRLS[i+n_grid+matSize*(j+n_grid)]+=
              weights[i]*struct2Kernel[l][i]
             *weights[j]*struct2Kernel[l][j]
             /(dataErrors.get(l)*dataErrors.get(l));
        }
      }
    }

    for(int i=0; i<n_grid; i++){
      for(int j=0; j<n_grid; j++){
        for(int l=0; l<n_inputKernels; l++){
          fitDataMatRLS[i+matSize*(j+n_grid)]+=
              weights[i]*struct1Kernel[l][i]
             *weights[j]*struct2Kernel[l][j]
             /(dataErrors.get(l)*dataErrors.get(l));
        }
      }
    }

    for(int i=0; i<n_grid; i++) {
      for(int n=0; n<dataCenter.n_surf; n++) {
        for(int l=0; l<n_inputKernels; l++) {
          fitDataMatRLS[i+matSize*(n+2*n_grid)]+=
              weights[i]*struct1Kernel[l][i]
             *surfPoly[l][n]
             /(dataErrors.get(l)*dataErrors.get(l));
          fitDataMatRLS[i+n_grid+matSize*(n+2*n_grid)]+=
              weights[i]*struct2Kernel[l][i]
             *surfPoly[l][n]
             /(dataErrors.get(l)*dataErrors.get(l));
        }
      }
    }

    for(int m=0; m<dataCenter.n_surf; m++) {
      for(int n=m; n<dataCenter.n_surf; n++) {
        for(int l=0; l<n_inputKernels; l++) {
          fitDataMatRLS[2*n_grid+m+matSize*(2*n_grid+n)]+=
              surfPoly[l][m]*surfPoly[l][n]
             /(dataErrors.get(l)*dataErrors.get(l));
        }
      }
    }

    // supplementary constraints for keeping the mass constant after inversion:  
    if (dataCenter.constantMass == 1) {
      for(int i=0; i<n_grid; i++) {
        fitDataMatRLS[i+matSize*(2*n_grid+dataCenter.n_surf)]=
          dataCenter.gridSInversion[i]*dataCenter.gridSInversion[i]*rho[i]*weights[i]*0.5;
      }
    }

    if (dataCenter.constantMass == 2) {
      for(int i=0; i<n_grid; i++) {
        fitDataMatRLS[i+n_grid+matSize*(2*n_grid+dataCenter.n_surf)]=
          dataCenter.gridSInversion[i]*dataCenter.gridSInversion[i]*rho[i]*weights[i]*0.5;
      }
    }

    if (dataCenter.constantMass == 3) {
      for(int i=0; i<n_grid; i++) {
        fitDataMatRLS[i+matSize*(2*n_grid+dataCenter.n_surf)]=
          dataCenter.gridSInversion[i]*dataCenter.gridSInversion[i]*rho[i]*weights[i]*0.5;
        fitDataMatRLS[i+n_grid+matSize*(2*n_grid+dataCenter.n_surf+1)]=
          dataCenter.gridSInversion[i]*dataCenter.gridSInversion[i]*rho[i]*weights[i]*0.5;
      }
    }

    // initialise the regularisation matrix:
    regularisationMatRLS = new double[n_grid*n_grid];
    for(int i=0; i<n_grid; i++){
      for(int j=i; j<n_grid; j++){
        for(int k=0; k<n_grid; k++){
          regularisationMatRLS[i+n_grid*j]+=
              weights[k]*derive[k+n_grid*i]
                        *derive[k+n_grid*j];
        }
      }
    }

    // initialise the matrix used for calculating the output kernels:
    kernelMatRLS = new double[matSize*n_inputKernels];
    for(int l=0; l<n_inputKernels; l++){
      for(int i=0; i<n_grid; i++) {
        kernelMatRLS[i+matSize*l]=
            weights[i]*struct1Kernel[l][i]
           /(dataErrors.get(l)*dataErrors.get(l));
        kernelMatRLS[i+n_grid+matSize*l]= 
            weights[i]*struct2Kernel[l][i]
           /(dataErrors.get(l)*dataErrors.get(l));
      }
      for(int m=0; m<dataCenter.n_surf; m++) {
        kernelMatRLS[m+2*n_grid+matSize*l]=
            surfPoly[l][m]
           /(dataErrors.get(l)*dataErrors.get(l));
      }
    }
  }

  /********************************************************
  * Do the inversion, using RLS
  ********************************************************/
  protected void RLSinversion() {  
    int n_grid = dataCenter.n_gridS;
    int matSize = 2*n_grid+dataCenter.n_surf;

    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      matSize++;
    if (dataCenter.constantMass==3)
      matSize+=2;

    int ipiv[] = new int[matSize];
    
    double Lambda = 0.0;
    for(int l=0; l<n_inputKernels; l++) Lambda+=1.0/(dataErrors.get(l)*dataErrors.get(l));
    Lambda=dataCenter.SLambda*Lambda/(double)n_inputKernels;
    
    double[] mat = fitDataMatRLS.clone();
    for(int i=0; i<n_grid; i++) {
      for(int j=i; j<n_grid; j++) {
        mat[i+matSize*j]+=Lambda*regularisationMatRLS[i+n_grid*j];
        mat[i+n_grid+matSize*(j+n_grid)]+=Lambda*regularisationMatRLS[i+n_grid*j];
      }
    }

    double[] mat_bis = kernelMatRLS.clone();
    
    double[] work = new double[lwork*matSize];
    Dsytrf.dsytrf("U",matSize,mat,0,matSize,ipiv,0,work,0,lwork*matSize,info);
    Dsytrs.dsytrs("U",matSize,n_inputKernels,mat,0,matSize,ipiv,0,mat_bis,0,matSize,info);

    // create data with or without noise:
    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.SNoiseSelected) {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l)+dataCenter.SNoiseLevel*dataCenter.SNoise[l];
      }
    } else {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l);
      }
    }

    // find the inverted functions:
    f1_invertedRLS = new double[n_grid];
    f2_invertedRLS = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int l=0; l<n_inputKernels; l++) {
        f1_invertedRLS[i]+=mat_bis[i+matSize*l]*dataPlusNoise[l];
        f2_invertedRLS[i]+=mat_bis[i+n_grid+matSize*l]*dataPlusNoise[l];
      }
    }

    // find the error bars:
    RLS_errorBars1 = new double[n_grid];
    RLS_errorBars2 = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int l=0; l<n_inputKernels; l++) {
        RLS_errorBars1[i]+=Math.pow(mat_bis[i+matSize*l]*dataErrors.get(l),2.0);
        RLS_errorBars2[i]+=Math.pow(mat_bis[i+n_grid+matSize*l]*dataErrors.get(l),2.0);
      }
      RLS_errorBars1[i] = Math.sqrt(RLS_errorBars1[i]);
      RLS_errorBars2[i] = Math.sqrt(RLS_errorBars2[i]);
    }
    
    // Averaging and cross-term kernels obtained from inversion
    // 1st coordinate: grid position
    // 2nd coordinate: target position
    averagingKernels1RLS = new double[n_grid*n_grid];
    averagingKernels2RLS = new double[n_grid*n_grid];
    crossTermKernels1RLS = new double[n_grid*n_grid];
    crossTermKernels2RLS = new double[n_grid*n_grid];
    for(int i=0; i<n_grid ;i++){
      for(int j=0; j<n_grid;j++){
        for(int l=0; l<n_inputKernels;l++){
          averagingKernels1RLS[i+n_grid*j]+=(struct1Kernel[l][i]*mat_bis[j+matSize*l]);
          crossTermKernels1RLS[i+n_grid*j]+=(struct2Kernel[l][i]*mat_bis[j+matSize*l]);
          averagingKernels2RLS[i+n_grid*j]+=(struct2Kernel[l][i]*mat_bis[j+n_grid+matSize*l]);
          crossTermKernels2RLS[i+n_grid*j]+=(struct1Kernel[l][i]*mat_bis[j+n_grid+matSize*l]);
        }
      }
    }
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion.
  ********************************************************/
  private void initSOLA() {
    int n_grid = dataCenter.n_gridS;
    int matSize = n_inputKernels+dataCenter.n_surf+1;
    double[] rho = null;

    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      matSize++;
    if (dataCenter.constantMass==3)
      matSize+=2;
    if (dataCenter.constantMass!=0)
      rho = Numerics.cubicSplineInterpol(dataCenter.gridModel,
                     dataCenter.rho,dataCenter.gridSInversion);

    // initialise matrix associated with struct1 kernels:
    kernelMat1SOLA = new double[matSize*matSize];
    for(int l=0; l<n_inputKernels; l++){
      for(int ll=l; ll<n_inputKernels; ll++){
        for(int i=0; i<n_grid; i++){
          kernelMat1SOLA[l+matSize*ll]+=weights[i]
               *struct1Kernel[l][i]*struct1Kernel[ll][i];
        }
      }
    }
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==3)) {
      for(int l=0; l<n_inputKernels; l++){
        for(int i=0; i<n_grid; i++) {
          kernelMat1SOLA[l+matSize*n_inputKernels]+=weights[i]
              *struct1Kernel[l][i]
              *dataCenter.gridSInversion[i]
              *dataCenter.gridSInversion[i]
              *rho[i];
        }
      }
      for(int i=0; i<n_grid; i++) {
        kernelMat1SOLA[(1+matSize)*n_inputKernels]+=
           weights[i]*rho[i]*rho[i]
          *Math.pow(dataCenter.gridSInversion[i],4.0);
      }
    }


    // initialise matrix associated with struct2 kernels:
    kernelMat2SOLA = new double[matSize*matSize];
    for(int l=0; l<n_inputKernels; l++){
      for(int ll=l; ll<n_inputKernels; ll++){
        for(int i=0; i<n_grid; i++){
          kernelMat2SOLA[l+matSize*ll]+=weights[i]
               *struct2Kernel[l][i]*struct2Kernel[ll][i];
        }
      }
    }
    if ((dataCenter.constantMass==2)||(dataCenter.constantMass==3)) {
      int ndx = dataCenter.constantMass-2;
      for(int l=0; l<n_inputKernels; l++){
        for(int i=0; i<n_grid; i++) {
          kernelMat2SOLA[l+matSize*(n_inputKernels+ndx)]+=weights[i]
              *struct2Kernel[l][i]
              *dataCenter.gridSInversion[i]
              *dataCenter.gridSInversion[i]
              *rho[i];
        }
      }
      for(int i=0; i<n_grid; i++) {
        kernelMat2SOLA[(1+matSize)*(n_inputKernels+ndx)]+=
           weights[i]*rho[i]*rho[i]
          *Math.pow(dataCenter.gridSInversion[i],4.0);
      }
    }


    // initialise matrix associated with supplementary constraints for
    // struct1 kernels:
    double sum;
    supplMat1SOLA = new double[n_inputKernels*(1+dataCenter.n_surf)];
    for(int l=0; l<n_inputKernels; l++){
      sum = 0.0;
      for(int i=0; i<n_grid; i++) sum+=(weights[i]*struct1Kernel[l][i]);
      supplMat1SOLA[l] = sum*0.5;
      for(int n=0; n<dataCenter.n_surf; n++) {
        supplMat1SOLA[l+n_inputKernels*(1+n)] = surfPoly[l][n]*0.5;
      }
    }

    // initialise matrix associated with supplementary constraints for
    // struct2 kernels:
    supplMat2SOLA = new double[n_inputKernels*(1+dataCenter.n_surf)];
    for(int l=0; l<n_inputKernels; l++){
      sum = 0.0;
      for(int i=0; i<n_grid; i++) sum+=(weights[i]*struct2Kernel[l][i]);
      supplMat2SOLA[l] = sum*0.5;
      for(int n=0; n<dataCenter.n_surf; n++) {
        supplMat2SOLA[l+n_inputKernels*(1+n)] = surfPoly[l][n]*0.5;
      }
    }

    // initialise regularisation "matrix":
    double mu = 0.0;
    for(int l=0; l<n_inputKernels; l++) mu+=dataErrors.get(l)*dataErrors.get(l);
    mu = n_inputKernels/mu;
    
    regularisationMatSOLA = new double[n_inputKernels];
    for(int l=0; l<n_inputKernels; l++){
      regularisationMatSOLA[l]=mu*dataErrors.get(l)*dataErrors.get(l);
    }
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion.
  ********************************************************/
  protected void initDelta1SOLA() {
    int n_grid = dataCenter.n_gridS;
    int matSize = n_inputKernels+dataCenter.n_surf+1;
    int suppl = 0;
    double[] rho = null;
    
    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      suppl=1;
    if (dataCenter.constantMass==3)
      suppl=2;
    if (dataCenter.constantMass!=0)
      rho = Numerics.cubicSplineInterpol(dataCenter.gridModel,
                     dataCenter.rho,dataCenter.gridSInversion);
    matSize+=suppl;

    // obtain the sound velocity:
    double[] c2 = Numerics.cubicSplineInterpol(dataCenter.gridModel,dataCenter.c2,
                                               dataCenter.gridSInversion);

    // this is to avoid division by zero later on:
    double max_c2 = 0.0;
    for(int i=0; i<c2.length; i++) if (max_c2 < c2[i]) max_c2 = c2[i];
    for(int i=0; i<c2.length; i++) if ((c2[i]/max_c2) < 1e-6) c2[i] = max_c2*1e-6;

    // initialise the matrix with the target kernels:
    double Norm;
    double FunctionValue;
    targetKernelMat1SOLA = new double[matSize*n_grid];
    for(int i=0; i<n_grid; i++) {
      Norm = 0.0;
      for(int j=0; j<n_grid; j++) {
        FunctionValue=Math.exp(-Math.pow((dataCenter.gridSInversion[j]-dataCenter.gridSInversion[i]),2.0)/
                                         (c2[i]*dataCenter.SDelta1*dataCenter.SDelta1))*weights[j];
        Norm+=FunctionValue;
        for(int l=0; l<n_inputKernels; l++) 
          targetKernelMat1SOLA[l+matSize*i]+=(FunctionValue*struct1Kernel[l][j]);
        if ((dataCenter.constantMass==1)||(dataCenter.constantMass==3)) 
          targetKernelMat1SOLA[n_inputKernels+matSize*i]+=(FunctionValue*rho[j]
             *dataCenter.gridSInversion[j]*dataCenter.gridSInversion[j]);
      }
      Norm = 1.0/Norm;
      for(int l=0; l<n_inputKernels; l++) 
        targetKernelMat1SOLA[l+matSize*i]*=Norm;
      if ((dataCenter.constantMass==1)||(dataCenter.constantMass==3)) 
        targetKernelMat1SOLA[n_inputKernels+matSize*i]*=Norm;
      targetKernelMat1SOLA[n_inputKernels+suppl+matSize*i]=0.5;
    }
  }

  /********************************************************
  * Initialise various matrices used in a SOLA inversion.
  ********************************************************/
  protected void initDelta2SOLA() {
    int n_grid = dataCenter.n_gridS;
    int matSize = n_inputKernels+dataCenter.n_surf+1;
    int suppl = 0, ndx = 0;
    double[] rho = null;
    
    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      suppl=1;
    if (dataCenter.constantMass==3)
      suppl=2;
    if (dataCenter.constantMass!=0)
      rho = Numerics.cubicSplineInterpol(dataCenter.gridModel,
                     dataCenter.rho,dataCenter.gridSInversion);
    if (dataCenter.constantMass>1) ndx = dataCenter.constantMass-2;
    matSize+=suppl;

    // obtain the sound velocity:
    double[] c2 = Numerics.cubicSplineInterpol(dataCenter.grid,dataCenter.c2,
                                               dataCenter.gridSInversion);

    // this is to avoid division by zero later on:
    double max_c2 = 0.0;
    for(int i=0; i<c2.length; i++) if (max_c2 < c2[i]) max_c2 = c2[i];
    for(int i=0; i<c2.length; i++) if ((c2[i]/max_c2) < 1e-6) c2[i] = max_c2*1e-6;

    // initialise the matrix with the target kernels:
    double Norm;
    double FunctionValue;
    targetKernelMat2SOLA = new double[matSize*n_grid];
    for(int i=0; i<n_grid; i++) {
      Norm = 0.0;
      for(int j=0; j<n_grid; j++) {
        FunctionValue=Math.exp(-Math.pow((dataCenter.gridSInversion[j]-dataCenter.gridSInversion[i]),2.0)/
                                         (c2[i]*dataCenter.SDelta2*dataCenter.SDelta2))*weights[j];
        Norm+=FunctionValue;
        for(int l=0; l<n_inputKernels; l++) 
          targetKernelMat2SOLA[l+matSize*i]+=(FunctionValue*struct2Kernel[l][j]);
        if ((dataCenter.constantMass==2)||(dataCenter.constantMass==3)) 
          targetKernelMat2SOLA[n_inputKernels+ndx+matSize*i]+=(FunctionValue*rho[j]
             *dataCenter.gridSInversion[j]*dataCenter.gridSInversion[j]);
      }
      Norm = 1.0/Norm;
      for(int l=0; l<n_inputKernels; l++) 
        targetKernelMat2SOLA[l+matSize*i]*=Norm;
      if ((dataCenter.constantMass==2)||(dataCenter.constantMass==3)) 
        targetKernelMat2SOLA[n_inputKernels+ndx+matSize*i]*=Norm;
      targetKernelMat2SOLA[n_inputKernels+suppl+matSize*i]=0.5;
    }

  }

  /********************************************************
  * Do the inversion, using SOLA
  ********************************************************/
  protected void SOLAinversion1() {
    int n_grid = dataCenter.n_gridS;
    int matSize = n_inputKernels+dataCenter.n_surf+1;
    int suppl = 0;
    
    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      suppl=1;
    if (dataCenter.constantMass==3)
      suppl=2;
    matSize+=suppl;

    int matSize2 = matSize*matSize;
    int ipiv[] = new int[matSize];

    double[] mat = kernelMat1SOLA.clone();
    double tanTheta = Math.tan(dataCenter.STheta1);
    for(int l=0; l<n_inputKernels; l++) 
      mat[l+matSize*l]+=tanTheta*regularisationMatSOLA[l];
    for(int l=0; l<n_inputKernels; l++) {
      for(int n=0; n<(1+dataCenter.n_surf); n++)
        mat[l+matSize*(n+suppl+n_inputKernels)]+=supplMat1SOLA[l+n_inputKernels*n];
    }
    for(int i=0; i<matSize2; i++) 
      mat[i]+=dataCenter.Beta1*kernelMat2SOLA[i];

    double[] mat_bis = targetKernelMat1SOLA.clone();
    double[] work = new double[lwork*matSize];
    Dsytrf.dsytrf("U",matSize,mat,0,matSize,ipiv,0,work,0,lwork*matSize,info);
    Dsytrs.dsytrs("U",matSize,n_grid,mat,0,matSize,ipiv,0,mat_bis,0,matSize,info);

    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.SNoiseSelected) {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l)+dataCenter.SNoiseLevel*dataCenter.SNoise[l];
      }
    } else {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l);
      }
    }
    
    // find the inverted function:
    f1_invertedSOLA = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int l=0; l<n_inputKernels; l++) {
        f1_invertedSOLA[i]+=dataPlusNoise[l]*mat_bis[l+matSize*i];
      }
    }

    // find the error bars:
    SOLA_errorBars1 = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int l=0; l<n_inputKernels; l++) {
        SOLA_errorBars1[i]+=Math.pow(mat_bis[l+matSize*i]*dataErrors.get(l),2.0);
      }
      SOLA_errorBars1[i]=Math.sqrt(SOLA_errorBars1[i]);
    }
    
    // Averaging and cross-term kernels obtained from the inversion:
    // 1st coordinate: grid position
    // 2nd coordinate: target position
    averagingKernels1SOLA = new double[n_grid*n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int j=0; j<n_grid; j++) {
        for(int l=0; l<n_inputKernels; l++) {
          averagingKernels1SOLA[i+n_grid*j]+=(struct1Kernel[l][i]*mat_bis[l+matSize*j]);
        }
      }
    }
    crossTermKernels1SOLA = new double[n_grid*n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int j=0; j<n_grid; j++) {
        for(int l=0; l<n_inputKernels; l++) {
          crossTermKernels1SOLA[i+n_grid*j]+=(struct2Kernel[l][i]*mat_bis[l+matSize*j]);
        }
      }
    }
  }

  /********************************************************
  * Do the inversion, using SOLA
  ********************************************************/
  protected void SOLAinversion2() {
    int n_grid = dataCenter.n_gridS;
    int matSize = n_inputKernels+dataCenter.n_surf+1;
    int suppl = 0;
    
    // this is for supplementary constraints
    if ((dataCenter.constantMass==1)||(dataCenter.constantMass==2))
      suppl=1;
    if (dataCenter.constantMass==3)
      suppl=2;
    matSize+=suppl;

    int matSize2 = matSize*matSize;
    int ipiv[] = new int[matSize];

    double[] mat = kernelMat2SOLA.clone();
    double tanTheta = Math.tan(dataCenter.STheta2);
    for(int l=0; l<n_inputKernels; l++) 
      mat[l+matSize*l]+=tanTheta*regularisationMatSOLA[l];
    for(int l=0; l<n_inputKernels; l++) {
      for(int n=0; n<(1+dataCenter.n_surf); n++)
        mat[l+matSize*(n+suppl+n_inputKernels)]+=supplMat2SOLA[l+n_inputKernels*n];
    }
    for(int i=0; i<matSize2; i++) 
      mat[i]+=dataCenter.Beta2*kernelMat1SOLA[i];

    double[] mat_bis = targetKernelMat2SOLA.clone();
    double[] work = new double[lwork*matSize];
    Dsytrf.dsytrf("U",matSize,mat,0,matSize,ipiv,0,work,0,lwork*matSize,info);
    Dsytrs.dsytrs("U",matSize,n_grid,mat,0,matSize,ipiv,0,mat_bis,0,matSize,info);

    double[] dataPlusNoise = new double[n_inputKernels];
    if (dataCenter.SNoiseSelected) {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l)+dataCenter.SNoiseLevel*dataCenter.SNoise[l];
      }
    } else {
      for(int l=0; l<n_inputKernels; l++) {
        dataPlusNoise[l] = dataValues.get(l);
      }
    }
    
    // find the inteverted function:
    f2_invertedSOLA = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int l=0; l<n_inputKernels; l++) {
        f2_invertedSOLA[i]+=dataPlusNoise[l]*mat_bis[l+matSize*i];
      }
    }

    // find the error bars:
    SOLA_errorBars2 = new double[n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int l=0; l<n_inputKernels; l++) {
        SOLA_errorBars2[i]+=Math.pow(mat_bis[l+matSize*i]*dataErrors.get(l),2.0);
      }
      SOLA_errorBars2[i]=Math.sqrt(SOLA_errorBars2[i]);
    }
    
    // Averaging and cross-term kernels obtained from the inversion:
    // 1st coordinate: grid position
    // 2nd coordinate: target position
    averagingKernels2SOLA = new double[n_grid*n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int j=0; j<n_grid; j++) {
        for(int l=0; l<n_inputKernels; l++) {
          averagingKernels2SOLA[i+n_grid*j]+=(struct2Kernel[l][i]*mat_bis[l+matSize*j]);
        }
      }
    }
    crossTermKernels2SOLA = new double[n_grid*n_grid];
    for(int i=0; i<n_grid; i++) {
      for(int j=0; j<n_grid; j++) {
        for(int l=0; l<n_inputKernels; l++) {
          crossTermKernels2SOLA[i+n_grid*j]+=(struct1Kernel[l][i]*mat_bis[l+matSize*j]);
        }
      }
    }
  }
}
