/*--------------------------------------------------------------------------
*    This file is part of InversionKit.
*
*    InversionKit is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    InversionKit is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with InversionKit.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************
*
*    Copyright (c) Daniel Reese, Sergei Zharkov, 2008
*
****************************************************************************/
package inversion_applet;

import graph.IPlot;
import graph.Plot;
import inversion_applet.Utilities;
import inversion_applet.DataCenter;
import inversion_applet.StructuralInversion;
import inversion_applet.FrequencyShiftsPage;
import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import javax.swing.JSeparator;
import javax.swing.JCheckBox;
import javax.swing.JSlider;
import javax.swing.JComponent;
import javax.swing.JComboBox;
import javax.swing.JRadioButton;
import javax.swing.ButtonGroup;
import javax.swing.SwingConstants;
import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.awt.Color;
import java.awt.GridLayout;
import java.awt.BorderLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import java.util.StringTokenizer;
import java.util.ArrayList;

public class StructuralInversionPage extends JPanel {

  private DataCenter dataCenter;
  private IPlot graphInvertedFunction;
  private IPlot graphAveragingKernel;
  private IPlot graphCrossTermKernel;
  private JPanel graphContainer;
  private StructuralInversion my_inversion;
  private JPanel SliderContainer;

  private JLabel numberOfKernels;
  private JButton Refresh;
  private JCheckBox SOLACheckBox;
  private JCheckBox RLSCheckBox;
  private JCheckBox TargetFunctionCheckBox;

  private JCheckBox showErrorBars;
  private JCheckBox calculateWhileSliding;
  
  private JComboBox kernelTypeComboBox;
  private JRadioButton struct1Inversion;
  private JRadioButton struct2Inversion;
  
  private JPanel NoiseTitle;
  private JCheckBox NoiseCheckBox;
  private JLabel NoiseLabel;
  private JTextField NoiseTextField;
  private JSlider NoiseSlider;
  private JButton NoiseRefresh;

  private JLabel RLStitle;
  private JPanel LambdaTitle;
  private JLabel LambdaLabel;
  private JTextField LambdaTextField;
  private JSlider LambdaSlider;

  private JLabel SOLAtitle;
  private JPanel BetaTitle;
  private JLabel BetaLabel;
  private JTextField BetaTextField;
  private JSlider BetaSlider;
  private JPanel ThetaTitle;
  private JLabel ThetaLabel;
  private JTextField ThetaTextField;
  private JSlider ThetaSlider;
  private JPanel DeltaTitle;
  private JLabel DeltaLabel;
  private JTextField DeltaTextField;
  private JSlider DeltaSlider;

  private JPanel KernelTitle;
  private JLabel KernelLabel;
  private JTextField KernelTextField;
  private JSlider KernelSlider;
  private JPanel NGridPanel;
  private JLabel NGridLabel;
  private JTextField NGridTextField;

  private JPanel NSurfPanel;
  private JLabel NSurfLabel;
  private JTextField NSurfTextField;

  private JButton LoadTarget;
  private JButton ClearTarget;
  private JButton CalculateDF;
  private static final double ln10 = 2.302585092994046;

  /************************************************************************
  * A constructor.
  ************************************************************************/
  public StructuralInversionPage(DataCenter dataCenter) {
    this.dataCenter = dataCenter;
  }

  //////////////////////////////////////////////////////////////
  // These methods give the relationship between the slider
  // value and that of various parameters of the inversion.
  //////////////////////////////////////////////////////////////
  private int LambdaToSlider(double Lambda) {
    if (Lambda == 0.0) return -180;
    int result = (int)Math.round(Math.log(Lambda)*20.0/ln10);
    if (result < -180) result = -180;
    if (result > 20) result = 20;
    return result;
  }
  private double SliderToLambda() {
    return Math.pow(10.0,LambdaSlider.getValue()/20.0);
  }
  private int NoiseToSlider(double NoiseLevel) {
    if (NoiseLevel == 0.0) return -180;
    int result = (int)Math.round(Math.log(NoiseLevel)*20.0/ln10);
    if (result < -180) result = -180;
    if (result > 20) result = 20;
    return result;
  }
  private double SliderToNoise() {
    return Math.pow(10.0,NoiseSlider.getValue()/20.0);
  }
  private int BetaToSlider(double Beta) {
    if (Beta == 0.0) return -100;
    int result = (int)Math.round(Math.log(Beta)*25.0/ln10);
    if (result < -100) result = -100;
    if (result > 100) result = 100;
    return result;
  }
  private double SliderToBeta() {
    return Math.pow(10.0,BetaSlider.getValue()/25.0);
  }
  private int ThetaToSlider(double Theta) {
    if (Theta == 0.0) return -200;
    int result = (int)Math.round(Math.log(Theta/1.5)*50.0/ln10);
    if (result < -200) result = -200;
    if (result > 0) result = 0;
    return result;
  }
  private double SliderToTheta() {
    return 1.5*Math.pow(10.0,ThetaSlider.getValue()/50.0);
  }
  private int DeltaToSlider(double Delta) {
    if (Delta == 0.0) return -200;
    int result = (int)Math.round(Math.log(Delta)*50.0/ln10);
    if (result < -200) result = -200;
    if (result > 0) result = 0;
    return result;
  }
  private double SliderToDelta() {
    return Math.pow(10.0,DeltaSlider.getValue()/50.0);
  }

  private int XKernelToSlider(double xKernel) {
    if (dataCenter.gridSInversion == null) return 0;

    int istart = 0;
    int ifinish = dataCenter.gridSInversion.length - 1;
    int inext;
    while((istart+1) < ifinish) {
      inext = (istart+ifinish)/2;
      if (dataCenter.gridSInversion[inext] > xKernel) {
        ifinish = inext;
      } else {
        istart = inext;
      }
    }

    if ((xKernel-dataCenter.gridSInversion[istart]) < (dataCenter.gridSInversion[ifinish]-xKernel)) {
      return istart;
    } else {
      return ifinish;
    }
  }

  private double SliderToXKernel () {
    if (dataCenter.gridSInversion == null) return 0.0;
    return dataCenter.gridSInversion[KernelSlider.getValue()];
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the noise slider and refresh button
  /////////////////////////////////////////////////////////////////////////////
  private void initNoiseButtons() {

    double NoiseLevel = dataCenter.SNoiseLevel;
    boolean NoiseSelected = dataCenter.SNoiseSelected;
    
    NoiseLabel = new JLabel(" Noise level: ");
    NoiseLabel.setEnabled(NoiseSelected);
    NoiseTextField = new JTextField(String.format("%.4e",NoiseLevel));
    NoiseTextField.setEnabled(NoiseSelected);
    NoiseTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double Noise = Utilities.readDouble(NoiseTextField.getText(),
                       "noise level",0.0,1e30,NoiseTextField);
        if (Noise == null) {
          NoiseTextField.setText(String.format("%.4e",dataCenter.SNoiseLevel));
        } else {
          // I use this home-made method to change the slider without
          // triggering the ChangeListener:
          ChangeJSliderValue(NoiseSlider, NoiseToSlider(Noise));
          NoiseTextField.setText(String.format("%.4e",Noise));
          dataCenter.SNoiseLevel = Noise;
          boolean averagingKernelsUpToDate = true;
          if (!dataCenter.structInversionUpToDate) {
            updateInversion(NoiseTextField);
            averagingKernelsUpToDate = false;
          } else {
            if (my_inversion.canInvert) {
              my_inversion.RLSinversion();
              my_inversion.SOLAinversion1();
              my_inversion.SOLAinversion2();
            }
          }
          replotInvertedFunctions();
          if (!averagingKernelsUpToDate) replotAllKernels();
        }
      }
    });
    NoiseTitle = new JPanel();
    NoiseTitle.setLayout(new GridLayout(1,2));
    NoiseTitle.add(NoiseLabel);
    NoiseTitle.add(NoiseTextField);

    NoiseSlider = new JSlider(JSlider.HORIZONTAL,-180,20,NoiseToSlider(NoiseLevel));
    NoiseSlider.setEnabled(NoiseSelected);
    NoiseSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double NoiseLevel = SliderToNoise();
        NoiseTextField.setText(String.format("%.4e",NoiseLevel));
        if ((!calculateWhileSliding.isSelected())&&
            (NoiseSlider.getModel().getValueIsAdjusting())) return;
        boolean averagingKernelsUpToDate = true;
        dataCenter.SNoiseLevel = NoiseLevel;
        if (!dataCenter.structInversionUpToDate) {
          updateInversion(NoiseSlider);
          averagingKernelsUpToDate = false;
        } else {
          if (my_inversion.canInvert) {
            my_inversion.RLSinversion();
            my_inversion.SOLAinversion1();
            my_inversion.SOLAinversion2();
          }
        }
        replotInvertedFunctions();
        if (!averagingKernelsUpToDate) replotAllKernels();
      }
    });

    NoiseRefresh = new JButton("New noise array");
    NoiseRefresh.setEnabled(NoiseSelected);
    NoiseRefresh.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        boolean averagingKernelsUpToDate = true;
        if (!dataCenter.structInversionUpToDate) {
          updateInversion(NoiseRefresh);
          averagingKernelsUpToDate = false;
        } else {
          if (my_inversion.canInvert) {
            my_inversion.initNoise();
            my_inversion.RLSinversion();
            my_inversion.SOLAinversion1();
            my_inversion.SOLAinversion2();
          }
        }
        replotInvertedFunctions();
        if (!averagingKernelsUpToDate) replotAllKernels();
      }
    });
    
    NoiseCheckBox = new JCheckBox("Noise",NoiseSelected);
    NoiseCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.SNoiseSelected = NoiseCheckBox.isSelected();
        NoiseLabel.setEnabled(NoiseCheckBox.isSelected());
        NoiseTextField.setEnabled(NoiseCheckBox.isSelected());
        NoiseSlider.setEnabled(NoiseCheckBox.isSelected());
        NoiseRefresh.setEnabled(NoiseCheckBox.isSelected());
        boolean averagingKernelsUpToDate = true;
        if (!dataCenter.structInversionUpToDate) {
          updateInversion(NoiseCheckBox);
          averagingKernelsUpToDate = false;
        } else {
          if (my_inversion.canInvert) {
            my_inversion.RLSinversion();
            my_inversion.SOLAinversion1();
            my_inversion.SOLAinversion2();
          }
        }
        replotInvertedFunctions();
        if (!averagingKernelsUpToDate) replotAllKernels();
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the RLS slider
  /////////////////////////////////////////////////////////////////////////////
  private void initRLSslider() {
    double Lambda = dataCenter.SLambda;
    RLStitle = new JLabel("RLS",SwingConstants.CENTER);
    LambdaSlider = new JSlider(JSlider.HORIZONTAL,-180,20,LambdaToSlider(Lambda));
    LambdaLabel = new JLabel(" Lambda: ");
    
    LambdaTextField = new JTextField(String.format("%.4e",Lambda));
    LambdaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double Lambda = Utilities.readDouble(LambdaTextField.getText(),
                       "value of Lambda",0.0,1e30,LambdaTextField);
        if (Lambda == null) {
          LambdaTextField.setText(String.format("%.4e",dataCenter.SLambda));
        } else {
          // I use this home-made method to change the slider without
          // triggering the ChangeListener:
          ChangeJSliderValue(LambdaSlider, LambdaToSlider(Lambda));
          LambdaTextField.setText(String.format("%.4e",Lambda));
          dataCenter.SLambda = Lambda;
          if (!dataCenter.structInversionUpToDate) {
            updateInversion(LambdaTextField);
          } else {
            if (my_inversion.canInvert) my_inversion.RLSinversion();
          }
          replotInvertedFunctions();
          replotAllKernels();
        }
      }
    });
    
    LambdaTitle = new JPanel();
    LambdaTitle.setLayout(new GridLayout(1,2));
    LambdaTitle.add(LambdaLabel);
    LambdaTitle.add(LambdaTextField);
    
    LambdaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double Lambda = SliderToLambda();
        LambdaTextField.setText(String.format("%.4e",Lambda));
        if ((!calculateWhileSliding.isSelected())&&
            (LambdaSlider.getModel().getValueIsAdjusting())) return;
        dataCenter.SLambda = Lambda;
        if (!dataCenter.structInversionUpToDate) {
          updateInversion(LambdaSlider);
        } else {
          if (my_inversion.canInvert) my_inversion.RLSinversion();
        }
        replotInvertedFunctions();
        replotAllKernels();
      }
    });
  }
    
  /////////////////////////////////////////////////////////////////////////////
  // Setting up the SOLA sliders
  /////////////////////////////////////////////////////////////////////////////
  private void initSOLAsliders() {
    SOLAtitle = new JLabel("SOLA",SwingConstants.CENTER);

    // later on, struct1Inversion will be initialised to true, hence this choice:
    double Beta = dataCenter.Beta1;
    double Theta = dataCenter.STheta1;
    double Delta = dataCenter.SDelta1;

    BetaLabel = new JLabel(" Beta: ");
    BetaTextField = new JTextField(String.format("%.4e",Beta));
    BetaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double Beta = Utilities.readDouble(BetaTextField.getText(),
                       "value of Beta",0.0,1e30,BetaTextField);
        if (Beta == null) {
          if (struct1Inversion.isSelected()) {
            BetaTextField.setText(String.format("%.4e",dataCenter.Beta1));
          } else {
            BetaTextField.setText(String.format("%.4e",dataCenter.Beta2));
          }
        } else {
          // I use this home-made method to change the slider without
          // triggering the ChangeListener:
          ChangeJSliderValue(BetaSlider, BetaToSlider(Beta));
          BetaTextField.setText(String.format("%.4e",Beta));
          if (struct1Inversion.isSelected()) {
            dataCenter.Beta1 = Beta;
          } else {
            dataCenter.Beta2 = Beta;
          }
          if (!dataCenter.structInversionUpToDate) {
            updateInversion(BetaTextField);
          } else {
            if (my_inversion.canInvert) {
              if (struct1Inversion.isSelected()) {
                my_inversion.SOLAinversion1();
              } else {
                my_inversion.SOLAinversion2();
              }
            }
          }
          replotInvertedFunctions();
          replotAllKernels();
        }
      }
    });

    BetaTitle = new JPanel();
    BetaTitle.setLayout(new GridLayout(1,2));
    BetaTitle.add(BetaLabel);
    BetaTitle.add(BetaTextField);

    BetaSlider = new JSlider(JSlider.HORIZONTAL,-100,100,BetaToSlider(Beta));
    BetaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double Beta = SliderToBeta();
        BetaTextField.setText(String.format("%.4e",Beta));
        if ((!calculateWhileSliding.isSelected())&&
            (BetaSlider.getModel().getValueIsAdjusting())) return;
        if (struct1Inversion.isSelected()) {
          dataCenter.Beta1 = Beta;
        } else {
          dataCenter.Beta2 = Beta;
        }
        if (!dataCenter.structInversionUpToDate) {
          updateInversion(BetaSlider);
        } else {
          if (my_inversion.canInvert){
            if (struct1Inversion.isSelected()) {
              my_inversion.SOLAinversion1();
            } else {
              my_inversion.SOLAinversion2();
            }
          }
        }
        replotInvertedFunctions();
        replotAllKernels();
      }
    });
    
    ThetaLabel = new JLabel(" Theta: ");
    ThetaTextField = new JTextField(String.format("%.4e",Theta));
    ThetaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double Theta = Utilities.readDouble(ThetaTextField.getText(),
                       "value of Theta",0.0,Math.PI/2.0,ThetaTextField);
        if (Theta == null) {
          if (struct1Inversion.isSelected()) {
            ThetaTextField.setText(String.format("%.4e",dataCenter.STheta1));
          } else {
            ThetaTextField.setText(String.format("%.4e",dataCenter.STheta2));
          }
        } else {
          // I use this home-made method to change the slider without
          // triggering the ChangeListener:
          ChangeJSliderValue(ThetaSlider, ThetaToSlider(Theta));
          ThetaTextField.setText(String.format("%.4e",Theta));
          if (struct1Inversion.isSelected()) {
            dataCenter.STheta1 = Theta;
          } else {
            dataCenter.STheta2 = Theta;
          }
          if (!dataCenter.structInversionUpToDate) {
            updateInversion(ThetaTextField);
          } else {
            if (my_inversion.canInvert) {
              if (struct1Inversion.isSelected()) {
                my_inversion.SOLAinversion1();
              } else {
                my_inversion.SOLAinversion2();
              }
            }
          }
          replotInvertedFunctions();
          replotAllKernels();
        }
      }
    });

    ThetaTitle = new JPanel();
    ThetaTitle.setLayout(new GridLayout(1,2));
    ThetaTitle.add(ThetaLabel);
    ThetaTitle.add(ThetaTextField);

    ThetaSlider = new JSlider(JSlider.HORIZONTAL,-200,0,ThetaToSlider(Theta));
    ThetaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double Theta = SliderToTheta();
        ThetaTextField.setText(String.format("%.4e",Theta));
        if ((!calculateWhileSliding.isSelected())&&
            (ThetaSlider.getModel().getValueIsAdjusting())) return;
        if (struct1Inversion.isSelected()) {
          dataCenter.STheta1 = Theta;
        } else {
          dataCenter.STheta2 = Theta;
        }
        if (!dataCenter.structInversionUpToDate) {
          updateInversion(ThetaSlider);
        } else {
          if (my_inversion.canInvert){
            if (struct1Inversion.isSelected()) {
              my_inversion.SOLAinversion1();
            } else {
              my_inversion.SOLAinversion2();
            }
          }
        }
        replotInvertedFunctions();
        replotAllKernels();
      }
    });
        
    DeltaSlider = new JSlider(JSlider.HORIZONTAL,-200,0,DeltaToSlider(Delta));
    DeltaLabel = new JLabel(" Delta: ");
    DeltaTextField = new JTextField(String.format("%.4e",Delta));
    DeltaTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Double Delta = Utilities.readDouble(DeltaTextField.getText(),
                       "value of Delta",0.0,1e2,DeltaTextField);
        if (Delta == null) {
          if (struct1Inversion.isSelected()) {
            DeltaTextField.setText(String.format("%.4e",dataCenter.SDelta1));
          } else {
            DeltaTextField.setText(String.format("%.4e",dataCenter.SDelta2));
          }
        } else {
          // I use this home-made method to change the slider without
          // triggering the ChangeListener:
          ChangeJSliderValue(DeltaSlider, DeltaToSlider(Delta));
          DeltaTextField.setText(String.format("%.4e",Delta));
          if (struct1Inversion.isSelected()) {
            dataCenter.SDelta1 = Delta;
          } else {
            dataCenter.SDelta2 = Delta;
          }
          Boolean averagingKernelsUpToDate = true;
          if (!dataCenter.structInversionUpToDate) {
            updateInversion(DeltaTextField);
          } else {
            if (my_inversion.canInvert) {
              if (struct1Inversion.isSelected()) {
                my_inversion.initDelta1SOLA();
                my_inversion.SOLAinversion1();
              } else {
                my_inversion.initDelta2SOLA();
                my_inversion.SOLAinversion2();
              }
            }
          }
          replotInvertedFunctions();
          replotAllKernels();
        }
      }
    });

    DeltaTitle = new JPanel();
    DeltaTitle.setLayout(new GridLayout(1,2));
    DeltaTitle.add(DeltaLabel);
    DeltaTitle.add(DeltaTextField);

    DeltaSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double Delta = SliderToDelta();
        DeltaTextField.setText(String.format("%.4e",Delta));
        if ((!calculateWhileSliding.isSelected())&&
            (DeltaSlider.getModel().getValueIsAdjusting())) return;
        if (struct1Inversion.isSelected()) {
          dataCenter.SDelta1 = Delta;
        } else {
          dataCenter.SDelta2 = Delta;
        }
        if (!dataCenter.structInversionUpToDate) {
          updateInversion(DeltaSlider);
        } else {
          if (my_inversion.canInvert) {
            if (struct1Inversion.isSelected()) {
              my_inversion.initDelta1SOLA();
              my_inversion.SOLAinversion1();
            } else {
              my_inversion.initDelta2SOLA();
              my_inversion.SOLAinversion2();
            }
          }
        }
        replotInvertedFunctions();
        replotAllKernels();
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise buttons for the selection of kernel pairs, and specific kernels.
  /////////////////////////////////////////////////////////////////////////////
  private void initStructKernelButtons() {
    kernelTypeComboBox = new JComboBox();
    kernelTypeComboBox.addItem("c2-rho kernels");
    kernelTypeComboBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        String newKernelType = (String)kernelTypeComboBox.getSelectedItem();
        if (!dataCenter.kernelType.equals(newKernelType)) {
          dataCenter.kernelType = newKernelType;
          dataCenter.findStructKernels();
          updateInversion(kernelTypeComboBox);
          replotInvertedFunctions();
          replotAllKernels();
        }
      }
    });

    struct1Inversion = new JRadioButton("1st structural function",true);
    struct2Inversion = new JRadioButton("2nd structural function");
    ActionListener l = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (!dataCenter.structInversionUpToDate) 
          updateInversion((JRadioButton)e.getSource());
        if (struct1Inversion.isSelected()) {
          ChangeJSliderValue(BetaSlider,BetaToSlider(dataCenter.Beta1));
          BetaTextField.setText(String.format("%.4e",dataCenter.Beta1));
          ChangeJSliderValue(ThetaSlider,ThetaToSlider(dataCenter.STheta1));
          ThetaTextField.setText(String.format("%.4e",dataCenter.STheta1));
          ChangeJSliderValue(DeltaSlider,DeltaToSlider(dataCenter.SDelta1));
          DeltaTextField.setText(String.format("%.4e",dataCenter.SDelta1));
        } else {
          ChangeJSliderValue(BetaSlider,BetaToSlider(dataCenter.Beta2));
          BetaTextField.setText(String.format("%.4e",dataCenter.Beta2));
          ChangeJSliderValue(ThetaSlider,ThetaToSlider(dataCenter.STheta2));
          ThetaTextField.setText(String.format("%.4e",dataCenter.STheta2));
          ChangeJSliderValue(DeltaSlider,DeltaToSlider(dataCenter.SDelta2));
          DeltaTextField.setText(String.format("%.4e",dataCenter.SDelta2));
        }
        replotInvertedFunctions();
        replotAllKernels();
      }
    };
    struct1Inversion.addActionListener(l);
    struct2Inversion.addActionListener(l);
    ButtonGroup group2 = new ButtonGroup();
    group2.add(struct1Inversion);
    group2.add(struct2Inversion);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Setting up the Kernel slider:   
  /////////////////////////////////////////////////////////////////////////////
  private void initKernelSlider() {
    KernelLabel = new JLabel(" Position, r/R: ");
    KernelTextField = new JTextField(String.format("%.4f",0.0));
    KernelTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (!dataCenter.structInversionUpToDate) {
          updateInversion(KernelTextField);
          replotInvertedFunctions();
        }
        if (!my_inversion.canInvert) {
          KernelTextField.setText(String.format("%.4f",0.0));
          replotAllKernels();
          return;
        }
        Double XKernel = Utilities.readDouble(KernelTextField.getText(),
                        "value of Kernel",dataCenter.gridSInversion[0],
                        dataCenter.gridSInversion[dataCenter.gridSInversion.length-1],
                        KernelTextField);
        if (XKernel == null) {
          KernelTextField.setText(String.format("%.4f",SliderToXKernel()));
        } else {
          // I use this home-made method to change the slider without
          // triggering the ChangeListener:
          ChangeJSliderValue(KernelSlider, XKernelToSlider(XKernel));
          KernelTextField.setText(String.format("%.4f",SliderToXKernel()));
          replotAllKernels();
        }
      }
    });

    KernelTitle = new JPanel();
    KernelTitle.setLayout(new GridLayout(1,2));
    KernelTitle.add(KernelLabel);
    KernelTitle.add(KernelTextField);


    // VERY IMPORTANT: the number of positions on the Kernel Slider MUST
    // correspond to the number of points in gridInversion
    KernelSlider = new JSlider(JSlider.HORIZONTAL,0,dataCenter.n_gridS-1,0);
    KernelSlider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        KernelTextField.setText(String.format("%.4f",SliderToXKernel()));
        if (!dataCenter.structInversionUpToDate) {
          updateInversion(KernelSlider);
          replotInvertedFunctions();
        }
        replotAllKernels();
      }
    });

    NGridLabel = new JLabel(" # of grid points: ");
    NGridTextField = new JTextField(String.format("%d",dataCenter.n_gridS));
    NGridTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Integer NGrid = Utilities.readInt(NGridTextField.getText(), "number of grid points",
                                10,999,NGridTextField);
        if (NGrid == null) {
          NGridTextField.setText(String.format("%d",dataCenter.n_gridS));
          return;
        }
        if (!NGrid.equals(dataCenter.n_gridS)) {
          KernelSlider.setMaximum(NGrid-1);
          dataCenter.n_gridS = NGrid;
          if (dataCenter.structInversionUpToDate&&!my_inversion.canInvert)
            return;
          updateInversion(NGridTextField);
          replotInvertedFunctions();
          replotAllKernels();
        }
      }
    });
    NGridPanel = new JPanel();
    NGridPanel.setLayout(new GridLayout(1,2));
    NGridPanel.add(NGridLabel);
    NGridPanel.add(NGridTextField);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Initialise the text field and label that allows one to adjust the number
  // of Legendre polynomials to fit/remove surface effects.
  /////////////////////////////////////////////////////////////////////////////
  private void initNSurfPanel() {
    NSurfLabel = new JLabel(" # of surf. poly.: ");
    NSurfTextField = new JTextField(String.format("%d",dataCenter.n_surf));
    NSurfTextField.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Integer NSurf = Utilities.readInt(NSurfTextField.getText(),
                                "number of polynomials",
                                0,50,NGridTextField);
        if (NSurf == null) {
          NSurfTextField.setText(String.format("%d",dataCenter.n_surf));
          return;
        }
        if (!NSurf.equals(dataCenter.n_surf)) {
          dataCenter.n_surf = NSurf;
          if (dataCenter.structInversionUpToDate&&!my_inversion.canInvert)
            return;
          updateInversion(NSurfTextField);
          replotInvertedFunctions();
          replotAllKernels();
        }
      }
    });
    NSurfPanel = new JPanel();
    NSurfPanel.setLayout(new GridLayout(1,2));
    NSurfPanel.add(NSurfLabel);
    NSurfPanel.add(NSurfTextField);
  }

  /////////////////////////////////////////////////////////////////////////////
  // Setting up the refresh button and the RLS, SOLA and Target function check
  // boxes.
  /////////////////////////////////////////////////////////////////////////////
  private void initCheckBoxes() {
    Refresh = new JButton("Update Inversion");
    RLSCheckBox = new JCheckBox("RLS (red lines)",true);
    SOLACheckBox = new JCheckBox("SOLA (blue lines)",true);
    TargetFunctionCheckBox = new JCheckBox("Target function (black line)",true);
    showErrorBars = new JCheckBox("Show error bars",true);

    Refresh.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        updateInversion(Refresh);
        replotInvertedFunctions();
        replotAllKernels();
      }
    });

    RLSCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        RLStitle.setEnabled(RLSCheckBox.isSelected());
        LambdaLabel.setEnabled(RLSCheckBox.isSelected());
        LambdaTextField.setEnabled(RLSCheckBox.isSelected());
        LambdaSlider.setEnabled(RLSCheckBox.isSelected());
        if (!dataCenter.structInversionUpToDate) updateInversion(RLSCheckBox);
        replotInvertedFunctions();
        replotAllKernels();
      }
    });
          
    SOLACheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        SOLAtitle.setEnabled(SOLACheckBox.isSelected());
        BetaLabel.setEnabled(SOLACheckBox.isSelected());
        BetaTextField.setEnabled(SOLACheckBox.isSelected());
        BetaSlider.setEnabled(SOLACheckBox.isSelected());
        ThetaLabel.setEnabled(SOLACheckBox.isSelected());
        ThetaTextField.setEnabled(SOLACheckBox.isSelected());
        ThetaSlider.setEnabled(SOLACheckBox.isSelected());
        DeltaLabel.setEnabled(SOLACheckBox.isSelected());
        DeltaTextField.setEnabled(SOLACheckBox.isSelected());
        DeltaSlider.setEnabled(SOLACheckBox.isSelected());
        if (!dataCenter.structInversionUpToDate) updateInversion(SOLACheckBox);
        replotInvertedFunctions();
        replotAllKernels();
      }
    });

    TargetFunctionCheckBox.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (!dataCenter.structInversionUpToDate) updateInversion(TargetFunctionCheckBox);
        replotInvertedFunctions();
        replotAllKernels();
      }
    });

    showErrorBars.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (!dataCenter.structInversionUpToDate) {
          updateInversion(showErrorBars);
          replotAllKernels();
        }
        replotInvertedFunctions();
      }
    });

  }

  /////////////////////////////////////////////////////////////////////////////
  // init buttons which deal with a target function
  /////////////////////////////////////////////////////////////////////////////
  private void initTargetFunctionButtons() {
    LoadTarget = new JButton("Load target functions");
    LoadTarget.setEnabled(!dataCenter.isApplet);
    LoadTarget.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        // open in current directory:
        JFileChooser fc = new JFileChooser(dataCenter.currentDirectory);

        // show the filechooser
        int result = fc.showOpenDialog(LoadTarget);  // put parent component

        // get filename
        if(result == JFileChooser.APPROVE_OPTION) {
          File aFile = fc.getSelectedFile();

          // check to see if file exists:
          if (aFile.exists()) {
            loadStructTargets(aFile);
            replotInvertedFunctions();
            dataCenter.currentDirectory = aFile.getParentFile();
          } else {
            JOptionPane.showMessageDialog(
              LoadTarget,
              "File \""+aFile.getName()+"\" doesn't exist.",
              "File not found",
              JOptionPane.WARNING_MESSAGE
            );
          }
        }
      }
    });
    
    ClearTarget = new JButton("Clear target functions");
    ClearTarget.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        dataCenter.targetStruct1 = null;
        dataCenter.targetStruct2 = null;
        dataCenter.gridTargetStruct = null;
        dataCenter.targetStructLoaded = false;
        replotInvertedFunctions();
      }
    });

    CalculateDF = new JButton("Find \u0394f/f from target functions");
    CalculateDF.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        ((FrequencyShiftsPage)dataCenter.frequencyShiftsPage).calculateDF(CalculateDF);
      }
    });
  }

  /////////////////////////////////////////////////////////////////////////////
  // This changes the value on JSlider without activating its change listeners.
  // This can be done by removing the change listeners, changing the value on
  // the JSlider, then adding the change listeners back (it does seem like
  // unecessary complications...).
  /////////////////////////////////////////////////////////////////////////////
  private static void ChangeJSliderValue(JSlider aSlider, int aValue) {
    ChangeListener[] cl = aSlider.getChangeListeners();
    for(int i=0; i<cl.length; i++) aSlider.removeChangeListener(cl[i]);
    aSlider.setValue(aValue);
    for(int i=0; i<cl.length; i++) aSlider.addChangeListener(cl[i]);
  }
  
  /***************************************************************************
  *  This updates the inversion.
  ***************************************************************************/
  public void updateInversion(JComponent c) {
    my_inversion.init(c);
    numberOfKernels.setText(String.format(" This inversion uses %d kernels",
                            my_inversion.n_inputKernels));
    // no need to give an error message if canInvert = false, since
    // my_inversion.init(c) already gives error messages.
    if (my_inversion.canInvert) {
      my_inversion.RLSinversion();
      my_inversion.SOLAinversion1();
      my_inversion.SOLAinversion2();
    }
  }

  /************************************************************************
  * This initialises the program.
  ************************************************************************/
  public void initProgram() {
    graphInvertedFunction = new IPlot("Inversion results","Position, r/R", "Profile",7);
    graphInvertedFunction.setIOenabled(!dataCenter.isApplet);
    graphAveragingKernel = new IPlot("Averaging kernels","Position, r/R", "Kernel",2);
    graphAveragingKernel.setIOenabled(!dataCenter.isApplet);
    graphCrossTermKernel = new IPlot("Cross-term kernels","Position, r/R", "Kernel",2);
    graphCrossTermKernel.setIOenabled(!dataCenter.isApplet);
    graphContainer = new JPanel();
    graphContainer.setLayout(new GridLayout(3,1));
    graphContainer.add(graphInvertedFunction);
    graphContainer.add(graphAveragingKernel);
    graphContainer.add(graphCrossTermKernel);

    setLayout(new BorderLayout());
    add("Center", graphContainer);

    // init the inversion
    my_inversion = new StructuralInversion(dataCenter);

    calculateWhileSliding = new JCheckBox("Calculate while using sliders",false);
    initNoiseButtons();
    initRLSslider();
    initSOLAsliders();
    initStructKernelButtons();
    initKernelSlider();
    initCheckBoxes();
    initTargetFunctionButtons();
    initNSurfPanel();
    numberOfKernels = new JLabel(String.format(" This inversion uses %d kernels",
                                 my_inversion.n_inputKernels));
    
    SliderContainer = new JPanel();
    SliderContainer.setLayout(new GridLayout(39,1));
    SliderContainer.add(numberOfKernels);
    SliderContainer.add(Refresh);
    SliderContainer.add(RLSCheckBox);
    SliderContainer.add(SOLACheckBox);
    SliderContainer.add(TargetFunctionCheckBox);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(showErrorBars);
    SliderContainer.add(calculateWhileSliding);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(kernelTypeComboBox);
    SliderContainer.add(struct1Inversion);
    SliderContainer.add(struct2Inversion);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(NoiseCheckBox);
    SliderContainer.add(NoiseTitle);
    SliderContainer.add(NoiseSlider);
    SliderContainer.add(NoiseRefresh);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(RLStitle);
    SliderContainer.add(LambdaTitle);
    SliderContainer.add(LambdaSlider);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(SOLAtitle);
    SliderContainer.add(BetaTitle);
    SliderContainer.add(BetaSlider);
    SliderContainer.add(ThetaTitle);
    SliderContainer.add(ThetaSlider);
    SliderContainer.add(DeltaTitle);
    SliderContainer.add(DeltaSlider);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(KernelTitle);
    SliderContainer.add(KernelSlider);
    SliderContainer.add(NGridPanel);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(NSurfPanel);
    SliderContainer.add(new JSeparator(SwingConstants.HORIZONTAL));
    SliderContainer.add(LoadTarget);
    SliderContainer.add(ClearTarget);
    SliderContainer.add(CalculateDF);
    add("East",SliderContainer);
       
    graphInvertedFunction.repaint();
    graphAveragingKernel.repaint();
  }

  /////////////////////////////////////////////////////////
  // load a file with target structural functions:
  /////////////////////////////////////////////////////////
  private void loadStructTargets(File aFile){
    BufferedReader f;
    String line;
    Boolean skipLine = false;
    ArrayList<Double> x = new ArrayList<Double>();
    ArrayList<Double> y = new ArrayList<Double>();
    ArrayList<Double> z = new ArrayList<Double>();
    try {
      f = new BufferedReader(new FileReader(aFile));
      while (true) {
        line = f.readLine();     // read a line of text
        if (line == null) break; // test for the end of the buffer
        line = trim(line);       // remove comments
        StringTokenizer t = new StringTokenizer(line," ");
        // only read lines with EXACTLY 3 numbers
        if (t.countTokens() == 3) {
          x.add(Double.parseDouble(t.nextToken()));
          y.add(Double.parseDouble(t.nextToken()));
          z.add(Double.parseDouble(t.nextToken()));
        } else {
          if (t.countTokens() != 0) skipLine = true;
        }
      }
      f.close();
    } catch(Exception e) {
      JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "Unable to read file \""+aFile.getName()+"\"",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }
    if (x.size() == 0) {
      JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "No data in file",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
      return;
    }

    // perform a basic check on the grid:
    for(int i=1; i<x.size(); i++) {
      if (x.get(i).compareTo(x.get(i-1)) < 0) {
        JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "Please define the target functions\n"+
                    "on a strictly ascending grid.",
                    "Error",
                    JOptionPane.ERROR_MESSAGE
                  );
        return;
      }
    }
    
    if (skipLine) {
      JOptionPane.showMessageDialog(
                    LoadTarget,  // the parent component
                    "Some lines were skipped\nwhen reading the file.",
                    "Warning",
                    JOptionPane.WARNING_MESSAGE
                  );
    }
    dataCenter.gridTargetStruct = new double[x.size()];
    dataCenter.targetStruct1 = new double[y.size()];
    dataCenter.targetStruct2 = new double[z.size()];
    for(int i=0; i<x.size(); i++) {
      dataCenter.gridTargetStruct[i] = (double)(x.get(i));
      dataCenter.targetStruct1[i] = (double)(y.get(i));
      dataCenter.targetStruct2[i] = (double)(z.get(i));
    }
    dataCenter.targetStructLoaded = true;
  }

  /////////////////////////////////////////////////////////
  // This remove comments (i.e. anything following a "#")
  // from a string and is used by the loadRotaTarget method:
  /////////////////////////////////////////////////////////
  private static String trim(String str) {
    int len = str.length();
    int i = 0;
    while ((i < len)&&(str.charAt(i) != '#')) i++;
    return str.substring(0,i);
  }
     
  /***************************************************************************
  *  Plot the inverted and target rotation functions
  ***************************************************************************/
  private void replotInvertedFunctions() {
    if (TargetFunctionCheckBox.isSelected()&&dataCenter.targetStructLoaded) {
      if (struct1Inversion.isSelected()) {
        graphInvertedFunction.replaceDataSet(6,dataCenter.gridTargetStruct,
                                               dataCenter.targetStruct1,Plot.solid);
      } else {
        graphInvertedFunction.replaceDataSet(6,dataCenter.gridTargetStruct,
                                               dataCenter.targetStruct2,Plot.solid);
      } 
    } else {
      graphInvertedFunction.nullify(6);
    }

    if (dataCenter.structInversionUpToDate&&my_inversion.canInvert) {
      int n_grid = dataCenter.gridSInversion.length;
      double[] aux = new double[n_grid];

      if (RLSCheckBox.isSelected()) {
        if (struct1Inversion.isSelected()) {
          graphInvertedFunction.replaceDataSet(0,dataCenter.gridSInversion,my_inversion.f1_invertedRLS,
                                               Plot.solid,Color.red);
        } else {
          graphInvertedFunction.replaceDataSet(0,dataCenter.gridSInversion,my_inversion.f2_invertedRLS,
                                               Plot.solid,Color.red);
        }
      } else {
        graphInvertedFunction.nullify(0);
      }
      
      if (RLSCheckBox.isSelected()&&showErrorBars.isSelected()) {
        if (struct1Inversion.isSelected()) {
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f1_invertedRLS[i]+my_inversion.RLS_errorBars1[i];
          graphInvertedFunction.replaceDataSet(1,dataCenter.gridSInversion,aux,Plot.dotted,Color.red);
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f1_invertedRLS[i]-my_inversion.RLS_errorBars1[i];
          graphInvertedFunction.replaceDataSet(2,dataCenter.gridSInversion,aux,Plot.dotted,Color.red);
        } else {
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f2_invertedRLS[i]+my_inversion.RLS_errorBars2[i];
          graphInvertedFunction.replaceDataSet(1,dataCenter.gridSInversion,aux,Plot.dotted,Color.red);
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f2_invertedRLS[i]-my_inversion.RLS_errorBars2[i];
          graphInvertedFunction.replaceDataSet(2,dataCenter.gridSInversion,aux,Plot.dotted,Color.red);
        }
      } else {
        graphInvertedFunction.nullify(1);
        graphInvertedFunction.nullify(2);
      }

      if (SOLACheckBox.isSelected()) {
        if (struct1Inversion.isSelected()) {
          graphInvertedFunction.replaceDataSet(3,dataCenter.gridSInversion,my_inversion.f1_invertedSOLA,
                                               Plot.solid,Color.blue);
        } else {
          graphInvertedFunction.replaceDataSet(3,dataCenter.gridSInversion,my_inversion.f2_invertedSOLA,
                                               Plot.solid,Color.blue);
        }
      } else {
        graphInvertedFunction.nullify(3);
      }
      
      if (SOLACheckBox.isSelected()&&showErrorBars.isSelected()) {
        if (struct1Inversion.isSelected()) {
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f1_invertedSOLA[i]+my_inversion.SOLA_errorBars1[i];
          graphInvertedFunction.replaceDataSet(4,dataCenter.gridSInversion,aux,Plot.dotted,Color.blue);
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f1_invertedSOLA[i]-my_inversion.SOLA_errorBars1[i];
          graphInvertedFunction.replaceDataSet(5,dataCenter.gridSInversion,aux,Plot.dotted,Color.blue);
        }else {
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f2_invertedSOLA[i]+my_inversion.SOLA_errorBars2[i];
          graphInvertedFunction.replaceDataSet(4,dataCenter.gridSInversion,aux,Plot.dotted,Color.blue);
          for(int i=0; i< n_grid; i++) aux[i] = my_inversion.f2_invertedSOLA[i]-my_inversion.SOLA_errorBars2[i];
          graphInvertedFunction.replaceDataSet(5,dataCenter.gridSInversion,aux,Plot.dotted,Color.blue);
        }
      } else {
        graphInvertedFunction.nullify(4);
        graphInvertedFunction.nullify(5);
      }
    } else {
      for(int i=0; i<=5; i++) graphInvertedFunction.nullify(i);
    }

    graphInvertedFunction.findRange();

    if (dataCenter.structInversionUpToDate&&my_inversion.canInvert) {
      graphInvertedFunction.setXmin(dataCenter.gridSInversion[0]);
      graphInvertedFunction.setXmax(dataCenter.gridSInversion[dataCenter.gridSInversion.length-1]);
    }
    graphInvertedFunction.repaint();
  }

  /***************************************************************************
  *  Plot all kernels
  ***************************************************************************/
  private void replotAllKernels() {
    replotAveragingKernels();
    replotCrossTermKernels();
  }

  /***************************************************************************
  *  Plot averaging kernels
  ***************************************************************************/
  private void replotAveragingKernels() {
    // quick exit if possible:
    if (dataCenter.gridSInversion == null) {
      graphAveragingKernel.nullify();
      graphAveragingKernel.repaint();
      return;
    }
    
    // number of grid points
    int n_grid = dataCenter.gridSInversion.length;
    
    // this specifies which kernel to plot:
    int iKernel = KernelSlider.getValue();
    
    // update KernelTextField (just in case):
    KernelTextField.setText(String.format("%.4f",SliderToXKernel()));

    // RLS averaging kernel:
    if (RLSCheckBox.isSelected()) {
      double[] y = new double[n_grid];
      if (struct1Inversion.isSelected()) {
        for(int i=0; i<n_grid; i++) y[i] = my_inversion.averagingKernels1RLS[i+n_grid*iKernel];
      } else {
        for(int i=0; i<n_grid; i++) y[i] = my_inversion.averagingKernels2RLS[i+n_grid*iKernel];
      }
      graphAveragingKernel.replaceDataSet(0,dataCenter.gridSInversion,y,Plot.solid,Color.red);
    } else {
      graphAveragingKernel.nullify(0);
    }
    
    // SOLA averaging kernel:
    if (SOLACheckBox.isSelected()) {
      double[] y = new double[n_grid];
      if (struct1Inversion.isSelected()) {
        for(int i=0; i<n_grid; i++) y[i] = my_inversion.averagingKernels1SOLA[i+n_grid*iKernel];
      } else {
        for(int i=0; i<n_grid; i++) y[i] = my_inversion.averagingKernels2SOLA[i+n_grid*iKernel];
      }
      graphAveragingKernel.replaceDataSet(1,dataCenter.gridSInversion,y,Plot.solid,Color.blue);
    } else {
      graphAveragingKernel.nullify(1);
    }

    graphAveragingKernel.findRange();
    graphAveragingKernel.repaint();
  }

  /***************************************************************************
  *  Plot cross-term kernels
  ***************************************************************************/
  private void replotCrossTermKernels() {
    // quick exit if possible:
    if (dataCenter.gridSInversion == null) {
      graphCrossTermKernel.nullify();
      graphCrossTermKernel.repaint();
      return;
    }
    
    // number of grid points
    int n_grid = dataCenter.gridSInversion.length;
    
    // this specifies which kernel to plot:
    int iKernel = KernelSlider.getValue();
    
    // update KernelTextField (just in case):
    KernelTextField.setText(String.format("%.4f",SliderToXKernel()));

    // RLS crossTerm kernel:
    if (RLSCheckBox.isSelected()) {
      double[] y = new double[n_grid];
      if (struct1Inversion.isSelected()) {
        for(int i=0; i<n_grid; i++) y[i] = my_inversion.crossTermKernels1RLS[i+n_grid*iKernel];
      } else {
        for(int i=0; i<n_grid; i++) y[i] = my_inversion.crossTermKernels2RLS[i+n_grid*iKernel];
      }
      graphCrossTermKernel.replaceDataSet(0,dataCenter.gridSInversion,y,Plot.solid,Color.red);
    } else {
      graphCrossTermKernel.nullify(0);
    }
    
    // SOLA crossTerm kernel:
    if (SOLACheckBox.isSelected()) {
      double[] y = new double[n_grid];
      if (struct1Inversion.isSelected()) {
        for(int i=0; i<n_grid; i++) y[i] = my_inversion.crossTermKernels1SOLA[i+n_grid*iKernel];
      } else {
        for(int i=0; i<n_grid; i++) y[i] = my_inversion.crossTermKernels2SOLA[i+n_grid*iKernel];
      }
      graphCrossTermKernel.replaceDataSet(1,dataCenter.gridSInversion,y,Plot.solid,Color.blue);
    } else {
      graphCrossTermKernel.nullify(1);
    }

    graphCrossTermKernel.findRange();
    graphCrossTermKernel.repaint();
  }
}
